/*
 * routines to evaluate the quadratic form v^t C^{-1} v 
 * and determinant |C| using a Cholesky decomposition.
 * HISTORY
 *    ??-??-??  MPH  Fortran version
 *    ??-??-01  KSM  C port
 */

/*  Time-stamp: <24 Jan 2002 11:42 GMT cquadform.c maising> */

#include <stdio.h>
#include <math.h>
#include "my_malloc2.h"
#ifndef SCALAPACK
#include "clapack.h"
#else
#include "cscalapack.h"
#ifdef MPI
#include <mpi.h> 
#endif
#endif
#include "cquadform.h"
#include "like_matvec.h"

#define MAX(x, y) (((x) > (y) ? (x) : (y)))
#define MIN(x, y) (((x) < (y) ? (x) : (y)))

#define ROOT 0			/* MPI root process */

/*-----*-----------------------------------------------------------------*/
/*************************************************************************
 * converts covariance matrix from sparse storage scheme to 
 * ScaLAPACK's distributed block-cyclic storage. 
 * (compare p1_p0()) 
 * PARAMS:
 *     icn  [I]  column index array for compressed column storage
 *     irn  [I]  row index array
 *     c    [I]  array elements
 *     nv   [I]  dimension of matrix
 *     lu   [I]  lu == 2: symmetrise matrix; lu == 0, 1: general matrix 
 *     desc [I]  ScaLAPACK matrix descriptor for distributed matrix
 *     pp0  [O]  pointer to output matrix
 *     lrow [I]  number of rows of local matrix
 *     lcol [I]  number of columns of local matrix
 *     flag [I]  flag: convert full matrix (0) or only southeastern quarter (1)
 *               in the second case, nv is half of the total matrix dimension
 */
void sca_p1_p0(int *icn, int *irn, double *c, int nv, int lu, int desc[], 
	       double ***pp0, int lrow, int lcol, int flag)
{
#ifdef SCALAPACK
  int i, j, p;

  (*pp0) = malloc_2dmatrix(lcol, lrow);

  for (i = 1; i <= lcol; i++) 
    for (j = 1; j <= lrow; j++) 
      (*pp0)[i][j] = 0.;

  if (! flag) {
    for (j = 1; j <= nv; j++) 
      for (p = icn[j]; p <= icn[j+1]-1; p++) {
	i = irn[p];
	cpdelset((*pp0)[C_OFFSET]+C_OFFSET, i, j, desc, c[p]); /* upper half in C notation */
	if (i != j) 
	  cpdelset((*pp0)[C_OFFSET]+C_OFFSET, j, i, desc, c[p]); /* lower half */
      }
  }
  else {
    /* only use southeastern quarter of matrix */
    for (j = 1; j <= nv; j++) 
      for (p = icn[j+nv]; p <= icn[j+nv+1]-1; p++) {
	i = irn[p]-nv;
	cpdelset((*pp0)[C_OFFSET]+C_OFFSET, i, j, desc, c[p]); /* upper half in C notation */
	if (i != j) 
	  cpdelset((*pp0)[C_OFFSET]+C_OFFSET, j, i, desc, c[p]);
      }
  }
#endif

  return;
}
/*-----*-----------------------------------------------------------------*/

/*
 * evaluate the quadratic form v^t C^{-1} v and the determinant
 * ln |C| using LAPACK or ScaLAPACK routines
 * PARAMS
 *     icn      [I]  column index for packed SLAP storage scheme
 *     irn      [I]  row index
 *     cm       [I]  covariance matrix in sparse storage
 *     nz       [I]  number of non-zero elements in covariance matrix
 *     b        [I]  vector v
 *     nv       [I]  dimension of covariance matrix, length of v
 *     quad     [O]  pointer to quadratic form
 *     lndet    [O]  pointer to ln |C|
 *     flag     [I]  flag: use only southeastern quarter of covariance matrix
 *                   (i.e. imaginary parts only)
 *     rank     [I]  rank of current processor
 *     npe      [I]  number of processors
 */
void claquadform(int *icn, int *irn, double *cm, int nz, double *b,
		 int nv, double *quad, double *lndet,
		 int flag, int rank, int npe)
{
  int i;

  *lndet = 0.;
  *quad = 0.;

  {
    double **a;

#ifndef SCALAPACK
    /*
     * LAPACK version
     */
    int uplo, trans, diag, nrhs, info;
    double *x;

    /* convert covariance matrix to dense storage */
    if (!flag)
      p1_p0(icn, irn, cm, nv, 2, &a); 
    else
      p1_p0_2(icn, irn, cm, nv, 2, &a); 

    x = malloc_1dmatrix(nv);
    for(i = 1; i <= nv; i++)
      x[i] = b[i];   /* copy b into x since it is overwritten by DTRTRS */

    /* 
     * compute Cholesky factorization of a symmetric positive definite matrix 
     */
    uplo = 1;
    CDPOTRF(&uplo, &nv, a[C_OFFSET]+C_OFFSET, &nv, &info);
    if (info != 0)
      printf("ERROR: LAPACK dpotrf info == %i\n", info); 
    for(i = 1; i <= nv; i++) 
      *lndet += 2. * log(a[i][i]);

    /* solve triangular system of linear equation AX=B */
    uplo = 1; trans = 0; diag = 0;
    nrhs = 1;
    CDTRTRS(&uplo, &trans, &diag, &nv, &nrhs, a[C_OFFSET]+C_OFFSET, &nv, 
	    x+C_OFFSET, &nv, &info);
    if (info != 0)
      printf("ERROR: LAPACK dtrtrs info == %i\n", info); 
    for(i = 1; i <= nv; i++)
      *quad += x[i]*x[i];
    free_1dmatrix(x);
    free_2dmatrix(a);

#else
    /* 
     * ScaLAPACK version
     */
    static char c_uplo  = 'L'; 
    static char c_scope = 'A';
    static char c_top   = ' ';
    static char c_diag  = 'N';
    static char c_trans = 'N';
    static int  i_zero = 0;
    static int  i_one  = 1;
    int    info;
    double *x, val;
    int    blocksize = BLOCKSIZE;
    int    icontxt;
    int    lrow, lcol, prow, pcol, nprow, npcol;
    int    desca[9], descx[9];

    cblacs_get(-1, 0, &icontxt);
    cblacs_gridinfo(icontxt, &nprow, &npcol, &prow, &pcol); 
    if (prow >= 0) {
      lrow = cnumroc(nv, blocksize, prow, 0, nprow);
      lcol = cnumroc(nv, blocksize, pcol, 0, npcol);
      /* make sure that the numbers are > 0 to keep ScaLAPACK happy */
      lrow = MAX(1, lrow);
      lcol = MAX(1, lcol);
      descinit(desca, &nv, &nv, &blocksize, &blocksize,
	       &i_zero, &i_zero, &icontxt, &lrow, &info); 

      /* convert covariance matrix to distributed dense storage */
      sca_p1_p0(icn, irn, cm, nv, 2, desca, &a, lrow, lcol, flag); 
      
      
      /* compute Cholesky factorization 
       * of a symmetric positive definite matrix 
       */
      cpdpotrf(&c_uplo, nv, a[C_OFFSET]+C_OFFSET, 1, 1, desca, 
	      &info);
      if ((info != 0) && (rank == ROOT))
	printf("ERROR: pdpotrf info == %i\n", info); 
      for(i = 1; i <= nv; i++) { 
	cpdelget(&c_scope, &c_top, &val, a[C_OFFSET]+C_OFFSET, i, i, desca);
	*lndet += 2. * log(val);
      }
      
      descinit(descx, &nv, &i_one, &blocksize, &blocksize, &i_zero, &i_zero, 
	       &icontxt, &lrow, &info);
      x = malloc_1dmatrix(lrow);
      for(i = 1; i <= nv; i++)
	cpdelset(x+C_OFFSET, i, 1, descx, b[i]);

      /* solve triangular system of linear equation AX=B */
      cpdtrtrs(&c_uplo, &c_trans, &c_diag, nv, 1, 
	       a[C_OFFSET]+C_OFFSET, 1, 1, desca,
	       x+C_OFFSET, 1, 1, descx, &info);
      if (info != 0)
	printf("ERROR: pdtrtrs info == %i\n", info); 
      for(i = 1; i <= nv; i++) {
	cpdelget(&c_scope, &c_top, &val, x+C_OFFSET, i, 1, descx);
	*quad += val * val;
      }

      free_1dmatrix(x);
      free_2dmatrix(a);
    }
    /* broadcast results to processors outside of process grid */
#ifdef MPI
    if (prow * pcol < npe) {
      MPI_Bcast(quad, 1, MPI_DOUBLE, ROOT, MPI_COMM_WORLD);
      MPI_Bcast(lndet, 1, MPI_DOUBLE, ROOT, MPI_COMM_WORLD);
    }
#else
    myerror("ERROR: Need MPI to use all processes.");
#endif
#endif
  }

  return;
}
  
/*-----*-----------------------------------------------------------------*/
/*
 * evaluate the quadratic form v^t C^{-1} v and the determinant
 * ln |C| using sparse SLATEC routines 
 * (wrapper for spquadform.f)
 * PARAMS
 *     icn      [I]  column index for packed SLAP storage scheme
 *     irn      [I]  row index
 *     cm       [I]  covariance matrix in sparse storage
 *     nz       [I]  number of non-zero elements in covariance matrix
 *     vis      [I]  visibility vector v
 *     nv       [I]  dimension of covariance matrix, length of v
 *     quad     [O]  pointer to quadratic form
 *     lndet    [O]  pointer to ln |C|
 *     quaderr  [O]  numerical error returned by SLATEC 
 *     imagflag [I]  flag: calculate real and imaginary parts separately
 *                   (assumes no correlations between real and imag. parts!!)
 */
void cspquadform(int *icn, int *irn, double *cm, int nz, double *vis, int nv,
		 double *quad, double *lndet, double *quaderr, int imagflag)
{
  int    *iwork;
  double *rwork, *x;
  int    lenw, leniw;

  /* 
   * use full covariance matrix 
   */
  if (imagflag) {
    lenw = nz + 5 * nv; 
    leniw = nz + nv + 11;
    iwork = malloc_1imatrix(leniw);
    rwork = malloc_1dmatrix(lenw);      
    x = malloc_1dmatrix(nv);
    SPQUADFORM(&nv, &nz, irn+C_OFFSET, icn+C_OFFSET, cm+C_OFFSET, vis+C_OFFSET,
	       quad, lndet, quaderr, 
	       iwork+C_OFFSET, rwork+C_OFFSET, x+C_OFFSET);
  }
  /* 
   * do real and imaginary parts separately 
   */
  else {
    int    *icn2, *irn2;
    double *cm2;
    double rquad, iquad, rlndet, ilndet, rquaderr, iquaderr; 
    int    j, p, nz2, nv2;
 
    nz2 = nz/2;
    nv2 = nv/2;
    lenw = nz2 + 5 * nv2; 
    leniw = nz2 + nv2 + 11;
    iwork = malloc_1imatrix(leniw);
    rwork = malloc_1dmatrix(lenw);      
    x = malloc_1dmatrix(nv);

    /* 
     * real parts 
     */
    SPQUADFORM(&nv2, &nz2, irn+C_OFFSET, icn+C_OFFSET, cm+C_OFFSET, 
	       vis+C_OFFSET, &rquad, &rlndet, &rquaderr, 
	       iwork+C_OFFSET, rwork+C_OFFSET, x+C_OFFSET);

    /* 
     * imaginary parts 
     */
    icn2 = malloc_1imatrix(nv2+1);
    irn2 = malloc_1imatrix(nz2);
    cm2 = malloc_1dmatrix(nz2);
    
    /* sanity check on index table */
    if ((icn[nv2+1] != nz2+1) || (icn[2*nv2+1] != 2*nz2+1)
	|| (irn[icn[nv2+1]] < nv2+1) || (irn[icn[2*nv2]-1] > 2*nv2))
      myerror("  error in index table in spquadform()");
    /* move imaginary parts to beginning of new sparse matrix */
    for (j = nv2+1; j <= 2*nv2; j++) {
      icn2[j-nv2] = icn[j] - nz2;
      for (p = icn[j]; p <= icn[j+1]-1; p++) {
	irn2[p-nz2] = irn[p]-nv2;
	cm2[p-nz2] = cm[p];
      }
    }
    icn2[nv2+1] = nz2+1;
    SPQUADFORM(&nv2, &nz2, irn2+C_OFFSET, icn2+C_OFFSET, cm2+C_OFFSET, 
	       vis+C_OFFSET+nv2, &iquad, &ilndet, &iquaderr, 
	       iwork+C_OFFSET, rwork+C_OFFSET, x+C_OFFSET);
    free_1imatrix(icn2);
    free_1imatrix(irn2);
    free_1dmatrix(cm2);
    *quaderr = rquaderr + iquaderr;
    *lndet = rlndet + ilndet;
    *quad = rquad + iquad;
  }
  free_1imatrix(iwork);
  free_1dmatrix(rwork);
  free_1dmatrix(x);
  
  return;
}
/*-----*-----------------------------------------------------------------*/
