/* 
 * A program to calculate the likelihood of a flat band power CMB spectrum
 * from interferometric observations of CMB and Galactic fluctuations
 * at several frequencies using signal to noise eigenmodes;
 * includes mosaicing and w-distortions. 
 */
/*  Time-stamp: <12 Nov 2002 18:22 GMT eigenlike.c maising> */
/*  
 *  HISTORY
 *    ??-??-1997  MPH   single-field Fortran version (vsaeigen.f)
 *    10-07-2001  KSM   C version
 *    25-11-2001  KSM   includes MPI, w-distortions
 */

/*-----*-----------------------------------------------------------------*/
/* 	$Id: eigenlike.c,v 1.10 2002/11/13 12:13:45 maising Exp $	 */

#ifndef lint
static char vcid[] = "$Id: eigenlike.c,v 1.10 2002/11/13 12:13:45 maising Exp $";
#endif /* lint */

#include <stdio.h>
#include <math.h>

#if defined(MPI)
#include <mpi.h>
#endif
#if defined (SCALAPACK)
#include "cscalapack.h"
#include "like_scalapack.h"
#else
#include "like_lapack.h"
#endif

#include "my_malloc2.h"
#include "globals.h"
#include "like_temp.h"
#include "like_io.h"
#include "like_sort.h"
#include "like_integrals.h"
#include "like_covariance.h"
#include "like_matvec.h"
#include "like_beam.h"

#define PI 3.14159265358979323844
#define MAX(x, y) (((x) > (y) ? (x) : (y)))
#define MIN(x, y) (((x) < (y) ? (x) : (y)))

#define ROOT 0			/* MPI root process */
#define DISCARD_NEG_VALS 0
#define DISCARD_THRESHOLD 1.e-3
#define SIMPLE_DRIVER 1		/* use LAPACK simple or expert driver */

/*-----*-----------------------------------------------------------------*/

/* solve the general eigenvalue problem S x = \lambda N x,
 * using the sparse matrices S and N
 * PARAMS
 *     sm      [I,O] the signal covariance matrix S; destroyed on exit
 *     nm      [I,O] the noise covariance matrix N; destroyed on exit
 *     nc        [I] dimension of S and N
 *     snval   [I,O] the SN eigenvalues; 
 *                   the memory has to be allocated outside this routine
 *     eigenvecs [O] pointer to the SN eigenvectors
 *     cholesky_diag [O] pointer to the diagonal of the Cholesky 
 *                   decomposition L of N = L L^T
 *     nev       [O] pointer to the number of eigenvalues computed; 
 *                   nev <= 2*nv
 */
void do_gen_eval_problem(double **sm, double **nm, int nc, 
			 double *snval, double ***eigenvecs, 
			 double **cholesky_diag, int *nev)
{
  int i;

#ifndef SCALAPACK
#if SIMPLE_DRIVER
  gen_sym_eigenval(sm, nm, snval, nc);
  *nev = nc;
  *eigenvecs = sm;
#else
  {				/* use LAPACK expert driver */
    double **eigvec;
    int j;
    
    eigvec = malloc_2dmatrix(nc, nc);
    gen_sym_exp_eigenval(sm, nm, nc, DISCARD_THRESHOLD, 1.e3, 
			 snval, eigvec, nev);
    free_2dmatrix(sm);
    *eigenvecs = eigvec;
    
    for (i = *nev+1; i <= nc; i++) {
      snval[i] = 0.;
      for (j = 1; j <= nc; j++) 
	*eigenvecs[i][j] = 0.;
    } 
  }
#endif
#endif /* ifndef SCALAPACK */

  /* nm now contains the Cholesky decomposition L of N */
  *cholesky_diag = malloc_1dmatrix(nc);
  for (i = 1; i <= nc; i++)
    (*cholesky_diag)[i] = nm[i][i];
  free_2dmatrix(nm);
  
  return;
} 
			 

/*************************************************************************
 * obtain SN eigenmodes and transform visibilities to SN eigenbasis.
 * PARAMS 
 *     vis      [I]   vector of visibilities; for complex visibilities,
 *                    the first nc/2 elements are the real parts
 *     nc       [I]   dimension of covariance matrix
 *     snvec    [O]   visibilities rotated to SN basis
 *     snval    [O]   SN eigenvalues
 *     nev      [O]   number of eigenvalues and vectors
 *     lndetn   [O]   pointer to ln det N
 *     sm     [I,O]   signal covariance; on exit, destroyed
 *     nm     [I,O]   noise covariance; on exit, destroyed
 */
void sn_trafo(double *vis, int nc,
	      double *snvec, double *snval, int *nev, double *lndetn,
	      double  **sm, double **nm)
{
  int    i, j;
  double **eigenvecs, *cholesky_diag;
	    
  /* solve the generalised eigenvalue problem Sx = \lambda Nx; 
   * return eigenvalues, vectors and diagonal L_{ii} of 
   * Choleksy decomposition L of N = L L^T 
   */
  do_gen_eval_problem(sm, nm, nc, snval, &eigenvecs, &cholesky_diag, nev);

  /* 
   * log |N| = 2 \sum_i log L_{ii} is a normalisation constant,
   * where L Cholesky decomposition of N=L L^T.
   */
  *lndetn = 0.;
  for (i = 1; i <= nc; i++) 
    *lndetn += 2. * log(cholesky_diag[i]); 

  /* 
   * perform S/N eigenmode rotation.
   */
  for (i = 1; i <= nc; i++) {
    snvec[i] = 0.;
    for (j = 1; j <= nc; j++) 
      snvec[i] += eigenvecs[i][j] * vis[j]; 
  } 
    
  free_2dmatrix(eigenvecs); free_1dmatrix(cholesky_diag);

  return;
}

#ifdef SCALAPACK
/*
 * sn_trafo() using ScaLAPACK.
 * PARAMS 
 *     see sn_trafo(); additional parameters are:
 *     descs, descn [I]  descriptor arrays for the block-cyclic matrices
 *                       S and N (compare ScaLAPACK documentation)
 */ 
void sca_sn_trafo(double *vis, int nc,
		  double *snvec, double *snval, int *nev, double *lndetn,
		  double  **sm, int descs[], double **nm, int descn[])
{

  static char   c_trans = 'T';	    /* 'N' no transpose, 'T' use transpose */
  static char   c_scope = 'A';
  static char   c_top   = ' ';
  static int    i_zero = 0;
  static int    i_one  = 1;
  double *cholesky_diag;
  double **z, *y, *x;
  double valmin, valmax;
  int    i;
  int    descx[9], descy[9], descz[9];
  int    nprow, npcol, prow, pcol, lrow, lcol;
  int    rblocksize, lblocksize, info, icontxt;

  /* 
   * determine process context, blocksizes, process grid and local array size
   */
  icontxt = descs[1];
  rblocksize = descs[4];
  lblocksize = descs[5];
  cblacs_gridinfo(icontxt, &nprow, &npcol, &prow, &pcol); 
  lrow = cnumroc(nc, rblocksize, prow, 0, nprow);
  lcol = cnumroc(nc, lblocksize, pcol, 0, npcol);
  /* make sure that the numbers are > 0 to keep ScaLAPACK happy */
  lrow = MAX(1, lrow);
  lcol = MAX(1, lcol);
  z = malloc_2dmatrix(lcol, lrow);
  descinit(descz, &nc, &nc, &rblocksize, &lblocksize, &i_zero, &i_zero, 
	   &icontxt, &lrow, &info);

  /*
   * solve the distributed generalised eigenvalue problem Sx = \lambda Nx; 
   * return eigenvalues, vectors and diagonal L_{ii} of 
   * Choleksy decomposition L of N = L L^T 
   */
  valmin = -1.; 
  valmax = 1.e5;
  sca_gen_sym_exp_eigenval(sm, descs, nm, descn, valmin, valmax, 
			   snval, nc, z, descz, nev);
  free_2dmatrix(sm);    /* snval is global output */

  /* 
   * copy distributed diagonal of Cholesky decomposition 
   * to non-distributed vector 
   */
  cholesky_diag = malloc_1dmatrix(nc);
  for (i = 1; i <= nc; i++)
    cpdelget(&c_scope, &c_top, cholesky_diag+i, 
	     nm[C_OFFSET]+C_OFFSET, i, i, descz);
  free_2dmatrix(nm);

  /*
   * perform S/N eigenmode rotation using PBLAS level 2 routine pdgemv
   * for matrix-vector multiplication: y = Z x, where Z is the matrix
   * formed by the eigenvectors and x is the visibility vector
   */
  x = malloc_1dmatrix(lrow);
  y = malloc_1dmatrix(lrow);
  descinit(descx, &nc, &i_one, &rblocksize, &lblocksize, &i_zero, &i_zero, 
	   &icontxt, &lrow, &info);
  descinit(descy, &nc, &i_one, &rblocksize, &lblocksize, &i_zero, &i_zero,
	   &icontxt, &lrow, &info);
  for (i = 1; i <= nc; i++) {
    cpdelset(x+C_OFFSET, i, 1, descx, vis[i]);
    cpdelset(y+C_OFFSET, i, 1, descy, 0.);
  }
  cpdgemv(&c_trans, nc, nc, 1., 
	 z[C_OFFSET]+C_OFFSET, 1, 1, descz,
	 x+C_OFFSET, 1, 1, descx, 1, 0.,
	 y+C_OFFSET, 1, 1, descy, 1);
  free_2dmatrix(z);
  for (i = 1; i <= nc; i++) 
    cpdelget(&c_scope, &c_top, snvec+i, y+C_OFFSET, i, 1, descy);
  free_1dmatrix(x);
  free_1dmatrix(y);

  /* 
   * log |N| = 2 \sum_i log L_{ii} is a normalisation constant,
   * where L Cholesky decomposition of N=L L^T.
   */
  *lndetn = 0.;
  for (i = 1; i <= nc; i++) 
    *lndetn += 2. * log(cholesky_diag[i]); 

  free_1dmatrix(cholesky_diag);

  return;
}
#endif

/*************************************************************************
 * compute the logarithm of the likelihood using the SN eigenmodes 
 * PARAMS
 *     theta  [I] dimensionless parameter: power spectrum in investigated
 *                bin divided by power used to obtain SN trafo
 *     snvec  [I] the visibility vector transformed to the S/N eigenbasis
 *     snval  [I] the SN eigenvalues
 *     n      [I] length of the visibility vector = 2 x number of visibilities
 *     lndetn [I] ln det N
 *     chisq  [O] pointer to value of \chi^2
 *     lndet  [O] pointer to value ln det C
 * RETURNS
 *     log likelihood
 */
double snloglike(double theta, 
		 double *snvec, double *snval, int n, double lndetn,
		 double *chisq, double *lndet)
{
  int i;

  *chisq = 0.;
  *lndet = 0.;
  for (i = 1; i <= n; i++) {
#if DISCARD_NEG_VALS
    if (snval[i] >= DISCARD_THRESHOLD)
#endif
    {
      /*
       * S/N eigenmodes: x_i S/N eigenvector, \lambda_i eigenvalue
       *    S x_i = \lambda_i N x_i, normalisation: X^T N X = 1.
       * the covariance matrix in S/N basis is given by: 
       *    C' = N^{-1/2} C N^{-1/2} = \diag[1 + \theta \lambda_i].
       * from which we obtain:
       *    \chi^2 = \sum_i y_i^2 / (1 + \theta \lambda_i),
       *    | C' | = \prod_i (1 +  \theta \lambda_i)
       */
      *chisq += snvec[i]*snvec[i] / (1. + theta * snval[i]);
      *lndet += log(1. + theta * snval[i]);
    }
  }


  /* add normalisation constant to take into account the  normalisation 
   * of the determinant:
   * |C'| = |C| |N^{-1}| ==> |C| = |C'| |N|.
   */
  *lndet += lndetn;

  return -0.5*(*lndet + *chisq);
}

/*
 * count the number of negative values in an array
 * PARAMS
 *     eval    [I]  the array
 *     n       [I]  array dimension
 *     min     [O]  pointer to the value of the minimum array element
 * RETURNS
 *     number of negative elements in array
 */
int count_negvals(double *eval, int n, double *min)
{
  int i, num;

  num = 0;
  *min = 0.;
  for (i = 1; i <= n; i++) {
    if (eval[i] < 0.) {
      num ++;
      if (eval[i] < *min) *min = eval[i];
    }
  }

  return num;
}

/*************************************************************************
 * converts covariance matrix from sparse storage scheme to 
 * ScaLAPACK's distributed block-cyclic storage 
 * (p1_p0() for ScaLAPACK)
 * PARAMS:
 *     icn  [I]  column index array for compressed column storage
 *     irn  [I]  row index array
 *     c    [I]  array elements
 *     nv   [I]  dimension of matrix
 *     lu   [I]  lu == 2: symmetrise matrix; lu == 0, 1: general matrix 
 *     desc [I]  ScaLAPACK matrix descriptor for distributed matrix
 *     pp0  [O]  pointer to output matrix
 *     lrow [I]  number of rows of local matrix
 *     lcol [I]  number of columns of local matrix
 *     flag [I]  flag: convert full matrix (0) or only southeastern quarter (1)
 *               in the second case, nv is half of the total matrix dimension
 */
void sca_p1_p0(int *icn, int *irn, double *c, int nv, int lu, int desc[], 
	       double ***pp0, int lrow, int lcol, int flag)
{
#ifdef SCALAPACK
  int i, j, p;

  (*pp0) = malloc_2dmatrix(lcol, lrow);

  for (i = 1; i <= lcol; i++) 
    for (j = 1; j <= lrow; j++) 
      (*pp0)[i][j] = 0.;

  if (! flag) {
    for (j = 1; j <= nv; j++) 
      for (p = icn[j]; p <= icn[j+1]-1; p++) {
	i = irn[p];
	cpdelset((*pp0)[C_OFFSET]+C_OFFSET, i, j, desc, c[p]); /* upper half in C notation */
	if (i != j) 
	  cpdelset((*pp0)[C_OFFSET]+C_OFFSET, j, i, desc, c[p]); /* lower half */
      }
  }
  else {
    /* only use southeastern quarter of matrix */
    for (j = 1; j <= nv; j++) 
      for (p = icn[j+nv]; p <= icn[j+nv+1]-1; p++) {
	i = irn[p]-nv;
	cpdelset((*pp0)[C_OFFSET]+C_OFFSET, i, j, desc, c[p]); /* upper half in C notation */
	if (i != j) 
	  cpdelset((*pp0)[C_OFFSET]+C_OFFSET, j, i, desc, c[p]);
      }
  }
#endif

  return;
}


/*-----*-----------------------------------------------------------------*/

int main(int argc, char *argv[])
{
  double *u, *v, *w, *visr, *visi, *rms;
  double *u2, *v2, *w2, *vis2, *rms2;
  int    nv, nv2;               /* number of visibilities */
  int    *freqn, *freqn2, nva[NAMAX+1], fieldn[NFMAX+1];
  double rbins[NAMAX+1+1];
  double cps1d[NAMAX+1], gps1d[NAMAX+1], sp;
  double cps1dmax[NAMAX+1], gps1dmax[NAMAX+1], spindmax[NAMAX+1];
  double cps1dold[NAMAX+1], gps1dold[NAMAX+1];

  double chisqmax[NAMAX+1], likemax[NAMAX+1];
  double freq[NFMAX+1], sigma[NFMAX+1], antwdth[NFMAX+1], freq0;
  double ra[NFMAX+1], dec[NFMAX+1];
  double ra_ref, dec_ref;
  double *likeval, *chisqval, *lndetval;
  double *snval, *snvec, lndetn;
  double **sm, **nm;

  int    galflag, plotflag, outflag, sparseflag, imagflag;
  int    gaussian_beam, converge, saveflag, read_ps_flag;
  int    itsp, itg, itc, ita, iter, ind, idx;
  int    ncps, ngps, nsp, nf, ni, na, i, j, ia, firstbin;
  double antmax, uvmin, uvmax, rmin, rmax;
  double cpsmin, cpsmax, cpsstep;
  double gpsmin, gpsmax, gpsstep;
  double spmin, spmax, spstep;
  double min = 0.;
	  
  char   datafn[NFMAX+1][FILENAME_LEN];
  char   outputfn[FILENAME_LEN], logfn[FILENAME_LEN], parlogfn[FILENAME_LEN];
  char   infile[FILENAME_LEN];
  int    std_input = 1;

  int    nev;                   /* eigenvectors */
  integral_struct integral_data;
  cm_struct sm_data, nm_data;
  int    rank = 0, npe = 1;
#ifdef SCALAPACK
  int    blocksize = BLOCKSIZE;
  int    icontxt;
  int    i_zero = 0;
  char   major_order='C';
  int    prow, pcol, nprow, npcol;
  int    mypnum, nprocs;
#endif

  /*
   * read command line arguments
   */
  if (argc > 1) {
    sprintf(infile, "%s", argv[1]);
    std_input = 0;
  }

#ifdef MPI
  /*
   * Initialise MPI.
   */
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &npe);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Process %i is running...\n", rank);
#endif
#ifdef SCALAPACK
  /*
   * Initialise ScaLAPACK.
   */
  cblacs_pinfo(&mypnum, &nprocs);
  if ((nprocs != npe) || (mypnum != rank))
    printf("  Rank and processor number do not match on process %i!\n", rank);

  nprow = (int)sqrt(nprocs) + 1;
  while (nprocs%nprow > 0) nprow--;
  npcol = nprocs / nprow;
  if (rank == ROOT)
    printf("  Using %i x %i processor grid.\n", nprow, npcol);
  cblacs_get(-1, 0, &icontxt);
  cblacs_gridinit(&icontxt, &major_order, nprow, npcol);
  cblacs_gridinfo(icontxt, &nprow, &npcol, &prow, &pcol); 
#endif

  /* initialise checks for memory leaks */
  init_mymalloc();

  /* read input parameters*/
  if (std_input)
    read_inps(&nf, freq, &freq0, &ni, fieldn, 
	      ra, dec, &ra_ref, &dec_ref, sigma, antwdth, &antmax,
	      datafn, rank, npe);
  else
    fread_inps(infile, &nf, freq, &freq0, &ni, fieldn,
	       ra, dec, &ra_ref, &dec_ref, sigma, antwdth, &antmax,
	       datafn, rank, npe);

  /* find limits on uv positions*/
  uvlimits(nf, datafn, &nv, &uvmin, &uvmax, rank, npe);
  if (rank == ROOT)
    printf("  Minimum and maximum u: %f %f\n", uvmin, uvmax);

  /* determine annuli to use*/
  if (std_input)
    annuli(uvmin, uvmax, antmax, rbins, &na, rank, npe);
  else
    fannuli(infile, uvmin, uvmax, antmax, rbins, &na, rank, npe);

  /* 
   * read in input data (assuming units are lambda and Jy), plot uv tracks
   */
  if (rank == ROOT)
    printf("  Number of visibilities: %i.\n", nv);
  freqn = malloc_1imatrix(nv);
  u = malloc_1dmatrix(nv); v = malloc_1dmatrix(nv); w = malloc_1dmatrix(nv);
  visr = malloc_1dmatrix(nv); visi = malloc_1dmatrix(nv); 
  rms = malloc_1dmatrix(nv);

  /* read visibility data from file */
  read_data(nf, freqn, u, v, w, visr, visi, rms, &nv, freq, 
	    rbins[1], rbins[na+1], datafn, rank, npe);

  /* sort visibilities by radius in uv-plane */
  sort_data(freqn, u, v, w, visr, visi, rms, nv, radius);
  /* count visibilities per bin */
  for (ia = 1; ia <= na; ia++)
    nva[ia] = 0;
  for (j = 1; j <= nv; j++)
    for (ia = 1; ia <= na; ia++)
      if (radius(u[j], v[j], w[j]) >= rbins[ia] && 
	  radius(u[j], v[j], w[j]) < rbins[ia+1])
	nva[ia] ++;

  if (rank == ROOT)
    for (ia = 1; ia <= na; ia++)
      printf("  In bin %2i: %5i visibilities.\n", ia, nva[ia]);

  /* read in limits on CMB and GAL parameters*/
  if (std_input) 
    read_lim(&cpsmin, &cpsmax, &cpsstep, &ncps, &gpsmin, &gpsmax, &gpsstep, 
	     &ngps, &spmin, &spmax, &spstep, &nsp, &galflag, rank, npe);
  else 
    fread_lim(infile,
	      &cpsmin, &cpsmax, &cpsstep, &ncps, &gpsmin, &gpsmax, &gpsstep, 
	      &ngps, &spmin, &spmax, &spstep, &nsp, &galflag, rank, npe);


  /* read in parameters controlling likelihood calculation*/
  if (std_input) 
    read_lfcn_params(& plotflag, &outflag, outputfn,
		     &sparseflag, &converge, &saveflag, rank, npe);
  else
    fread_lfcn_params(infile, & plotflag, &outflag, outputfn,
		      &sparseflag, &converge, &saveflag, rank, npe);
  /* write to log file */
  if (rank == ROOT) {
    sprintf(logfn, "%s.log", outputfn);
    log_file(logfn, vcid);
    for (i = 1; i <= nf; i++)
      log_file(logfn, datafn[i]);
  }

  /* read in profile of primary beam - default is Gaussian */
  {
    char beamfilename[FILENAME_LEN];
    double beamuvcell, beamfreq;

#if 0
    read_beam_info(& gaussian_beam, beamfilename, &beamuvcell, &beamfreq);
#else
    gaussian_beam = 1;
    beamuvcell = 3.;
    beamfreq = freq0;
#endif

    if (gaussian_beam) {
      set_gaussian(sigma[1], freq[1]);
      set_beamfft(gaussian_2d_fft);
    }      
    else {
      read_beam(beamfilename, 1., beamuvcell, beamfreq);
      set_gaussian(sigma[1], freq[1]);
      set_beamfft(my_beamfft);
    }
#if 0
    /* write out beam profile (for debugging) */
    if (!gaussian_beam) {
      for (i = 0; i <= 20; i++)
	for (j = 0; j <= 20; j++) {
	  double u, v;
	  dcomplex z1, z2;

	  u = i-10; v = j-10;
	  z1 = gaussian_2d_fft(u, v, freq[1]);
	  z2 = my_beamfft(u, v, freq[1]);
	  printf("%f %f: Gaussian: (%f %f) beam: (%f %f)\n", u, v,
		 z1.re, z1.im, z2.re, z2.im);
	}
    }
#endif
  }

  /* check whether there is more then one pointing (mosaicing) */
  imagflag = 0;
  for (i = 2; i <= nf; i++)
    if ((ra[i] != ra[1]) || (dec[i] != dec[1]))
      imagflag = 1;
#if defined(WCORRECTIONS)
  imagflag = 1;
#endif
  if (rank == ROOT) {
    if (imagflag)
      printf("  Using mosaicing algorithm...\n");
    else 
      printf("  No mosaicing required...\n");
  }

  /* 
   * set initial assumed CMB and GAL power spectrum
   */
  if (rank == ROOT)
    sprintf(parlogfn, "%s.plog", outputfn);

  if (std_input)
    read_initial_ps(cps1dmax, gps1dmax, &sp, na, galflag, &read_ps_flag,
		    rank, npe);
  else
    fread_initial_ps(infile, cps1dmax, gps1dmax, &sp, na, galflag, 
		     &read_ps_flag, rank, npe);

  /*
   * initialise likelihood values
   */
  firstbin = 1;
  iter = 0;
  likeval = malloc_1dmatrix(na*nsp*ngps*ncps);
  chisqval = malloc_1dmatrix(na*nsp*ngps*ncps);
  lndetval = malloc_1dmatrix(na*nsp*ngps*ncps);
  for (i = 1; i <= na*nsp*ngps*ncps; i++)
    likeval[i] = 0.;

  /* 
   * read parameters from previous iteration 
   */
  if (read_ps_flag == 2) {
    if (rank == ROOT)
      printf("  Reading parameters from logfile %s.\n", parlogfn);
    read_param_log(parlogfn, cps1dmax, gps1dmax, 
		   &sp, rbins, na, galflag, &iter, &firstbin, rank, npe);
    firstbin++;		       /* first bin after the last saved bin */
    iter--;		       /* this gets incremented again later */
    if (firstbin > na) {
      firstbin = 1; iter++; 
    }
    read_like(outputfn, rbins, na, cpsmin, cpsmax, ncps, 
	      gpsmin, gpsmax, ngps, spmin, spmax, nsp, likeval, rank, npe);
  }

  if (rank == ROOT) {
    printf("  Initial CMB PS: "); 
    for (ita = 1; ita <= na; ita++)
      printf("%g ", cps1dmax[ita]);
    printf("\n");
  }

  if (rank == ROOT) {
    printf("  Initial GAL PS: "); 
    for (ita = 1; ita <= na; ita++)
      printf("%g ", gps1dmax[ita]);
    printf("\n");
  }

  /* 
   * iterate likelihood calculation
   */
  do {
    iter++;

    /* update CMB and GAL power spectrum */
    for(ita = 1; ita <= na; ita++) {
      cps1d[ita] = cps1dmax[ita];
      gps1d[ita] = gps1dmax[ita];
      cps1dold[ita] = cps1dmax[ita];
      gps1dold[ita] = gps1dmax[ita];
    }

    /* start loop over annuli*/
    for (ita = firstbin; ita <= na; ita++) {
      if (rank == ROOT)
	printf("\n  Beginning analysis of annulus %2i\n\n",ita);

      /* set likelihood values in this bin to zero */
      for (i = (ita-1)*nsp*ngps*ncps+1; i <= ita*nsp*ngps*ncps; i++) {
	likeval[i] = 0.;
	chisqval[i] = 0.;
	lndetval[i] = 0.;
      }

      /* find minimum and maximum radius sensitive to power in this bin */
      rmin = rbins[ita] - antmax;
      rmax = rbins[ita+1] + antmax;

      /* loop over independent mosaics */
      for (ind = 1; ind <= ni; ind++) {

	/* calculate reduced number of visibilities for this bin */
	nv2 = 0;
	for (j = 1; j <= nv; j++)
	  if ((fieldn[freqn[j]] == ind)
	      && (radius(u[j], v[j], w[j]) >= rmin)
	      && (radius(u[j], v[j], w[j]) < rmax))
	    nv2 ++;
	if (rank == ROOT)
	  printf("  Number of observed visibilities = %i\n", nv2);

	/* skip bin if empty */
	if (nv2 == 0) {
	  likemax[ita] = 0.; chisqmax[ita] = 0.; 
	  cps1dmax[ita] = 0.; gps1dmax[ita] = 0.; spindmax[ita] = 0.;
	  continue;
	}

	freqn2 = malloc_1imatrix(nv2);
	u2 = malloc_1dmatrix(nv2); v2 = malloc_1dmatrix(nv2); 
	w2 = malloc_1dmatrix(nv2);
	vis2 = malloc_1dmatrix(2*nv2); rms2 = malloc_1dmatrix(nv2);

	/* copy data */
	i = 0;
	for (j = 1; j <= nv; j++)
	  if ((fieldn[freqn[j]] == ind)
	      && (radius(u[j], v[j], w[j]) >= rmin)
	      && (radius(u[j], v[j], w[j]) < rmax)) {
	    i++;
	    freqn2[i] = freqn[j];
	    u2[i] = u[j]; v2[i] = v[j]; w2[i] = w[j];
	    vis2[i] = visr[j]; vis2[i+nv2] = visi[j]; rms2[i] = rms[j];
	  }

	/*
	 * calculate the integrals required to form the covariance matrix
	 */
	if (rank == ROOT)
	  printf("  Obtaining overlap integrals ...\n");
	get_integrals(&integral_data, freqn2, u2, v2, w2, nv2, 
		      ra, dec, ra_ref, dec_ref, freq, antwdth, sigma, nf,
		      rbins, na, ita, ind, ni, 1, imagflag, saveflag, 
		      "integrals", rank, npe);
	if (rank == ROOT) { 
	  char msg[255];
	  
	  sprintf(msg, "Finished %i integrals for %i visibilities.", 
		  integral_data.nz, integral_data.nv);
	  log_file(logfn, msg);
	}
	if (saveflag & INT_SAVE_ONLY)
	  goto label_freeintegrals;

	/* 
	 * calculate the log(likelihood) ...
	 */
	if (rank == ROOT)
	  printf("\n  Calculating log(likelihood) ...\n\n");
	likemax[ita] = -1.e30;
	snval = malloc_1dmatrix(2*nv2);
	snvec = malloc_1dmatrix(2*nv2);
      
	/* update Galactic flux spectral index */
	for(itsp = 1; itsp <= nsp; itsp++) {
	  sp = spmin + spstep * (double)(itsp-1);

	  /* update GAL power-spectrum */
	  for(itg = 1; itg <= ngps; itg++) {
	    gps1d[ita] = gpsmin + gpsstep * (double)(itg-1);

	    /* 
	     * no loop over CMB parameters here - S/N eigenmodes only
	     * need to be calculated once for different parameter values,
	     * as they are linear in the flat band powers 
	     */
	    cps1d[ita] = cpsstep;
    
	    /* 
	     * create index structure for the covariance matrices
	     */
	    {
	      int cnz;

	      cnz = countcms(&integral_data);
	      if (rank == ROOT) {
		printf("  Number of entries in covariance matrix: %i\n", cnz);
		printf("  Creating covariance matrix...\n");
	      }
	      newcms(&sm_data, nv2, cnz); 
	      newcms(&nm_data, nv2, cnz);
	      if (rank == ROOT)
		printf("  Creating index table for covariance matrix...\n");
	      makeidxcms(&sm_data, &integral_data);
	      makeidxcms(&nm_data, &integral_data);
	    }

	    /* 
	     * compute signal and noise covariance matrices 
	     */
	    makesmnm(rms2, nv2, cps1d, gps1d, na, sp, freq, nf, freq0, 
		     ps_conversion(T_CMB, freq0), ita,
		     &integral_data, &sm_data, &nm_data, rank, npe);
	    freeidxcms(&sm_data); freeidxcms(&nm_data);

	    /* 
	     * compute SN eigenmodes
	     */
#if !defined(SCALAPACK)
	    if (imagflag) {		/* mosaicing */
	      /* convert from column indexed storage to dense storage */
	      p1_p0(sm_data.icn, sm_data.irn, sm_data.cm, 2*nv2, 2, &sm);
	      p1_p0(nm_data.icn, nm_data.irn, nm_data.cm, 2*nv2, 2, &nm);
	      freecms(&sm_data); freecms(&nm_data);

	      /* construct SN eigenmodes */
	      sn_trafo(vis2, 2*nv2, snvec, snval, &nev, &lndetn, sm, nm);
	    }
	    else {			/* no mosaicing */
	      int nevr, nevi;
	      double lndetnr, lndetni;

	      /* diagonalise real parts of covariance matrices */
	      if (rank == ROOT)
		printf("  Performing eigenmode rotation for real parts... \n");
	      p1_p0(sm_data.icn, sm_data.irn, sm_data.cm, nv2, 2, &sm);
	      p1_p0(nm_data.icn, nm_data.irn, nm_data.cm, nv2, 2, &nm);
	      sn_trafo(vis2, nv2, snvec, snval, &nevr, &lndetnr, sm, nm);

	      /* diagonalise imaginary parts of covariance matrices */
	      if (rank == ROOT)
		printf("  Performing eigenmode rotation for imag parts...\n");
	      p1_p0_2(sm_data.icn, sm_data.irn, sm_data.cm, nv2, 2, &sm);
	      p1_p0_2(nm_data.icn, nm_data.irn, nm_data.cm, nv2, 2, &nm);
	      freecms(&sm_data); freecms(&nm_data);
	      sn_trafo(vis2+nv2, nv2, snvec+nv2, snval+nv2, &nevi, &lndetni, 
		     sm, nm);

	      nev = nevr + nevi;
	      lndetn = lndetnr + lndetni;
	    }
#else  /* ScaLAPACK */
	    /*
	     * calculate eigenmodes using distributed ScaLAPACK routines
	     */
	    if (rank == 0)
	      printf("  Performing eigenmode rotation...\n");

	    /* if process is on the process grid */
	    if (prow >= 0) {
	      int lcol, lrow, info;
	      int descs[9], descn[9];

	      if (imagflag) {	/* mosaicing */
		int nc;

		nc = 2 *nv2;
		lrow = cnumroc(nc, blocksize, prow, 0, nprow);
		lcol = cnumroc(nc, blocksize, pcol, 0, npcol);
		/* make sure that numbers are > 0 to keep ScaLAPACK happy */
		lrow = MAX(1, lrow);
		lcol = MAX(1, lcol);
		descinit(descs, &nc, &nc, &blocksize, &blocksize,
			 &i_zero, &i_zero, &icontxt, &lrow, &info); 
		descinit(descn, &nc, &nc, &blocksize, &blocksize,
			 &i_zero, &i_zero, &icontxt, &lrow, &info);
		sca_p1_p0(sm_data.icn, sm_data.irn, sm_data.cm, nc, 2, 
			  descs, &sm, lrow, lcol, 0);
		sca_p1_p0(nm_data.icn, nm_data.irn, nm_data.cm, nc, 2, 
			  descn, &nm, lrow, lcol, 0);
		freecms(&sm_data); freecms(&nm_data);
		sca_sn_trafo(vis2, nc, snvec, snval, &nev, &lndetn, 
			     sm, descs, nm, descn);
	      }
	      else {		/* no mosaicing */
		int nevr, nevi;
		double lndetnr, lndetni;

		lrow = MAX(cnumroc(nv2, blocksize, prow, 0, nprow), 1);
		lcol = MAX(cnumroc(nv2, blocksize, pcol, 0, npcol), 1);
		descinit(descs, &nv2, &nv2, &blocksize, &blocksize,
			 &i_zero, &i_zero, &icontxt, &lrow, &info); 
		descinit(descn, &nv2, &nv2, &blocksize, &blocksize,
			 &i_zero, &i_zero, &icontxt, &lrow, &info);
		/* diagonalise real parts of covariance matrices */
		sca_p1_p0(sm_data.icn, sm_data.irn, sm_data.cm, nv2, 2, 
			  descs, &sm, lrow, lcol, 0);
		sca_p1_p0(nm_data.icn, nm_data.irn, nm_data.cm, nv2, 2, 
			  descn, &nm, lrow, lcol, 0);
		sca_sn_trafo(vis2, nv2, snvec, snval, &nevr, &lndetnr, 
			     sm, descs, nm, descn);
		/* imaginary parts */
		sca_p1_p0(sm_data.icn, sm_data.irn, sm_data.cm, nv2, 2, 
			  descs, &sm, lrow, lcol, 1);
		sca_p1_p0(nm_data.icn, nm_data.irn, nm_data.cm, nv2, 2, 
			  descn, &nm, lrow, lcol, 1);
		freecms(&sm_data); freecms(&nm_data);
		sca_sn_trafo(vis2+nv2, nv2, snvec+nv2, snval+nv2, 
			     &nevi, &lndetni, sm, descs, nm, descn);
		lndetn = lndetnr + lndetni;
		nev = nevr + nevi;
	      }
	    }
	    else {
	      freecms(&sm_data); freecms(&nm_data);
	    }

	    /* make sure all processors have the same results, even if they
	       are not on the ScaLAPACK grid */
	    if (nprow * npcol < npe) {
	      MPI_Bcast(&nev, 1, MPI_INT, ROOT, MPI_COMM_WORLD);
	      MPI_Bcast(snval+C_OFFSET, 2*nv2, MPI_DOUBLE, ROOT, 
			MPI_COMM_WORLD);
	      MPI_Bcast(snvec+C_OFFSET, 2*nv2, MPI_DOUBLE, ROOT, 
			MPI_COMM_WORLD);
	      MPI_Bcast(&lndetn, 1, MPI_DOUBLE, ROOT, MPI_COMM_WORLD);
	    }
#endif

	    /* count negative S/N eigenvalues (as a diagnostic) */
	    if (rank == ROOT)
	      printf("\n  Negative S/N eigenvalues: %i, minimum %g\n\n", 
		     count_negvals(snval, nev, &min), min);

	    /* update CMB power-spectrum */
	    for(itc = 1; itc <= ncps; itc++) {
	      double ll, chisq, lndet;

	      cps1d[ita] = cpsmin + cpsstep * (double)(itc-1);

	      idx = (ita-1)*nsp*ngps*ncps +(itsp-1)*ngps*ncps+(itg-1)*ncps+itc;

	      /*
	       * calculate log[likelihood]
	       */
	      ll = snloglike(cps1d[ita]/cpsstep, snvec, snval, nev, lndetn,
			     &chisq, &lndet);
	      likeval[idx] += ll;
	      chisqval[idx] += chisq;
	      lndetval[idx] += lndet;
	      
	      if (ind == ni) {
		/* write out current parameters and likelihood */
		if ((plotflag) && (rank == ROOT)) {
		  printf("  CMB PS, GAL PS, Index = %g %g %g\n",
			 cps1d[ita], gps1d[ita], sp);
		  printf("  lndet,chisq,loglike   = %8.3f %8.3f %8.3f\n",
			 0.5*lndetval[idx], 0.5*chisqval[idx], likeval[idx]);
		}

		/* store maximum likelihood in present annulus */
		if (likeval[idx] > likemax[ita]) {
		  likemax[ita] = likeval[idx];
		  chisqmax[ita] = chisqval[idx];
		  cps1dmax[ita] = cps1d[ita];
		  gps1dmax[ita] = gps1d[ita];
		  spindmax[ita] = sp;
		}
	      }
	    } /*...close CMB power-spectrum loop*/
	  } /*...close GAL power-spectrum loop*/
	} /*...close spectral index loop*/

	free_1dmatrix(snval); free_1dmatrix(snvec);

      label_freeintegrals:
	if (! (saveflag & INT_STORE_MEMORY))
	  freeintegrals(&integral_data);
	free_1imatrix(freqn2); 
	free_1dmatrix(u2); free_1dmatrix(v2); free_1dmatrix(w2);
	free_1dmatrix(vis2); free_1dmatrix(rms2);
      } /* loop over independent mosaics */

      /* reset CMB and GAL power spectrum to use for next annulus*/
      cps1d[ita] = cps1dold[ita];
      gps1d[ita] = gps1dold[ita];

      if (rank == ROOT)
	printf("  Maximum in iteration %2i, bin %2i occurs at %g %g %g\n",
	       iter, ita, cps1dmax[ita], gps1dmax[ita], spindmax[ita]);
#if 1
      /* log results to file - I find this useful on supercomputers 
       *  that don't flush their output regularly
       */
      if (rank == ROOT) {
	char msg[255];

	sprintf(msg, "Iteration %2i bin %2i  max %g %g %g", iter, ita, 
		cps1dmax[ita], gps1dmax[ita], spindmax[ita]);
	log_file(logfn, msg);
      }
#endif

      /*
       * write out likelihood values to file
       */
      write_param_log(parlogfn, cps1dmax, gps1dmax, spindmax[ita],
		      rbins, na, galflag, iter, ita, rank, npe);
      if (outflag)
	write_like(outputfn, rbins, na, cpsmin, cpsmax, ncps,
		   gpsmin, gpsmax, ngps, spmin, spmax, nsp,
		   likeval, rank, npe);

    } /* ... loop over annuli */

    /* 
     * write out peak of likelihood
     */
    if (rank == ROOT) {
      printf("\nMaximum of log(likelihood) occurs at:\n\n");
      for(ita = 1; ita <= na; ita++) {
	double leff;

	leff = PI * (rbins[ita]+rbins[ita+1]);
	printf("In annulus %2i centred at l = %8.2f\n", ita, leff);
	printf("\n");
	printf("  CMB power spectrum = %8.2g\n", cps1dmax[ita]);
	printf("  GAL power spectrum = %8.2g\n", gps1dmax[ita]);
	printf("  GAL spectral index = %8.2g\n", spindmax[ita]);
	printf("  Chi-sq,log[likehd] = %8.2f %8.2f\n",
	       0.5 * chisqmax[ita], 0.5*likemax[ita]);
	printf("\n");
      }
    }

    /* 
     * restart from current maximum likelihood estimates
     */
    firstbin = 1;
    if (converge) {
      /* check for convergence of likelihood maximum.
       * if either the change in one parameter is greater than the step size,
       * or the total change is more than (na)*0.6, start new iteration 
       */
      double eps = 0.6 * na;
      double error, ec, eg;
      int    flag;

      error = 0.;
      flag = 0;
      for (ita = 1; ita <= na; ita ++) {
	ec = 0.; eg = 0.;
	if (cpsstep > 0.)
	  ec = fabs(cps1dmax[ita] - cps1dold[ita]) / cpsstep;
	if (gpsstep > 0.)
	  eg = fabs(gps1dmax[ita] - gps1dold[ita]) / gpsstep;
	if ((ec > 1.001) || (eg > 1.001))
	  flag = ita;
	error += ec + eg;
      }
      if (rank == ROOT)
	printf("  Convergence of maximum: %g %g, %i\n", error, eps, flag);
      if (!((fabs(error) > eps) || (flag)) || (iter > 10))
	converge = 0;
    }
  } while (converge);

  if (rank == ROOT)
    printf("  Likelihood iteration converged after iteration %i.\n", iter);

  /* 
   * free allocated memory
   */
  reset_integrals(na, ni);
  free_1dmatrix(likeval);
  free_1dmatrix(chisqval);
  free_1dmatrix(lndetval);
  free_1imatrix(freqn); free_1dmatrix(u); free_1dmatrix(v); free_1dmatrix(w);
  free_1dmatrix(visr); free_1dmatrix(visi); free_1dmatrix(rms);

  /* check for memory leaks */
  done_mymalloc();

  /*
   * exit ScaLAPACK and MPI
   */
#ifdef SCALAPACK
  if (prow >= 0) {
    cblacs_gridexit(icontxt);
    cblacs_exit(0);
  }
#endif
#if defined(MPI) && (!defined(SCALAPACK))
  MPI_Finalize();
#endif

  return 0;
}
/*-----*-----------------------------------------------------------------*/
