/*
 * routine to calculate Hessian matrix by second-differencing
 * converted from equivalent FORTRAN routine
 */
/*  Time-stamp: <21 Jul 2002 13:11 BST hessian.c maising> */

#include <stdio.h>
#include <math.h>
#include "my_malloc2.h"
#include "hessian.h"

#define MAX(x, y) ((((x) > (y)) ? (x) : (y)))

/* PURPOSE
 * -------
 * FIND SECOND ORDER FORWARD FINITE DifFERENCE APPROXIMATION "A"
 * TO THE SECOND DERIVATIVE (HESSIAN) OF THE DEFINED BY THE SUBP
 * "FCN" EVALUATED AT THE NEW ITERATE "XPLS"
 *
 * FOR OPTIMIZATION USE THIS ROUTINE TO ESTIMATE
 * 1) THE SECOND DERIVATIVE (HESSIAN) OF THE OPTIMIZATION 
 *    if NO ANALYTICAL USER HAS BEEN SUPPLIED FOR EITHER
 *    THE GRADIENT OR THE HESSIAN AND if THE OPTIMIZATION 
 *    "FCN" IS INEXPENSIVE TO EVALUATE.
 *
 * PARAMETERS
 * ----------
 * NR           --> DIMENSION OF MATRIX
 * N            --> DIMENSION OF PROBLEM
 * XPLS[N]      --> NEW ITERATE:   X[K]
 * FCN          --> NAME OF void TO EVALUATE 
 * A[N][N]      <--  FINITE DifFERENCE APPROXIMATION TO HESSIAN
 *                  ONLY LOWER TRIANGULAR MATRIX AND DIAGONAL
 *                  ARE RETURNED
 * SX[N]        --> DIAGONAL SCALING MATRIX FOR X
 * RNOISE       --> RELATIVE NOISE IN FNAME [F[X]]
 */
void hessian(double (*fcn)(double *), double *xpls, int n, double **a,
	     double *sx, double rnoise)
{
  int  i, j, ip1;
  double   ov3,xtmpi, xtmpj, fpls, fhat;
  double *stepsz, *anbr;
  /* STEPSZ[N]    --> WORKSPACE (STEPSIZE IN I-TH COMPONENT DIRECTION)*/
  /* ANBR[N]      --> WORKSPACE (NEIGHBOR IN I-TH DIRECTION)*/

  stepsz = malloc_1dmatrix(n);
  anbr = malloc_1dmatrix(n);

  /* CALCULATE VALUE AT INPUT POINT */
  fpls = (*fcn)(xpls);

  /* FIND I-TH STEPSIZE AND EVALUATE NEIGHBOR IN DIRECTION
   * OF I-TH UNIT VECTOR.
   */
  ov3 = 1.0/3.0;
  for (i = 1; i <= n; i++) {
    stepsz[i] = pow(rnoise, ov3) * MAX((fabs(xpls[i])), sx[i]);
    xtmpi = xpls[i];
    xpls[i] = xtmpi + stepsz[i];
    anbr[i] = (*fcn)(xpls);
    xpls[i] = xtmpi;
  }
  /*
   * CALCULATE COLUMN I OF A
   */
  for (i = 1; i <= n; i++) {
    xtmpi = xpls[i];
    xpls[i] = xtmpi + 2.* stepsz[i];
    fhat = (*fcn)(xpls);
    a[i][i] = ((fpls-anbr[i])+(fhat-anbr[i]))/(stepsz[i]*stepsz[i]);

    /* CALCULATE SUB-DIAGONAL ELEMENTS OF COLUMN */
    if (i != n) {
      xpls[i] = xtmpi + stepsz[i];
      ip1 = i+1;
      for (j = ip1; j <= n; j++) {
	xtmpj=xpls[j];
	xpls[j] = xtmpj + stepsz[j];
	fhat = (*fcn)(xpls);
	a[j][i] = ((fpls-anbr[i])+(fhat-anbr[j]))/(stepsz[i]*stepsz[j]);
	xpls[j] = xtmpj;
      }
    }

    xpls[i] = xtmpi;
  }

  free_1dmatrix(stepsz);
  free_1dmatrix(anbr);
  return;
}
