/* 
 * NR routines for integration in double format 
 *
 * KSM, 2001
 * includes a hacked version of qsimp to do complex integrations 
 */

#include <stdio.h>
#include <stdlib.h>
#include "integration.h"
#include "my_malloc2.h"

/*---qtrap.c---*/
#include <math.h>
#define EPS 1.0e-5
#define JMAX 20

double qtrap(double (*func)(double), double a, double b)
{
	double trapzd(double (*func)(double), double a, double b, int n);
	void myerror(char error_text[]);
	int j;
	double s,olds;

	olds = -1.0e30;
	for (j=1;j<=JMAX;j++) {
		s=trapzd(func,a,b,j);
		if (fabs(s-olds) < EPS*fabs(olds)) return s;
		olds=s;
	}
	myerror("Too many steps in routine qtrap");
	return 0.0;
}
#undef EPS
#undef JMAX
/* (C) Copr. 1986-92 Numerical Recipes Software ?4210(9p#3. */

/*---qsimp.c---*/
/* on entry, iter contains the maximum number of iterations.
 * on exit, it returns the number of iterations used. */
double qsimp2(double (*func)(double), double a, double b, double eps,
	      int *iter)
{
	void trapzd2(double (*func)(double), double a, double b, 
		       double *s, int n);
	void myerror(char error_text[]);
	int j, jmax;
	double s,st,ost,os;

	s = 0.;
	jmax = *iter;
	*iter = 0;
	ost = os = -1.0e30;
	for (j=1;j<=jmax;j++) {
		trapzd2(func,a,b,&st,j);
		s=(4.0*st-ost)/3.0;
		if (fabs(s-os) < eps*fabs(os)) {
		  *iter = j;
		  return s;
		}
		os=s;
		ost=st;
	}
	return s;
}

double cmod(dcomplex z)
{
  double x,y,ans,temp;

  x = fabs(z.re);
  y = fabs(z.im);
  if (x == 0.0)
    ans = y;
  else if (y == 0.0)
    ans = x;
  else if (x > y) {
    temp = y/x;
    ans = x*sqrt(1.0+temp*temp);
  } else {
    temp = x/y;
    ans = y*sqrt(1.0+temp*temp);
  }
  return ans;
}

dcomplex cqsimp2(dcomplex (*func)(double), double a, double b, double eps,
		 int *iter)
{
  void ctrapzd2(dcomplex (*func)(double), double a, double b, 
		dcomplex *s, int n);
  void myerror(char error_text[]);
  int j, jmax;
  dcomplex s,st,ost,os,delta;

  jmax = *iter;
  *iter = 0;
  ost.re = os.re = -1.0e30;
  ost.im = os.im = -1.0e30;
  for (j=1;j<=jmax;j++) {
    ctrapzd2(func,a,b,&st,j);
    s.re=(4.0*st.re-ost.re)/3.0;
    s.im=(4.0*st.im-ost.im)/3.0;
    delta.re = s.re - os.re;
    delta.im = s.im - os.im;
    if (cmod(delta) < eps*cmod(os)) {
      *iter = j;
      return s;
    }
    os.re = s.re;
    os.im = s.im;
    ost.re = st.re;
    ost.im = st.im;
  }
  return s;
}
/* (C) Copr. 1986-92 Numerical Recipes Software ?4210(9p#3. */

/*---qromb---*/
#define EPS 1.0e-6
#define JMAX 20
#define JMAXP (JMAX+1)
#define K 5

double qromb2(double (*func)(double), double a, double b, double eps, 
	      int *iter)
{
	void polint(double xa[], double ya[], int n, double x, double *y, double *dy);
	void trapzd2(double (*func)(double), double a, double b, 
		       double *s, int n);
 	void myerror(char error_text[]);
	double ss,dss,oldss;
	double s[JMAXP+1],h[JMAXP+1];
 	int j,jmax;

	jmax=*iter;
	*iter = 0;
	oldss=-1.e30;
 	h[1]=1.0;
 	for (j=1;j<=jmax;j++) {
 		trapzd2(func,a,b,&(s[j]),j);
 		if (j >= K) {
 			polint(&h[j-K],&s[j-K],K,0.0,&ss,&dss);
 			if (fabs(ss-oldss) < eps*fabs(oldss)) {
			  *iter=j;
			  return ss;
			}
			oldss=ss;
 		}
 		s[j+1]=s[j];
 		h[j+1]=0.25*h[j];
 	}
 	return ss;
}

double qromb(double (*func)(double), double a, double b, double eps, 
	     int *ifail)
{
	void polint(double xa[], double ya[], int n, double x, double *y, double *dy);
	double trapzd(double (*func)(double), double a, double b, int n);
 	void myerror(char error_text[]);
	double ss,dss;
	double s[JMAXP+1],h[JMAXP+1];
 	int j;

	*ifail = 0;
 	h[1]=1.0;
 	for (j=1;j<=JMAX;j++) {
 		s[j]=trapzd(func,a,b,j);
 		if (j >= K) {
 			polint(&h[j-K],&s[j-K],K,0.0,&ss,&dss);
 			if (fabs(dss) < eps*fabs(ss)) return ss;
 		}
 		s[j+1]=s[j];
 		h[j+1]=0.25*h[j];
 	}
	*ifail = 1;
 	return 0.0;
}
#undef EPS
#undef JMAX
#undef JMAXP
#undef K
/* (C) Copr. 1986-92 Numerical Recipes Software ?4210(9p#3. */

/*---polint.c---*/
#define NRANSI

void polint(double xa[], double ya[], int n, double x, double *y, double *dy)
{
	int i,m,ns=1;
	double den,dif,dift,ho,hp,w;
	double *c,*d;

	dif=fabs(x-xa[1]);
	c=malloc_1dmatrix(n);
	d=malloc_1dmatrix(n);
	for (i=1;i<=n;i++) {
		if ( (dift=fabs(x-xa[i])) < dif) {
			ns=i;
			dif=dift;
		}
		c[i]=ya[i];
		d[i]=ya[i];
	}
	*y=ya[ns--];
	for (m=1;m<n;m++) {
		for (i=1;i<=n-m;i++) {
			ho=xa[i]-x;
			hp=xa[i+m]-x;
			w=c[i+1]-d[i];
			if ( (den=ho-hp) == 0.0) myerror("Error in routine polint");
			den=w/den;
			d[i]=hp*den;
			c[i]=ho*den;
		}
		*y += (*dy=(2*ns < (n-m) ? c[ns+1] : d[ns--]));
	}
	free_1dmatrix(d);
	free_1dmatrix(c);
}
#undef NRANSI
/* (C) Copr. 1986-92 Numerical Recipes Software ?4210(9p#3. */

#define FUNC(x) ((*func)(x))

double trapzd(double (*func)(double), double a, double b, int n)
{
	double x,tnm,sum,del;
	static double s;
	int it,j;

	if (n == 1) {
		return (s=0.5*(b-a)*(FUNC(a)+FUNC(b)));
	} else {
		for (it=1,j=1;j<n-1;j++) it <<= 1;
		tnm=it;
		del=(b-a)/tnm;
		x=a+0.5*del;
		for (sum=0.0,j=1;j<=it;j++,x+=del) sum += FUNC(x);
		s=0.5*(s+(b-a)*sum/tnm);
		return s;
	}
}

void trapzd2(double (*func)(double), double a, double b, double *s, int n)
{
  double x,tnm,sum,del;
  int it,j;

  if (n == 1) {
    *s = 0.5*(b-a)*(FUNC(a)+FUNC(b));
  } else {
    for (it=1,j=1;j<n-1;j++) it <<= 1;
    tnm=it;
    del=(b-a)/tnm;
    x=a+0.5*del;
    for (sum=0.0,j=1;j<=it;j++,x+=del) sum += FUNC(x);
    *s=0.5*(*s+(b-a)*sum/tnm);
  }
  return;
}

void ctrapzd2(dcomplex (*func)(double), double a, double b, 
	      dcomplex *s, int n)
{
  double x,tnm,del;
  int it,j;
  dcomplex fa, fb, fx, sum;

  if (n == 1) {
    fa = FUNC(a); fb = FUNC(b);
    (*s).re = 0.5*(b-a)*(fa.re+fb.re);
    (*s).im = 0.5*(b-a)*(fa.im+fb.im);
  } else {
    for (it=1,j=1;j<n-1;j++) it <<= 1;
    tnm=it;
    del=(b-a)/tnm;
    x=a+0.5*del;
    sum.re = 0.;
    sum.im = 0.;
    for (j=1;j<=it;j++) { 
      fx = FUNC(x);
      sum.re += fx.re;
      sum.im += fx.im;
      x += del;
    }
    (*s).re=0.5*((*s).re+(b-a)*sum.re/tnm);
    (*s).im=0.5*((*s).im+(b-a)*sum.im/tnm);
  }
  return;
}
#undef FUNC
/* (C) Copr. 1986-92 Numerical Recipes Software ?4210(9p#3. */






