/*
 * routines to store and manipulate beam patterns.
 * this is a bit messy; also see comments below about input data formats
 * KSM, 2001
 */
/*  Time-stamp: <19 Oct 2002 18:09 BST like_beam.c maising> */

#include <stdio.h>
#include <math.h>
#include "my_malloc2.h"
#include "like_beam.h"

#define PI 3.14159265358979323844
#define TINY 1.e-15

struct gaussian_struct {
  double sigma, freq;
  double a, b;
};

static struct gaussian_struct gaussian_data;

/* 
 * initialise Gaussian beam by setting the beam dispersion at a given
 * frequency
 */
void set_gaussian(double sigma, double freq)
{
  gaussian_data.a = 2.*PI*sigma*sigma*freq*freq;
  gaussian_data.b = -2.*PI*PI*sigma*sigma*freq*freq;
  return;
}

/* 
 * return Fourier tranform of Gaussian beam
 */
dcomplex gaussian_2d_fft(double u, double v, double freq)
{
  double x;
  dcomplex result;

  x = freq * freq;

  result.re = (gaussian_data.a / x) * exp((gaussian_data.b / x) * (u*u+v*v));
  result.im = 0.;
  return result;
}

struct beam_struct {
  double **re;
  double **im;
  double uvcellx, uvcelly, freq;
  int nx, ny;
  double xcentre, ycentre;
  double xmin, xmax, ymin, ymax;
};

static struct beam_struct beam;

void set_beam(double **re, double **im, int nx, int ny, 
	      double uvcellx, double uvcelly, double freq)
{
  beam.re = re;
  beam.im = im;
  beam.nx = nx;
  beam.ny = ny;
  beam.uvcellx = uvcellx;
  beam.uvcelly = uvcelly;
  beam.freq = freq;
  beam.xcentre = nx/2+1;
  beam.ycentre = ny/2+1;
  beam.xmin = -uvcellx*nx/2;
  beam.ymin = -uvcelly*ny/2;
  beam.xmax = uvcellx*(nx/2-1);
  beam.ymax = uvcelly*(ny/2-1);
  return;
}

/*
 * I think the peculiar file format used here is what you get when you
 * dump an array to a file in Fortran??
 */
void write_beam(char filename[])
{
  FILE *fp;
  int i, j, k;
  double reval, imval;

  fp = fopen(filename, "w");
  if (fp != NULL) {
    fprintf(fp, "%i %i\n", beam.nx, beam.ny);
    k = 0;
    for (i = 1; i <= beam.nx; i++)
      for (j = 1; j <= beam.ny; j++) {
	reval = beam.re[i][j];
	imval = beam.im[i][j];
	if (fabs(imval) < TINY) imval = 0.;
	if (fabs(reval) < TINY) reval = 0.;
	fprintf(fp, " ( %g, %g) ", reval, imval);
	k ++;
	if (! (k % 5))
	  fprintf(fp, "\n");
      }
    if (k % 5)
      fprintf(fp, "\n");
    fclose(fp);
  }
  else 
    printf("ERROR: Could not open %s.\n", filename);

  return;
}

void query_beam_params(double ***r, double ***i, int *nx, int *ny, 
		       double *cellx, double *celly, double *freq)
{
  *r = beam.re;
  *i = beam.im;
  *nx = beam.nx;
  *ny = beam.ny;
  *cellx = beam.uvcellx;
  *celly = beam.uvcelly;
  *freq = beam.freq;

  return;
}

/*
 * I think the peculiar file format used here is what you get when you
 * dump an array to a file in Fortran??
 */
void read_beam(char filename[], double scale, double cellsize, double freq)
{
  int i, j, imax, jmax;
  FILE *fp;
  double max;
  
  beam.uvcellx = cellsize;
  beam.uvcelly = cellsize;
  beam.freq = freq;

  fp = fopen(filename, "r");
  if (fp) {
    printf("  Opened %s.\n", filename);
    fscanf(fp, "%i %i\n", &beam.nx, &beam.ny);
    printf("  Grid size %i x %i\n", beam.nx, beam.ny);
    beam.re = malloc_2dmatrix(beam.nx, beam.ny);
    beam.im = malloc_2dmatrix(beam.nx, beam.ny);
    beam.xcentre = beam.nx/2+1;
    beam.ycentre = beam.ny/2+1;
    beam.xmin = -beam.uvcellx*beam.nx/2;
    beam.ymin = -beam.uvcelly*beam.ny/2;
    beam.xmax = beam.uvcellx*(beam.nx/2-1);
    beam.ymax = beam.uvcelly*(beam.ny/2-1);
    max = 0.;
    imax = 1; jmax = 1;
    i = 1; j = 1;
    while (fscanf(fp, " (%lg,%lg) ", &beam.re[i][j], &beam.im[i][j]) != EOF) {
      beam.re[i][j] *= scale;
      beam.im[i][j] *= scale;
      if (beam.re[i][j] > max) {
	max = beam.re[i][j];
	imax = i;
	jmax = j;
      }
      j++;
      if (j > beam.ny) {
	j = 1; i++;
      }
    }
    fclose(fp);
    printf("  Read beam profile from %s.\n", filename);
    printf("  Maximum of %f occurs at pixel (%i, %i)\n", max, imax, jmax);
  }
  else
    printf("ERROR: Could not load %s.\n", filename);

  return;
}

void free_beam()
{
  free_2dmatrix(beam.re);
  free_2dmatrix(beam.im);
  return;
}

/* interpolate an array
 * PARAMS
 *     image          [I] array to be interpolated
 *     i0,j0          [I] reference pixel in the array
 *     x0,y0          [I] coordinates of the reference pixel given by i0, j0
 *     xcell, ycell   [I] cell size of array 
 *     xpoint, ypoint [I] coordinates of point at which image is interpolated
 *     mode           [I] 0 - nearest neighbour, 1 - bilinear interpolation 
 * RETURNS
 *     interpolated image value at point (xpoint, ypoint)
 */
double myinterp(double **image, int i0, int j0, double x0, double y0,
		double xcell, double ycell, double xpoint, double ypoint,
		int mode)
{
  double val, t, u;
  int i, j;

  switch (mode) {
  case 1:			/* bilinear interpolation */
    i = i0 + (int)floor((xpoint-x0) / xcell);
    j = j0 + (int)floor((ypoint-y0) / ycell);
    t = (xpoint - x0) / xcell  - (i-i0);
    u = (ypoint - y0) / ycell  - (j-j0);
    
    val = (1-t) * (1-u) * image[i][j] + t * (1-u) * image[i+1][j]
      + u * (1-t) * image[i][j+1] + t * u * image[i+1][j+1];

    break;
  default:			/* nearest neighbour */
    i = i0 + (int)floor((xpoint-x0) / xcell+0.5);
    j = j0 + (int)floor((ypoint-y0) / ycell+0.5);
    val = image[i][j];
    break;
  }
  return val;
}

dcomplex my_beamfft(double u, double v, double freq)
{
  dcomplex result;
  int icentre, jcentre;
  double u2, v2, scale;

  if ((u >= beam.xmax) || (u <= beam.xmin) || 
      (v >= beam.ymax) || (v < beam.ymin)) {
    result.re = 0.; result.im = 0.;
    return result;
  }
  icentre = beam.nx/2 + 1;
  jcentre = beam.ny/2 + 1;
  scale = beam.freq / freq;
  u2 = u * scale;
  v2 = v * scale;
  result.re = scale*scale * myinterp(beam.re, icentre, jcentre, 0., 0., 
				     beam.uvcellx, beam.uvcelly, u2, v2, 1); 
  result.im = scale*scale * myinterp(beam.im, icentre, jcentre, 0., 0., 
				     beam.uvcellx, beam.uvcelly, u2, v2, 1);
  return result;
}
