/* 
 * routines for calculating the covariance matrix of visibilities
 * for mosaiced VSA observations from flat band powers.
 */
/*  Time-stamp: <12 Nov 2002 18:38 GMT like_covariance.c maising> */
/*  
 *  HISTORY
 *    ??-??-1997  MPH   original single-field Fortran version 
 *                      (vsa_indfuncsubs.f) 
 *    31-05-2001  KSM   C version 
 *    09-06-2001  KSM   mosaicing works
 *    24-11-2001  KSM   changes to reduce memory footprint, enable MPI
 *                      and include w-corrections
 */
/* 	$Id: like_covariance.c,v 1.3 2002/11/13 12:13:45 maising Exp $	 */

#ifndef lint
static char vcid[] = "$Id: like_covariance.c,v 1.3 2002/11/13 12:13:45 maising Exp $";
#endif /* lint */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "globals.h"
#include "my_malloc2.h"
#include "integration.h"
#include "bessi0.h"
#include "like_integrals.h"
#include "like_covariance.h"
#include "like_temp.h"
#if defined(MPI)
#include "mpi.h"
#endif

#define ROOT 0

/*-----*-----------------------------------------------------------------*/

/* 
 * allocate memory for sparse representation of covariance matrix 
 * PARAMS
 *     cms [I,O]  struct for the covariance matrix
 *     nv  [I]    number of visibilities
 *     nz  [I]    number of non-zero elements of covariance matrix
 */
void newcms(cm_struct *cms, int nv, int nz)
{
  cms->nv = nv;
  cms->nz = nz;
  cms->icn = malloc_1imatrix(2*nv+1);
  cms->irn = malloc_1imatrix(nz);
  cms->cm = malloc_1dmatrix(nz);

  return;
}

/* 
 * free memory allocated for covariance matrix 
 * PARAMS
 *     cms [I,O]  struct for the covariance matrix
 */
void freecms(cm_struct *cms)
{
  free_1dmatrix(cms->cm);
  free_1imatrix(cms->icn);
  free_1imatrix(cms->irn);

  return;
}

/* 
 * this subroutine creates the icn and irn arrays for the 
 * covariance matrix of visibility vectors of length 2*N 
 * containing the real parts in the first N elements and the imaginary 
 * parts in the next N.
 *
 * the matrix has the structure:
 * |  \
 * | <xx> \
 * |----------\
 * |  \  <xy> |   \
 * | <yx> \   |  <yy> \
 * |----------------------\    
 *
 * the <xx>, <yx> and <yy> segments have the same structure (N_z elements), 
 * whereas <xy> can be constructed from their transpose. 
 * the N_d diagonal elements of <yx> and <xy> are identical.
 * the new covariance matrix has 4*N_z - N_d elements.
 * lookup tables pxx, pyy, pxy, pyx help to find positions in new matrix
 * as a function of the N_z old elements.
 * 
 * PARAMS
 *     ints [I] struct containing the overlap integrals
 *     cms  [O] struct for the covariance matrix
 */
void makeidxcms(cm_struct *cms, integral_struct *ints)
{
  int niz = ints->nz;
  int nz = cms->nz;
  int nv = ints->nv;
  int i, j, p, p2, nd;
  int *icn, *irn, *ticn, *tirn, *iicn, *iirn;
  int *po, *count;

  ticn = NULL; tirn = NULL; po = NULL; 
  iicn = ints->icn;
  iirn = ints->irn;
  icn = cms->icn;
  irn = cms->irn;

  cms->pxx = malloc_1imatrix(niz);
  cms->pyy = malloc_1imatrix(niz);
  cms->pxy = malloc_1imatrix(niz);
  cms->pyx = malloc_1imatrix(niz);
  cms->imaginary_flag = ints->imaginary_flag;

  /* 
   * create column indices of the transpose in compressed storage.
   * this will be useful for the parts of the covariance matrix that can
   * be derived from symmetry properties 
   */ 
  if (ints->imaginary_flag) {

    ticn = malloc_1imatrix(nv+1);
    tirn = malloc_1imatrix(niz);
    count = malloc_1imatrix(nv);
    po = malloc_1imatrix(niz);	/* lookup table for index in original matrix 
				   as a function of index p in transpose  */

    for (j = 1; j <= nv; j++)
      count[j] = 0;

    /* 
     * count: old number of non-zero elements per row 
     *  == new number of elements per column. 
     * also count diagonal elements.
     */
    nd = 0;
    for(j = 1; j <= nv; j++)
      for(p = iicn[j]; p <= iicn[j+1]-1; p++) {
	i = iirn[p];
	count[i]++;
	if (i == j) nd++;
      }

    /* now use number counts to determine first index in each column */
    ticn[1] = 1;
    for(j = 1; j <= nv; j++)
      ticn[j+1] = ticn[j]+count[j];
    
    for (j = 1; j <= nv; j++)
      count[j] = 0;
    
    /* fill in row indices and create lookup table for original indices */
    for(j = 1; j <= nv; j++)
      for(p = iicn[j]; p <= iicn[j+1]-1; p++) {
	i = iirn[p];
	p2 = ticn[i]+count[i];
	tirn[p2] = j;
	po[p2] = p;
	count[i]++;
      }
    free_1imatrix(count);
  }

  /* 
   * now create new irn and icn arrays by walking through the columns
   * of the new matrix; compare the diagram above 
   */
  p2 = 1;
  for (j = 1; j <= nv; j++) {
    icn[j] = p2;
    for (p = iicn[j]; p <= iicn[j+1]-1; p++) { /* <xx>: real parts only */
      cms->pxx[p] = p2;
      irn[p2] = iirn[p];
      p2++;
    }
    /* include off-diagonal blocks <yx>, <xy> for mosaicing */
    if (ints->imaginary_flag) {
      for (p = ticn[j]; p <= ticn[j+1]-1; p++) { /* <xy> */
	i = tirn[p];
	cms->pxy[po[p]] = p2;
	if (i != j) {        /* skip diagonal elements to avoid 
				referencing them twice in <xy> and <yx> */
	  irn[p2] = i+nv;
	  p2++;		
	}
      }
      for (p = iicn[j]; p <= iicn[j+1]-1; p++) { /* <yx> */
	cms->pyx[p] = p2;
	irn[p2] = iirn[p]+nv;
	p2++;
      }
    }
    else 
      for (p = iicn[j]; p <= iicn[j+1]-1; p++) { 
	cms->pyx[p] = -9999.;	/* these will never be referenced! */
	cms->pxy[p] = -9999.;
      }
  }

  /* imaginary parts only */
  for (j = 1; j <= nv; j++) {
    icn[j+nv] = p2;
    for (p = iicn[j]; p <= iicn[j+1]-1; p++) { /* <yy> */
      cms->pyy[p] = p2;
      irn[p2] = iirn[p]+nv;
      p2++;
    }
  }
  icn[2*nv+1] = nz+1;

  if (ints->imaginary_flag) {
    free_1imatrix(po);
    free_1imatrix(ticn);
    free_1imatrix(tirn);
  }

  return;
}

/* 
 * free memory allocated for lookup tables 
 * PARAMS
 *     cms [I,O]  struct for the covariance matrix
 */
void freeidxcms(cm_struct *cms)
{
  free_1imatrix(cms->pxx);
  free_1imatrix(cms->pyy);
  free_1imatrix(cms->pxy);
  free_1imatrix(cms->pyx);

  return;
}

/*
 * returns the number of elements of a covariance matrix derived from N_z
 * elements. this is 2*N_z if no mosaicing is performed, and 4*N_z-N_v
 * for mosaicing.
 * PARAMS
 *     ints [I]  data struct containing the overlap integrals
 * RETURNS
 *     number of non-zero elements in lower half of covariance matrix
 */
int countcms(integral_struct *ints)
{
  int j;
  int nd;

  if (ints->imaginary_flag) {
    nd = 0;
    for (j = 1; j <= ints->nv; j++)
      if (j == ints->irn[ints->icn[j]])
	nd++;
    return 4*ints->nz - nd;
  }
  else 
    return 2 * ints->nz;
}

/* 
 * calculates the elements of the signal covariance matrix given a set of 
 * CMB and Galactic flat band powers.
 * uses the integrals computed in makeintegrals. 
 * PARAMS
 *     cps1d  [I]  the trial CMB band powers
 *     gps1d  [I]  the trial Galactic band powers
 *     na     [I]  number of power spectrum bins
 *     sp     [I]  spectral index of Galactic component
 *     freq   [I]  observing frequencies for each observed field
 *     nf     [I]  number of observed fields
 *     freq0  [I]  reference frequency at which power spectra are given
 *     convf  [I]  conversion factor for normalisation of fluctuations
 *     ints   [I]  struct containing the overlap integrals
 *     cms  [I,O]  struct containing the signal covariance matrix
 *     rank   [I]  rank of processor given by MPI; if MPI is not used, set to 0
 *     npe    [I]  number of processors; if MPI is not used set to 1
 */ 
void makesm(double *cps1d, double *gps1d, int na, 
	    double sp, double *freq, int nf, double freq0, double convf, 
	    integral_struct *ints, cm_struct *cms, int rank, int npe)
{
  int    i, j, m, p, kmax1, kmax2, nbin, nv;
  double cn, gn;
  double cnf[NAMAX+1], gnf[NAMAX+1];
  double value1r, value2r, value1i, value2i, pstot;
  int    *maxbin1 = ints->maxbin1;
  int    *maxbin2 = ints->maxbin2;
  int    *freqn = ints->freqn;
  float  *integral1r[N_BINS+1], *integral2r[N_BINS+1]; 
  float  *integral1i[N_BINS+1], *integral2i[N_BINS+1];
  int    n_bins, kwdth;
  int    *iirn = ints->irn;
  int    *iicn = ints->icn;
  int    *pxx = cms->pxx, *pyy = cms->pyy, *pxy = cms->pxy, *pyx = cms->pyx;
  double *cm = cms->cm;
  int    jfirst, jlast;
#if defined(MPI)
  int    *jidx = NULL;
#endif
  
  nv = ints->nv;
  n_bins = ints->n_bins;
  kwdth = n_bins/2;

  /* initialise variables*/
  for (i = 1; i <= n_bins; i++) {
    integral1r[i] = ints->integral1r[i];
    integral2r[i] = ints->integral2r[i];
    integral1i[i] = ints->integral1i[i];
    integral2i[i] = ints->integral2i[i];
  }

  /* frequency dependence */
  for (i = 1; i <= nf; i++) {
    cnf[i] = cmbfcn(T_CMB, freq[i]) / cmbfcn(T_CMB, freq0);
    gnf[i] = pow(freq[i]/freq0, sp);
  }

  /* 
   * determine what columns this processor needs to calculate 
   */
  jfirst = 1;
  jlast = nv;
#if defined(MPI)  
  if (npe > 1) {
    jidx = malloc_1imatrix(npe+1);
    get_mpi_integral_cols(ints->icn, nv, ints->nz, npe, jidx);
    jfirst = jidx[rank+1];
    jlast = jidx[rank+2]-1;
    for (p = 1; p <= cms->nz; p++)
      cm[p] = 0.;
  }
#endif

  /* 
   * calculate covariance matrix
   */
  for(j = jfirst; j <= jlast; j++) {
    for(p = iicn[j]; p <= iicn[j+1]-1; p++) {
      /* determine row index */
      i  = iirn[p];

      /* find the bins where the integrals have their maxima */
      kmax1 = maxbin1[p];
      kmax2 = maxbin2[p];

      /* calculate the frequency dependence of CMB and Galactic components
       *  (in brightness units) */
      cn = cnf[freqn[i]] * cnf[freqn[j]] * convf;
      gn = gnf[freqn[i]] * gnf[freqn[j]] * convf;

      /* calculate the corresponding element of covariance matrix
       * first sum over the components, then multiply by precomputed 
       * integral. sum over the three bins around the maximum. */ 
      /* for u_i - u_j... */
      value1r = 0.;		/* real part of <V_i V^*_j> */
      for(m = 1; m <= n_bins; m++) {
	nbin = kmax1 - 1 - kwdth + m;
	if ((nbin >= 1) && (nbin <= na)) {
	  pstot = (cn*cps1d[nbin] + gn*gps1d[nbin]);
	  value1r += pstot * integral1r[m][p];
	}
      }

      /* for u_i + u_j... */
      value2r = 0.;		/* real part of <V_i V_j> */
      for(m = 1; m <= n_bins; m++) {
	nbin = kmax2 - 1 - kwdth + m;
	if ((nbin >= 1) && (nbin <= na)) {
	  pstot = (cn*cps1d[nbin] + gn*gps1d[nbin]);
	  value2r += pstot * integral2r[m][p];
	}
      }

      /* compute the real-real and imag-imag correlations */
      cm[pxx[p]] = 0.5 * (value1r + value2r); /* <xx> */
      cm[pyy[p]] = 0.5 * (value1r - value2r); /* <yy> */


      /* for mosaiced observations include imaginary parts of integrals.*/
      if (ints->imaginary_flag) {
	value1i = 0.;		/* imag part of <V_i V^*_j> */
	for(m = 1; m <= n_bins; m++) {
	  nbin = kmax1 - 1 - kwdth + m;
	  if ((nbin >= 1) && (nbin <= na)) {
	    pstot = (cn*cps1d[nbin] + gn*gps1d[nbin]);
	    value1i += pstot * integral1i[m][p];
	  }
	}
	value2i = 0.;		/* imag part of <V_i V_j> */
	for(m = 1; m <= n_bins; m++) {
	  nbin = kmax2 - 1 - kwdth + m;
	  if ((nbin >= 1) && (nbin <= na)) {
	    pstot = (cn*cps1d[nbin] + gn*gps1d[nbin]);
	    value2i += pstot * integral2i[m][p];
	  }
	}
	cm[pyx[p]] = 0.5 * (value2i + value1i); /* <yx> */
	cm[pxy[p]] = 0.5 * (value2i - value1i); /* <xy> */
      }

    }
  }
  
#if defined(MPI)
  /* 
   * generate the global covariance matrix as a sum of the matrices held
   * on individual processes ... 
   */
  if (npe > 1) {
    double *newcms;

    newcms = malloc_1dmatrix(cms->nz);
    MPI_Allreduce(cms->cm+C_OFFSET, newcms+C_OFFSET, cms->nz, 
		  MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    free_1dmatrix(cms->cm);
    cms->cm = newcms;
    free_1imatrix(jidx);
  }
#endif

  return;
}

/* 
 * calculate the noise covariance matrix
 * PARAMS
 *     rms    [I]  rms per real or imaginary part of visibility
 *     nv     [I]  number of visibilities
 *     ints   [I]  struct containing the overlap integrals
 *     cms    [I,O]  struct containing the signal covariance matrix
 *     rank   [I]  rank of processor given by MPI; if MPI is not used, set to 0
 *     npe    [I]  number of processors; if MPI is not used set to 1
 */
void makenm(double *rms, int nv, integral_struct *ints,
	    cm_struct *cms, int rank, int npe)
{
  int *pxx = cms->pxx, *pyy = cms->pyy, *pxy = cms->pxy, *pyx = cms->pyx;
  int i, j, p;

  if (nv != ints->nv) myerror("wrong number of visibilities in makenm()");
    
  /* diagonal instrumental noise covariance N_{ij} */
  for (j = 1; j <= nv; j++) 
    for (p = ints->icn[j]; p <= ints->icn[j+1]-1; p++) {
      i = ints->irn[p];
      if (i == j) {
	cms->cm[pxx[p]] = rms[i]*rms[i];
	cms->cm[pyy[p]] = rms[i]*rms[i];
      }
      else {
	cms->cm[pxx[p]] = 0.;
	cms->cm[pyy[p]] = 0.;
      }
      if (ints->imaginary_flag) {
	cms->cm[pxy[p]] = 0.;
	cms->cm[pyx[p]] = 0.;
      }
      
    }

  return;
}


/* 
 * calculates the covariance matrix of signal and noise 
 * PARAMS
 *     rms    [I]  rms per real or imaginary part of visibility
 *     nv     [I]  number of visibilities
 *     cps1d  [I]  the trial CMB band powers
 *     gps1d  [I]  the trial Galactic band powers
 *     na     [I]  number of power spectrum bins
 *     sp     [I]  spectral index of Galactic component
 *     freq   [I]  observing frequencies for each observed field
 *     nf     [I]  number of observed fields
 *     freq0  [I]  reference frequency at which power spectra are given
 *     convf  [I]  conversion factor for normalisation of fluctuations
 *     ints   [I]  struct containing the overlap integrals
 *     cms    [I,O]  struct containing the signal covariance matrix
 *     rank   [I]  rank of processor given by MPI; if MPI is not used, set to 0
 *     npe    [I]  number of processors; if MPI is not used set to 1
 */ 
void makecm(double *rms, int nv, double *cps1d, double *gps1d, int na, 
	    double sp, double *freq, int nf, double freq0, double convf, 
	    integral_struct *ints, cm_struct *cms, int rank, int npe)
{
  int *pxx = cms->pxx, *pyy = cms->pyy;
  int i, j, p;

  if (nv != ints->nv) myerror("wrong number of visibilities in makecm()");

  /* calculate signal covariance */
  makesm(cps1d, gps1d, na, sp, freq, nf, freq0, convf, ints, cms, rank, npe);

  /* add diagonal instrumental noise covariance N_{ij} */
  for (j = 1; j <= nv; j++) 
    for (p = ints->icn[j]; p <= ints->icn[j+1]-1; p++) {
      i = ints->irn[p];
      if (i == j) {
	cms->cm[pxx[p]] += rms[i]*rms[i];
	cms->cm[pyy[p]] += rms[i]*rms[i];
      }
    }
  return;
}

/*************************************************************************
 * obtain the signal and noise covariance matrices for signal-to-noise
 * eigenmodes.
 * only one bandpower is considered as signal, the others are treated as noise
 * PARAMS 
 *     rms      [I]   vector of rms error per real/imag part of visibilty
 *     nv       [I]   number of visibilities
 *     cps1d    [I]   power spectrum of CMB in each bin
 *     gps1d    [I]   power spectrum of Galactic component
 *     na       [I]   total number of bins
 *     sp       [I]   spectral index of Galaxy
 *     freq     [I]   observing frequency for each field
 *     nf       [I]   number of observed fields
 *     freq0    [I]   reference frequency
 *     convf    [I]   conversion factor for PS units
 *     ita      [I]   bin wrt which the SN modes are calculated
 *     int_data [I]   integrals
 *     sm_data  [I,O] signal covariance
 *     nm_data  [I,O] noise covariance
 *     rank     [I]   rank of processor; if MPI is not used, set to 0
 *     npe      [I]   number of processors; if MPI is not used set to 1
 */
void makesmnm(double *rms, int nv, double *cps1d, double *gps1d, int na, 
	      double sp, double *freq, int nf, double freq0, double convf, 
	      int ita, integral_struct *int_data,
	      cm_struct *sm_data, cm_struct *nm_data, int rank, int npe)
{
  int    i, j, m, p, kmax1, kmax2, nbin;
  double cn, gn;
  double cnf[NAMAX+1], gnf[NAMAX+1];
  double value1rs, value2rs, value1is, value2is;
  double value1rn, value2rn, value1in, value2in;
  double pstots, pstotn;
  int    *maxbin1 = int_data->maxbin1;
  int    *maxbin2 = int_data->maxbin2;
  int    *freqn = int_data->freqn;
  float  *integral1r[N_BINS+1], *integral2r[N_BINS+1];
  float  *integral1i[N_BINS+1], *integral2i[N_BINS+1];
  int    n_bins, kwdth;
  int    *iirn = int_data->irn;
  int    *iicn = int_data->icn;
  int    *pxx = sm_data->pxx, *pyy = sm_data->pyy;
  int    *pxy = sm_data->pxy, *pyx = sm_data->pyx;
  int    jfirst, jlast;
#if defined(MPI)
  int    *jidx = NULL;
#endif
  
  n_bins = int_data->n_bins;
  kwdth = n_bins/2;

  /* initialise variables*/
  for (i = 1; i <= n_bins; i++) {
    integral1r[i] = int_data->integral1r[i];
    integral2r[i] = int_data->integral2r[i];
    integral1i[i] = int_data->integral1i[i];
    integral2i[i] = int_data->integral2i[i];
  }

  for (i = 1; i <= nf; i++) {
    cnf[i] = cmbfcn(T_CMB, freq[i]) / cmbfcn(T_CMB, freq0);
    gnf[i] = pow(freq[i]/freq0, sp);
  }

  /* 
   * determine what columns this processor needs to calculate 
   */
  jfirst = 1;
  jlast = nv;
#if defined(MPI)  
  if (npe > 1) {
    jidx = malloc_1imatrix(npe+1);
    get_mpi_integral_cols(int_data->icn, nv, int_data->nz, npe, jidx);
    jfirst = jidx[rank+1];
    jlast = jidx[rank+2]-1;
    for (p = 1; p <= sm_data->nz; p++)
      sm_data->cm[p] = 0.;
    for (p = 1; p <= sm_data->nz; p++)
      nm_data->cm[p] = 0.;
  }
#endif

  /* 
   * calculate covariance matrix
   */
  for(j = jfirst; j <= jlast; j++) {
    for(p = iicn[j]; p <= iicn[j+1]-1; p++) {
      /* determine row index */
      i  = iirn[p];

      /* find the bins where the integrals have their maxima */
      kmax1 = maxbin1[p];
      kmax2 = maxbin2[p];

      /* calculate the frequency dependence of CMB and Galactic components
       *  (in brightness units) */
      cn = cnf[freqn[i]] * cnf[freqn[j]] * convf;
      gn = gnf[freqn[i]] * gnf[freqn[j]] * convf;

      /* calculate the corresponding element of covariance matrix
       * sum over the n_bins bins around the maximum, using the
       * precomputed overlap integrals. 
       * any signal outside the current bin is treated as noise.
       */ 
      /* for u_i - u_j... */
      value1rs = 0.; value1rn = 0.;		/* real part of <V_i V^*_j> */
      for(m = 1; m <= n_bins; m++) {
	nbin = kmax1 - 1 - kwdth + m;
	if ((nbin >= 1) && (nbin <= na)) {
	  if (nbin == ita) {
	    pstots = cn*cps1d[nbin];
	    pstotn = gn*gps1d[nbin];
	  }
	  else {
	    pstots = 0.;
	    pstotn = (cn*cps1d[nbin] + gn*gps1d[nbin]);
	  }
	  value1rs += pstots * integral1r[m][p];
	  value1rn += pstotn * integral1r[m][p];
	}
      }

      /* for u_i + u_j... */
      value2rs = 0.; value2rn = 0.;		/* real part of <V_i V_j> */
      for(m = 1; m <= n_bins; m++) {
	nbin = kmax2 - 1 - kwdth + m;
	if ((nbin >= 1) && (nbin <= na)) {
	  if (nbin == ita) {
	    pstots = cn*cps1d[nbin];
	    pstotn = gn*gps1d[nbin];
	  }
	  else {
	    pstots = 0.;
	    pstotn = (cn*cps1d[nbin] + gn*gps1d[nbin]);
	  }
	  value2rs += pstots * integral2r[m][p];
	  value2rn += pstotn * integral2r[m][p];
	}
      }

      /* compute the different correlations */
      sm_data->cm[pxx[p]] = 0.5 * (value1rs + value2rs); /* <xx> */
      nm_data->cm[pxx[p]] = 0.5 * (value1rn + value2rn); /* <xx> */
      sm_data->cm[pyy[p]] = 0.5 * (value1rs - value2rs); /* <yy> */
      nm_data->cm[pyy[p]] = 0.5 * (value1rn - value2rn); /* <yy> */


      /* 
       * for mosaiced observations include imaginary parts of integrals.
       */
      if (int_data->imaginary_flag) {
	value1is = 0.; value1in = 0.;	       /* imag part of <V_i V^*_j> */
	for(m = 1; m <= n_bins; m++) {
	  nbin = kmax1 - 1 - kwdth + m;
	  if ((nbin >= 1) && (nbin <= na)) {
	    if (nbin == ita) {
	      pstots = cn*cps1d[nbin];
	      pstotn = gn*gps1d[nbin];
	    }
	    else {
	      pstots = 0.;
	      pstotn = (cn*cps1d[nbin] + gn*gps1d[nbin]);
	    }
	    value1is += pstots * integral1i[m][p];
	    value1in += pstotn * integral1i[m][p];
	  }
	}
	value2is = 0.; value2in = 0.;		/* imag part of <V_i V_j> */
	for(m = 1; m <= n_bins; m++) {
	  nbin = kmax2 - 1 - kwdth + m;
	  if ((nbin >= 1) && (nbin <= na)) {
	    if (nbin == ita) {
	      pstots = cn*cps1d[nbin];
	      pstotn = gn*gps1d[nbin];
	    }
	    else {
	      pstots = 0.;
	      pstotn = (cn*cps1d[nbin] + gn*gps1d[nbin]);
	    }
	    value2is += pstots * integral2i[m][p];
	    value2in += pstotn * integral2i[m][p];
	  }
	}
	sm_data->cm[pyx[p]] = 0.5 * (value2is + value1is); /* <yx> */
	sm_data->cm[pxy[p]] = 0.5 * (value2is - value1is); /* <xy> */
	nm_data->cm[pyx[p]] = 0.5 * (value2in + value1in); /* <yx> */
	nm_data->cm[pxy[p]] = 0.5 * (value2in - value1in); /* <xy> */
      }

      if (i == j) {
	nm_data->cm[pxx[p]] += rms[i]*rms[i];
	nm_data->cm[pyy[p]] += rms[i]*rms[i];
      }
    }
  }
  
#if defined(MPI)
    /* 
     * generate the global covariance matrix as a sum of the matrices held
     * on individual nodes
     */
  if (npe > 1) {
    double *newscm, *newncm;

    /* first start with signal covariance ... */
    newscm = malloc_1dmatrix(sm_data->nz);
    MPI_Allreduce(sm_data->cm+C_OFFSET, newscm+C_OFFSET, sm_data->nz, 
		  MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    newncm = sm_data->cm;
    sm_data->cm = newscm;
    /* now do the noise covariance */
    MPI_Allreduce(nm_data->cm+C_OFFSET, newncm+C_OFFSET, nm_data->nz, 
		  MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    free_1dmatrix(nm_data->cm);
    nm_data->cm = newncm;
  }
  free_1imatrix(jidx);
#endif

  return;
}

/*-----*-----------------------------------------------------------------*/

#define DER_CMB 0
#define DER_GAL 1
#define DER_SP 2

/* calculates the derivatives of the elements of the signal covariance matrix
 * with respect to the parameters (CMB and galactic band powers and
 * spectral index).
 * PARAMS
 *     cps1d  [I]  the trial CMB band powers
 *     gps1d  [I]  the trial Galactic band powers
 *     na     [I]  number of power spectrum bins
 *     sp     [I]  spectral index of Galactic component
 *     freq   [I]  observing frequencies for each observed field
 *     nf     [I]  number of observed fields
 *     freq0  [I]  reference frequency at which power spectra are given
 *     convf  [I]  conversion factor for normalisation of fluctuations
 *     idx    [I]  index of parameter wrt whom the derivative is calculated:
 *                 1..na CMB PS, na+1..2*na GAL PS, 2*na+1 spectral index 
 *     ints   [I]  struct containing the overlap integrals
 *     cms    [I,O]  struct containing the signal covariance matrix
 *     rank   [I]  rank of processor given by MPI; if MPI is not used, set to 0
 *     npe    [I]  number of processors; if MPI is not used set to 1
 */
void makedsm(double *cps1d, double *gps1d, int na, 
	     double sp, double *freq, int nf, double freq0,  double convf, 
	     int idx, integral_struct *ints, cm_struct *cms,
	     int rank, int npe)
{
  int    i, j, m, p, nbin, nv, derflag;
  int    kmax1, kmax2, m1, m2; 
  double cn, gn;
  double cnf[NAMAX+1], gnf[NAMAX+1], snf[NAMAX+1];
  double value1r, value2r, value1i, value2i;
  int    *maxbin1 = ints->maxbin1;
  int    *maxbin2 = ints->maxbin2;
  int    *freqn = ints->freqn;
  float  *integral1r[N_BINS+1], *integral2r[N_BINS+1];
  float  *integral1i[N_BINS+1], *integral2i[N_BINS+1];
  int    *iirn = ints->irn;
  int    *iicn = ints->icn;
  int    n_bins, kwdth;
  int    *pxx = cms->pxx, *pyy = cms->pyy, *pxy = cms->pxy, *pyx = cms->pyx;
  double *cm = cms->cm;
  int    jfirst, jlast;
#if defined(MPI)
  int    *jidx;
#endif

  nv = ints->nv;
  n_bins = ints->n_bins;
  kwdth = n_bins/2;

  derflag = DER_CMB;
  if (idx > 2 * na) {		/* compute derivative wrt spectral index */
    derflag = DER_SP;
  }
  else {
    if (idx > na) {		/* compute derivative wrt Galactic PS  */
      derflag = DER_GAL;
      idx -= na;
    }
  }

  /* initialise variables*/
  for (i = 1; i <= n_bins; i++) {
    integral1r[i] = ints->integral1r[i];
    integral2r[i] = ints->integral2r[i];
    integral1i[i] = ints->integral1i[i];
    integral2i[i] = ints->integral2i[i];
  }

  /* frequency dependence */
  for (i = 1; i <= nf; i++) {
    cnf[i] = cmbfcn(T_CMB, freq[i]) / cmbfcn(T_CMB, freq0);
    gnf[i] = pow(freq[i]/freq0, sp);
    snf[i] = log(freq[i]/freq0);
  }

  /* 
   * determine what columns this processor needs to calculate 
   */
  jfirst = 1;
  jlast = nv;
#if defined(MPI)  
  jidx = malloc_1imatrix(npe+1);
  get_mpi_integral_cols(ints->icn, nv, ints->nz, npe, jidx);
  jfirst = jidx[rank+1];
  jlast = jidx[rank+2]-1;
  for (p = 1; p <= cms->nz; p++)
    cm[p] = 0.;
#endif

  /* 
   * calculate derivative of covariance matrix
   */
  for(j = jfirst; j <= jlast; j++) {
    for(p = iicn[j]; p <= iicn[j+1]-1; p++) {
      /* determine row index */
      i  = iirn[p];

      /* find the bins where the two integrals have their maximum */
      kmax1 = maxbin1[p];
      kmax2 = maxbin2[p];
      m1 = idx - kmax1 + 1 + kwdth; 
      m2 = idx - kmax2 + 1 + kwdth; 

      value1r = 0.; value1i = 0.;
      value2r = 0.; value2i = 0.;
      if (derflag == DER_CMB) {	/* CMB derivatives */
	cn = cnf[freqn[i]] * cnf[freqn[j]] * convf;

	if ((m1 >= 1) && (m1 <= n_bins)) /* real parts */
	  value1r = cn * integral1r[m1][p];
	if ((m2 >= 1) && (m2 <= n_bins)) 
	  value2r = cn * integral2r[m2][p];

	if (ints->imaginary_flag) { /* imaginary parts */
	  if ((m1 >= 1) && (m1 <= n_bins)) 
	    value1i = cn * integral1i[m1][p];
	  if ((m2 >= 1) && (m2 <= n_bins)) 
	    value2i = cn * integral2i[m2][p];
	}
      }

      if (derflag == DER_GAL) {	/* Galactic derivatives */
	gn = gnf[freqn[i]] * gnf[freqn[j]] * convf;
	
	if ((m1 >= 1) && (m1 <= n_bins)) /* real parts */
	  value1r = gn * integral1r[m1][p];
	if ((m2 >= 1) && (m2 <= n_bins)) 
	  value2r = gn * integral2r[m2][p];

	if (ints->imaginary_flag) { /* imaginary parts */
	  if ((m1 >= 1) && (m1 <= n_bins)) 
	    value1i = gn * integral1i[m1][p];
	  if ((m2 >= 1) && (m2 <= n_bins)) 
	    value2i = gn * integral2i[m2][p];
	}
      }

      if (derflag == DER_SP) {	/* spectral index */
	gn = gnf[freqn[i]] * gnf[freqn[j]] *
	  (snf[freqn[i]] + snf[freqn[j]]) * convf;

	for(m = 1; m <= n_bins; m++) {
	  nbin = kmax1 - 1 - kwdth + m;
	  if ((nbin >= 1) && (nbin <= na))
	    value1r += gn * gps1d[nbin] * integral1r[m][p];
	}
	for(m = 1; m <= n_bins; m++) {
	  nbin = kmax2 - 1 - kwdth + m;
	  if ((nbin >= 1) && (nbin <= na))
	    value2r += gn * gps1d[nbin] * integral2r[m][p];
	}
	if (ints->imaginary_flag) {
	  for(m = 1; m <= n_bins; m++) {
	    nbin = kmax1 - 1 - kwdth + m;
	    if ((nbin >= 1) && (nbin <= na))
	      value1i += gn * gps1d[nbin] * integral1i[m][p];
	  }
	  for(m = 1; m <= n_bins; m++) {
	    nbin = kmax2 - 1 - kwdth + m;
	    if ((nbin >= 1) && (nbin <= na))
	      value2i += gn * gps1d[nbin] * integral2i[m][p];
	  }
	}
      }

      /* for the real and imaginary parts of the likelihood */
      cm[pxx[p]] = 0.5 * (value1r + value2r); /* <xx> */
      cm[pyy[p]] = 0.5 * (value1r - value2r); /* <yy> */

      if (cms->imaginary_flag) {
	cm[pyx[p]] = 0.5 * (value2i + value1i); /* <yx> */
	cm[pxy[p]] = 0.5 * (value2i - value1i); /* <xy> */
      }
    }
  }

#if defined(MPI)
  /* 
   * generate the global covariance matrix as a sum of the matrices held
   * on individual processes ... 
   */
  {
    double *newcms;

    newcms = malloc_1dmatrix(cms->nz);
    MPI_Allreduce(cms->cm+C_OFFSET, newcms+C_OFFSET, cms->nz, 
		  MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    free_1dmatrix(cms->cm);
    cms->cm = newcms;
  }
  free_1imatrix(jidx);
#endif

  return;
}

/*
 * derivatives of covariance matrix wrt to parameters 
 * PARAMS
 *     rms    [I]  rms per real or imaginary part of visibility
 *     nv     [I]  number of visibilities
 *     cps1d  [I]  the trial CMB band powers
 *     gps1d  [I]  the trial Galactic band powers
 *     na     [I]  number of power spectrum bins
 *     sp     [I]  spectral index of Galactic component
 *     freq   [I]  observing frequencies for each observed field
 *     nf     [I]  number of observed fields
 *     freq0  [I]  reference frequency at which power spectra are given
 *     convf  [I]  conversion factor for normalisation of fluctuations
 *     idx    [I]  index of parameter wrt whom the derivative is calculated:
 *                 1..na CMB PS, na+1..2*na GAL PS, 2*na+1 spectral index 
 *     ints   [I]  struct containing the overlap integrals
 *     cms    [I,O]  struct containing the signal covariance matrix
 *     rank   [I]  rank of processor given by MPI; if MPI is not used, set to 0
 *     npe    [I]  number of processors; if MPI is not used set to 1
 */
void makedcm(double *rms, int nv, double *cps1d, double *gps1d, int na, 
	     double sp, double *freq, int nf, double freq0, double convf, 
	     int idx, integral_struct *ints, cm_struct *cms,
	     int rank, int npe)
{
  makedsm(cps1d, gps1d, na, sp, 
	  freq, nf, freq0, convf, idx, ints, cms, rank, npe);
  return;
}

/*-----*-----------------------------------------------------------------*/

/*
 * copy a covariance matrix
 * PARAMS
 *     cm1 [I]  the matrix to be copied
 *     cm2 [O]  the target 
 */
void copycms(cm_struct *cm1, cm_struct *cm2)
{
  cm2->nv = cm1->nv;
  cm2->nz = cm1->nz;
  cm2->imaginary_flag = cm1->imaginary_flag;
  cm2->irn = cm1->irn;
  cm2->icn = cm1->icn;
  cm2->cm = cm1->cm;
  cm2->pxx = cm1->pxx;
  cm2->pyx = cm1->pyx;
  cm2->pxy = cm1->pxy;
  cm2->pyy = cm1->pyy;

  return;
}
/*-----*-----------------------------------------------------------------*/
