/*  Time-stamp: <12 Nov 2002 19:49 GMT like_func.c maising> */

#include <stdio.h>
#include <time.h>
#include <math.h>
#include "like_func.h"
#include "my_malloc2.h"
#include "cquadform.h"

#define ROOT 0

/*************************************************************************
 * compute the logarithm of the likelihood
 * PARAMS
 *     cps1d    [I]   power spectrum of CMB in each bin
 *     gps1d    [I]   power spectrum of Galactic component
 *     na       [I]   total number of bins
 *     sp       [I]   spectral index of Galaxy
 *     sparseflag [I]   use sparse matrix routines or not?
 *     vis      [I]   vector of real and imaginary parts of visibilities
 *     rms      [I]   vector of rms error per real/imag part of visibilty
 *     nv       [I]   number of visibilities
 *     freq     [I]   observing frequencies
 *     nf       [I]   number of observed fields
 *     freq0    [I]   reference frequency
 *     convf    [I]   conversion factor for PS units
 *     chisq    [O] pointer to value of \chi^2
 *     lndet    [O] pointer to value ln det C
 *     int_data [I]   integrals
 *     cm_data  [I,O] covariance matrix
 *     rank   [I]  rank of processor given by MPI; if MPI is not used, set to 0
 *     npe    [I]  number of processors; if MPI is not used set to 1
 * RETURNS
 *     log likelihood
 */
double loglike(double *cps1d, double *gps1d, int na, double sp,
	       int sparseflag, double *vis, double *rms, int nv,
	       double *freq, double nf, double freq0, double convf,
	       double *chisq, double *lndet,
	       integral_struct *int_data, cm_struct *cm_data,
	       int rank, int npe)
{
  double  err;

  /* calculate theoretical covariance matrix*/
  makecm(rms, nv, cps1d, gps1d, na, sp, freq, nf, freq0, convf, 
	 int_data, cm_data, rank, npe);

  /* evaluate the log(likelihood) */
  if (sparseflag != 0) {
    /* use sparse matrix routines */
    cspquadform(cm_data->icn, cm_data->irn, cm_data->cm, cm_data->nz, 
		vis, 2*nv, chisq, lndet, &err, int_data->imaginary_flag);
#if 0
    printf("  approximate error   = %g\n", err);
#endif
  } 
  else {
    /* use dense matrix routines */
    if (int_data->imaginary_flag)  /* mosaicing */
      claquadform(cm_data->icn, cm_data->irn, cm_data->cm, cm_data->nz, 
		  vis, 2*nv, chisq, lndet, 0, rank, npe);
    else {			/* no mosaicing */
      double chisqr, chisqi, lndetr, lndeti;

      claquadform(cm_data->icn, cm_data->irn, cm_data->cm, cm_data->nz, 
		  vis, nv, &chisqr, &lndetr, 0, rank, npe);
      claquadform(cm_data->icn, cm_data->irn, cm_data->cm, cm_data->nz, 
		  vis+nv, nv, &chisqi, &lndeti, 1, rank, npe);
      *lndet = lndetr + lndeti;
      *chisq = chisqr + chisqi;
    }
  }

  return (-0.5 * (*lndet + *chisq));
}

/*
 * global variables and wrapper routines to create a function suitable
 * for NR minimisation routines
 */
struct lfunc_struct {
  int np, na; 
  double sp;
  int sparseflag;
  double *ra, *dec, ra_ref, dec_ref;
  int *freqn;
  double *u, *v, *w;
  double *vis, *rms;
  int nv;
  double *freq, *antwdth, *sigma;
  int nf;
  double *rbins;
  int bin;
  int *fieldn;
  int ni;
  double freq0, convf;
  int saveflag, imagflag, ita;
  int verbose;
  int rank, npe;
  int fcount;
  double fnorm;
};

static struct lfunc_struct lfunc_data;

int query_loglike_func_count() { return lfunc_data.fcount; }
void reset_loglike_func_count() { lfunc_data.fcount = 0; return; }
void set_fnorm(double val) { lfunc_data.fnorm = val; return; }


/*
 * initialise data for func()
 */
void init_loglike_func(int np, int na, int sparseflag,
		       double *ra, double *dec, double ra_ref, double dec_ref,
		       int *freqn, double *u, double *v, double *w,
		       double *vis, double *rms, int nv,
		       double *freq, double *antwdth, double *sigma, int nf, 
		       double freq0, double convf, double *rbins, int ita,
		       int *fieldn, int ni, int imagflag, int saveflag,
		       int verbose, int rank, int npe)
{
  lfunc_data.np = np;
  lfunc_data.na = na;
  lfunc_data.sparseflag = sparseflag;
  lfunc_data.ra = ra;
  lfunc_data.dec = dec;
  lfunc_data.ra_ref = ra_ref;
  lfunc_data.dec_ref = dec_ref;
  lfunc_data.freqn = freqn;
  lfunc_data.u = u;
  lfunc_data.v = v;
  lfunc_data.w = w;
  lfunc_data.vis = vis;
  lfunc_data.rms = rms;
  lfunc_data.nv = nv;
  lfunc_data.freq = freq;
  lfunc_data.antwdth = antwdth;
  lfunc_data.sigma = sigma;
  lfunc_data.nf = nf;
  lfunc_data.freq0 = freq0;
  lfunc_data.convf = convf;
  lfunc_data.fieldn = fieldn;
  lfunc_data.ni = ni;
  lfunc_data.verbose = verbose;
  lfunc_data.fcount = 0;
  lfunc_data.rbins = rbins;
  lfunc_data.ita= ita;
  lfunc_data.imagflag = imagflag;
  lfunc_data.saveflag = saveflag;
  lfunc_data.rank = rank;
  lfunc_data.npe = npe;

  return;
}

/*
 * a wrapper function for loglike() to be used with NR minimisation
 * algorithms that expect a single vector argument
 * PARAMS
 *      x  [I]  parameter vector: 
 *              the first na entries are the CMB PS, the second na entries
 *              the GAL PS and the (2*na+1)-st entry the Gal spectral index
 * RETURNS
 *      the log likelihood 
 */  
double loglike_func(double *x)
{
  double   *gps, *cps;
  double   chisq, lndet, ll, value, sp;
  double   likeval, chisqval, lndetval;
  int      i, j, field, cnz;
  clock_t  timein, timeout;
  char     msg[256];
  cm_struct cm_data;
  integral_struct integral_data;
  int     *freqn2, nv2;
  double  *u2, *v2, *w2, *vis2, *rms2;

  lfunc_data.fcount++;
  if ((lfunc_data.rank == ROOT) && lfunc_data.verbose) {
    sprintf(msg, "fcount = %5i\n", lfunc_data.fcount);
    printf(msg);
    write_file("func.log", msg);
  }

  cps = malloc_1dmatrix(lfunc_data.na);
  for (i = 1; i <= lfunc_data.na; i++)
    cps[i] = fabs(x[i]) * 1.e-9;
 
  /* set galactic power spectrum */
  gps = malloc_1dmatrix(lfunc_data.na);
  if (lfunc_data.np >= 2*lfunc_data.na) {
    for (i = 1; i <= lfunc_data.na; i++)
      gps[i] = fabs(x[lfunc_data.na+i]) * 1.e-9;
  }
  else {
    for (i = 1; i <= lfunc_data.na; i++)
      gps[i] = 0.;
    sp = 0.;
  }
  if (lfunc_data.np >= 2*lfunc_data.na+1) 
    sp = x[2*lfunc_data.na+1];
  else
    sp = 0.;

  if ((lfunc_data.rank == ROOT) && lfunc_data.verbose) {
    for (i = 1; i <= lfunc_data.np; i++) {
      sprintf(msg,  "x(%2i) = %14.8f\n", i, x[i]);
      printf(msg);
      write_file("func.log", msg);
    }
  }
  timein = clock();

  likeval = 0.;
  chisqval = 0.;
  lndetval = 0.;
  for (field = 1; field <= lfunc_data.ni; field++) { 
    /* calculate reduced number of visibilities for each field */
    nv2 = 0;
    for (j = 1; j <= lfunc_data.nv; j++)
      if (lfunc_data.fieldn[lfunc_data.freqn[j]] == field)
	nv2 ++;
    if (lfunc_data.rank == ROOT)
      printf("  Number of observed visibilities = %i\n", nv2);

    /* skip bin if empty */
    if (nv2 > 0) {
      freqn2 = malloc_1imatrix(nv2);
      u2 = malloc_1dmatrix(nv2); v2 = malloc_1dmatrix(nv2); 
      w2 = malloc_1dmatrix(nv2);
      vis2 = malloc_1dmatrix(2*nv2); rms2 = malloc_1dmatrix(nv2);
      i = 0;
      for (j = 1; j <= lfunc_data.nv; j++)
	if (lfunc_data.fieldn[lfunc_data.freqn[j]] == field) {
	  i++;
	  freqn2[i] = lfunc_data.freqn[j];
	  u2[i] = lfunc_data.u[j]; v2[i] = lfunc_data.v[j]; 
	  w2[i] = lfunc_data.w[j];
	  vis2[i] = lfunc_data.vis[j]; 
	  vis2[i+nv2] = lfunc_data.vis[j+lfunc_data.nv]; 
	  rms2[i] = lfunc_data.rms[j];
	}
     
      /*
       * calculate the integrals required to form the covariance matrix
       */
      if ((lfunc_data.rank == ROOT) && lfunc_data.verbose)
	printf("  Obtaining overlap integrals ...\n");
      get_integrals(&integral_data, freqn2, u2, v2, w2, nv2,
		    lfunc_data.ra, lfunc_data.dec, 
		    lfunc_data.ra_ref, lfunc_data.dec_ref, 
		    lfunc_data.freq, lfunc_data.antwdth, 
		    lfunc_data.sigma, lfunc_data.nf,
		    lfunc_data.rbins, lfunc_data.na, lfunc_data.ita, 
		    field, lfunc_data.ni, 1, lfunc_data.imagflag, 
		    lfunc_data.saveflag, 
		    "integrals", lfunc_data.rank, lfunc_data.npe);
      if ((lfunc_data.rank == ROOT) && lfunc_data.verbose) { 
	char msg[255];
	
	sprintf(msg, "Finished %i integrals for %i visibilities.", 
		integral_data.nz, integral_data.nv);
      }

      /*
       * create covariance matrix
       */
      cnz = countcms(&integral_data);
      if ((lfunc_data.rank == ROOT) && lfunc_data.verbose) {
	printf("  Number of entries in covariance matrix: %i\n", cnz);
	printf("  Creating covariance matrix...\n");
      }
      newcms(&cm_data, nv2, cnz);
      if ((lfunc_data.rank == ROOT) && lfunc_data.verbose)
	printf("  Creating index table for covariance matrix...\n");
      makeidxcms(&cm_data, &integral_data);
      
      ll = loglike(cps, gps, lfunc_data.na, sp, lfunc_data.sparseflag,
		   vis2, rms2, nv2,
		   lfunc_data.freq, lfunc_data.nf,
		   lfunc_data.freq0, lfunc_data.convf,
		   &chisq, &lndet, &integral_data, &cm_data,
		   lfunc_data.rank, lfunc_data.npe);
      likeval += ll;
      chisqval += chisq;
      lndetval += lndet;
      freeidxcms(&cm_data);
      freecms(&cm_data);
      if (!(lfunc_data.saveflag & INT_STORE_MEMORY))
	freeintegrals(&integral_data);
      free_1imatrix(freqn2); 
      free_1dmatrix(u2); free_1dmatrix(v2); free_1dmatrix(w2);
      free_1dmatrix(vis2); free_1dmatrix(rms2);
      /* TODO: free integrals */
    }
  }
  value = -likeval - lfunc_data.fnorm; /* find minimum of scaled -log(likelihood) */
  
  timeout = clock();
  if ((lfunc_data.rank == ROOT) && lfunc_data.verbose) {
    printf("  lndet,chisq,loglike = %16.9f %16.9f %16.9f\n",
	   0.5 * lndetval, 0.5 * chisqval, likeval);
    printf("  func = %g\n", value);
    sprintf(msg, "  lndet,chisq,loglike,func = %18.9f %18.9f %18.9f %18.9f\n",
	    0.5 * lndetval, 0.5 * chisqval, likeval, value);
    write_file("func.log", msg);
    
    printf("  CPU time required (secs) = %li\n", 
	   ((timeout - timein)/CLOCKS_PER_SEC));
  }

  free_1dmatrix(cps);
  free_1dmatrix(gps);

  return value;
}
