/* 
 * routines for calculating the overlap integrals between uv-points
 * for mosaiced VSA observations.
 */
/*  Time-stamp: <4 Nov 2002 14:39 GMT like_integrals.c maising> */
/*  
 *  HISTORY
 *    ??-??-1997  MPH   Fortran version for a single field (vsa_indfuncsubs.f)
 *    31-05-2001  KSM   C version, dynamical memory allocation, arbitrary
 *                      beam shapes
 *    09-06-2001  KSM   mosaicing works
 *    24-11-2001  KSM   changes to reduce memory footprint, enable MPI
 *                      and include w-corrections
 */
/* the routines assume a Gaussian primary beam for a quick 
 * calculation of the angular integrals using modified Bessel functions.  
 * routines for arbitrary beam shapes are provided,
 * using (slow) numerical 2d integration.
 */

/* the compressed column storage scheme used for the sparse matrices works 
 * as follows:
 *  o  an NxN matrix is compressed to a vector containing 
 *     its M non-zero elements.
 *  o  icn is an integer array of dimension N+1.
 *  o  icn[j] is the index of the first element of the M-dimensional 
 *     vector that is in the j-th column. 
 *     thus the index of the last element of the j-th column is given by
 *     icn[j+1]-1. hence icn[N+1] = M+1.
 *  o  irn is an integer array of dimension M.
 *  o  irn[p] is the row of the p-th element of the M-dimensional vector.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "globals.h"
#include "my_malloc2.h"
#include "integration.h"
#include "bessi0.h"
#include "like_integrals.h"

#if defined(ZLIB)
#include "zlib.h"
#endif
#if defined(FITSIO)
#include "fitsio.h"
#endif
#if defined(MPI)
#include "mpi.h"
#endif

#define PI 3.14159265358979323844
#define EPS 5.e-5		/* rel conv tol for radial integration */
#define NITER 12		/* max no of iterations for radial int. */
#define EPS2 1.e-6		/* rel conv tol for angular int. */
#define NITER2 12		/* max no of iterations for angular int. */
#define ABS_THRESH 1.e-9
#define MAX(x, y) (((x) > (y) ? (x) : (y)))
#define MIN(x, y) (((x) < (y) ? (x) : (y)))
#define CUTOFF 1
#define ROOT 0

#if ! defined(ZLIB)

#define FOPEN fopen
#define FCLOSE fclose
#define FPRINTF fprintf
#define FGETS fgets

#else

#define FOPEN gzopen
#define FCLOSE gzclose
#define FPRINTF gzprintf
#define FGETS(buf, len, stream) gzgets(stream, buf, len)

#endif

/*-----*-----------------------------------------------------------------*/

/* 
 * compute the modulus of a complex number z = x+ iy 
 * minimising round-off errors - compare NR.  
 * PARAMS 
 *     xi [I]   real part of z
 *     yi [I]   imaginary part of z
 * RETURNS
 *     modulus of z
 */
double modulus(double xi, double yi)
{
  double x, y, ans, temp;
  x = fabs(xi);
  y = fabs(yi);
  if (x == 0.)
    ans = y;
  else if (y == 0.)
    ans = x;
  else if (x > y) {
    temp = y / x;
    ans = x * sqrt(1. + temp*temp);
  } else {
    temp = x / y;
    ans = y * sqrt(1. + temp*temp);
  }
  return ans;
}

/*
 * compute the square root of a complex number z = x + iy and
 * return zo = xo + i yo. compare NR. 
 * PARAMS
 *     xi [I]   real part of z
 *     yi [I]   imaginary part of z
 *     xo [O]   real part of sqrt(z)
 *     yo [O]   imaginary part of sqrt(z)
 */
void cdsqrt(double xi, double yi, double *xo, double *yo)
{
  double x, y, w, r;

  if ((xi == 0.) && (yi == 0.)) {
    *xo = 0.; *yo = 0.;
    return;
  } else {
    x = fabs(xi);
    y = fabs(yi);
    if (x >= y) {
      r = y / x;
      w = sqrt(x) * sqrt(0.5 * (1.+ sqrt(1. + r*r)));
    } else {
      r = x / y;
      w = sqrt(y) * sqrt(0.5 * (r + sqrt(1. + r*r)));
    }
    if (xi >= 0.0) {
      *xo = w;
      *yo = yi / (2. * w);
    } else {
      *yo =(yi >= 0) ? w : -w;
      *xo = yi /(2. * (*yo));
    }
    return;
  }
}

/*-----*-----------------------------------------------------------------*/
/*-----*-----------------------------------------------------------------*/
/* 
 * allocate memory for sparse representation of integrals required to
 * calculate the covariance matrix. see the top of this file for more
 * explanations on the storage scheme.
 * PARAMS
 *     data   [O]  struct for the overlap integrals
 *     nv     [I]  number of visibilities
 *     nz     [I]  number of non-zero integrals
 *     n_bins [I]  number of bins in which the overlap integrals are
 *                 calculated
 */
void newintegrals(integral_struct *data, int nv, int nz, int n_bins)
{
  int i;

  data->nv = nv;
  data->nz = nz;
  data->j1 = 1;
  data->j2 = nv;
  /* number of bins has to be an odd number = 
   *     the central bin + same number of bins on both sides */
  if (! (n_bins % 2))
    n_bins++;
  data->n_bins = n_bins;
  data->icn = malloc_1imatrix(nv+1);
  data->irn = malloc_1imatrix(nz);
  data->freqn = malloc_1imatrix(nv);
  data->maxbin1 = malloc_1imatrix(nz);
  data->maxbin2 = malloc_1imatrix(nz);
  for (i = 1; i <= n_bins; i++) {
    data->integral1r[i] = malloc_1fmatrix(nz);
    data->integral2r[i] = malloc_1fmatrix(nz);
    data->integral1i[i] = malloc_1fmatrix(nz);
    data->integral2i[i] = malloc_1fmatrix(nz);
  }
  return;
}

/* 
 * free memory allocated for integrals 
 * PARAMS
 *     data   [I,O]  struct for the overlap integrals
 */
void freeintegrals(integral_struct *data)
{
  int i;

  free_1imatrix(data->irn); free_1imatrix(data->icn);
  free_1imatrix(data->freqn);
  free_1imatrix(data->maxbin1); 
  free_1imatrix(data->maxbin2);
  for (i = 1; i <= data->n_bins; i++) {
    free_1fmatrix(data->integral1r[i]);
    free_1fmatrix(data->integral2r[i]);
    free_1fmatrix(data->integral1i[i]);
    free_1fmatrix(data->integral2i[i]);
  }

  return;
}

/* 
 * returns the number of integrals necessary in a sparse representation
 * of the integrals required to derive the covariance matrix.
 * PARAMS
 *     antwdth   [I]  width of 1/e diameter of aperture in wavelengths 
 *                    for each observed fields
 *     nf        [I]  number of observed fields
 *     u,v       [I]  uv positions of observed visibilities
 *     nv        [I]  number of calculated visibilities
 */
int countintegrals(double *antwdth, int nf, 
		   int *freqn, double *u, double *v, int nv)
{
  int nz;
  int i, j, ifr, jfr;
  double amax, sep1, sep2, anti, antj, ui, uj, vi, vj;

  nz = 0;
  for(j = 1; j <= nv; j++) {
    jfr = freqn[j];
    antj = antwdth[jfr];
    uj   = u[j];
    vj   = v[j];
    for (i = j; i <= nv; i++) { /* only calculate elements in lower triangle*/
      ifr = freqn[i];
      anti = antwdth[ifr];
      ui   = u[i];
      vi   = v[i];

      amax = MAX(anti, antj);
      sep1 = sqrt((ui-uj)*(ui-uj) + (vi-vj)*(vi-vj));
      sep2 = sqrt((ui+uj)*(ui+uj) + (vi+vj)*(vi+vj));
#if CUTOFF
      if ((sep1 > 2.*amax) && (sep2 > 2.*amax)) continue;
#endif
      nz++;
    }
  }
  return nz;
}

/*-----*-----------------------------------------------------------------*/
/*-----*-----------------------------------------------------------------*/

/************************************************************************/
/* the next routines calculate the real part of the integrals using 
 * code converted from MPH Fortran version.
 * assumes a Gaussian beam shape and does not work for mosaicing, 
 * but serves as a good test case.
 */
/************************************************************************/

/* 
 * logarithm of the modified Bessel function of the first kind and zeroth 
 * order for real argument x.
 * compare NR, Abramowitz & Stegun 
 */
double logbessi0(double x)
{
  double ax, bessi0, y;
  const double p[7+1] = { 0., 1.0, 3.5156229, 3.0899424,
			  1.2067492, 0.2659732, 0.360768e-1, 0.45813e-2 }; 
  const double q[9+1] = { 0., 0.39894228, 0.1328592e-1, 0.225319e-2, 
			  -0.157565e-2, 0.916281e-2, -0.2057706e-1, 
			  0.2635537e-1, -0.1647633e-1, 0.392377e-2 };

  if (fabs(x) < 3.75) {
    y = (x/3.75) * (x/3.75);
    bessi0 = p[1] 
      + y * (p[2]+y*(p[3]+y*(p[4]+y *(p[5]+y*(p[6]+y*p[7])))));
    return log(bessi0);
  } else {
    ax = fabs(x);
    y = 3.75/ax;
    return ax - 0.5 * log(ax) 
      + log(q[1]+y*(q[2]+y*(q[3]+y*(q[4]+y* 
		 (q[5]+y*(q[6]+y*(q[7]+y*(q[8]+y*q[9]))))))));
  }
}

/*-----*-----------------------------------------------------------------*/

struct com_struct {
  double ar1, ar2, ar3, b;
  double fmax;
};

static struct com_struct fcncom; 

/* calculates the integrand fcn*/
double fcn(double x)
{
  double logf;
  
  logf = log(fcncom.b) - fcncom.ar1 - fcncom.ar2 * x*x 
    + logbessi0(fcncom.ar3 * x) - log(x);

  return exp(logf);
}

/* calculates fcn divided by the maximum fmax */
double normfcn(double x)
{
  double logf;

  logf = log(fcncom.b) - fcncom.ar1 - fcncom.ar2 * x*x 
    + logbessi0(fcncom.ar3 *x ) - log(x) - log(fcncom.fmax);

  return exp(logf);
}

/*-----*-----------------------------------------------------------------*/
/*-----*-----------------------------------------------------------------*/

/************************************************************************/
/* routines to calculate integrals for a Gaussian primary beam.
 * uses complex modfied Bessel functions.
 * compare integral 3.937 of Gradshteyn & Rhyzik - note that the fourth
 * edition has the sign in front of the expression D wrong.
 * compare also Gr"obner & Hofreiter, Integraltafel, integral 337.9     */ 
/************************************************************************/
struct newfcn_struct {
  double ar, b1r, b2r;
#if defined(WCORRECTIONS)
  double ai, b1i, b2i;
#endif
  double cp, dp;
};

static struct newfcn_struct newfcn_data;

/* 
 * calculate parameters that are constant for two given visibilities i and j
 */
void set_newfcn(double ui, double vi, double wi,
		double xi, double yi, double si,
		double uj, double vj, double wj, 
		double xj, double yj, double sj)
{
  double ui2, vi2, si2, uj2, vj2, sj2;
  double x1, x2, w1, w2, c, d, sc, sd;
#if defined(WCORRECTIONS)
  double pi, pj;
#endif

  ui2 = ui * ui; vi2 = vi * vi;
  uj2 = uj * uj; vj2 = vj * vj;
  si2 = si * si; 
  sj2 = sj * sj; 

#if defined(WCORRECTIONS)
  pi = 2*PI * si2 * wi;
  pj = 2*PI * sj2 * wj;
  si2 /= (1. + pi * pi);
  sj2 /= (1. + pj * pj);
  newfcn_data.ar  = (4.*PI*PI*si2*sj2) * (1. + pi*pj);
  newfcn_data.ai  = (4.*PI*PI*si2*sj2) * (pj - pi);
  newfcn_data.b1r = -2.*PI*PI*(si2*(ui2+vi2) + sj2*(uj2+vj2));
  newfcn_data.b1i = -2.*PI*PI*(pj*sj2*(uj2+vj2) - pi*si2*(ui2+vi2));
  newfcn_data.b2r = -2.*PI*PI*(si2+sj2);
  newfcn_data.b2i = -2.*PI*PI*(pj*sj2-pi*si2);
  x1 = xj - xi + 2*PI * (pj*sj2*uj - pi*si2*ui); 
  x2 = yj - yi + 2*PI * (pj*sj2*vj - pi*si2*vi);
#else
  newfcn_data.ar  = (4.*PI*PI*si2*sj2);
  newfcn_data.b1r = -2.*PI*PI*(si2*(ui2+vi2) + sj2*(uj2+vj2));
  newfcn_data.b2r = -2.*PI*PI*(si2+sj2);
  x1 = xj - xi; 
  x2 = yj - yi;
#endif

  w1 = si2*ui+sj2*uj; 
  w2 = si2*vi+sj2*vj; 

  c = 4. * PI*PI * (w1*w1+w2*w2) - (x1*x1+x2*x2);
  d = 4. * PI * (w1*x1+w2*x2);
  
  cdsqrt(c, d, &sc, &sd);
  newfcn_data.cp = sc * 2. * PI;
  newfcn_data.dp = sd * 2. * PI;

  return;
}

/* 
 * compute the real and imaginary parts of the angular integral as a 
 * a function of radius 
 */
dcomplex newfcn(double rho)
{
  double rintp, iintp;
  double c, d, arg;
#if defined(WCORRECTIONS)
  double br, bi, kr, ki;
#endif
  dcomplex result;

  arg = newfcn_data.b1r + newfcn_data.b2r * rho * rho;
  c = newfcn_data.cp * rho; d = newfcn_data.dp * rho;
  CDEFACBESSI0(&arg, &c, &d, &rintp, &iintp);

#if defined(WCORRECTIONS)
  arg = newfcn_data.b1i + newfcn_data.b2i * rho * rho;
  br = 2.*PI * cos(arg) / rho;
  bi = 2.*PI * sin(arg) / rho; 

  kr = br * rintp - bi * iintp;
  ki = br * iintp + bi * rintp;
  result.re = newfcn_data.ar * kr - newfcn_data.ai * ki;
  result.im = newfcn_data.ar * ki + newfcn_data.ai * kr;
#else
  result.re = newfcn_data.ar * 2. * PI * rintp / rho;
  result.im = newfcn_data.ar * 2. * PI * iintp / rho;
#endif

  return result;
}

/*-----*-----------------------------------------------------------------*/
/*-----*-----------------------------------------------------------------*/

/************************************************************************/
/* routines to define a primary beam of arbitrary shape.                */
/************************************************************************/
struct beam_struct {
  dcomplex (*beamfft)(double, double, double);
};

static struct beam_struct beam_data;

/* 
 * set an arbitray aperture function. it is a function of 3 arguments:
 * u, v, and frequency
 */
void set_beamfft(dcomplex (*beamfft)(double, double, double))
{
  beam_data.beamfft = beamfft;
  return;
}

/* 
 * return real part of beam 
 */
double rbeamfft(double u, double v, double f)
{
  dcomplex r;

  r = beam_data.beamfft(u, v, f);

  return r.re;
}

/************************************************************************/
/* routines to compute integrals for arbitrary beam shapes,
 * using 2d numerical integration 
 * NOTE: w-corrections are not implemented
 */
/************************************************************************/

struct gnewfcn_struct {
  double rho;
  double ui, vi, uj, vj, fi, fj;
  double x1, x2;
};

static struct gnewfcn_struct gnewfcn_data;

/* 
 * initialise constant variables for subsequent calls of g1newfcn() 
 * and g2newfcn()
 */
void set_gnewfcn(double ui, double vi, double wi, 
		 double xi, double yi, double fi,
		 double uj, double vj, double wj,
		 double xj, double yj, double fj)
{
#if defined(WCORRECTIONS)
  myerror("FATAL ERROR: w correction not implemented "
	  "for arbitrary beam shapes");
#endif
  gnewfcn_data.ui = ui; gnewfcn_data.vi = vi; gnewfcn_data.fi = fi; 
  gnewfcn_data.uj = uj; gnewfcn_data.vj = vj; gnewfcn_data.fj = fj; 
  gnewfcn_data.x1 = 2. * PI * (xj - xi);
  gnewfcn_data.x2 = 2. * PI * (yj - yi);
  gnewfcn_data.rho = 0.;

  return;
}

/* calculate integrals in <V V^*> for arbitrary beam shapes.
 * this function cannot be called directly - it is used as an 
 * integration routine in g1newfcn()
 */
dcomplex g1newfcn1d(double theta)
{
  double   u1, u2, arg;
  dcomplex a, b, result;
  double   cr, ci, arbr, arbi, aibr, aibi;

  u1 = gnewfcn_data.rho * cos(theta);
  u2 = gnewfcn_data.rho * sin(theta);

  a = beam_data.beamfft(gnewfcn_data.ui - u1, gnewfcn_data.vi - u2,
			gnewfcn_data.fi);
  b = beam_data.beamfft(gnewfcn_data.uj - u1, gnewfcn_data.vj - u2,
			gnewfcn_data.fj);
  b.im = - b.im;		/* complex conjugate */

  arg = gnewfcn_data.x1 * u1 + gnewfcn_data.x2 * u2;
  cr = cos(arg); ci = sin(arg);

  arbr = a.re * b.re; arbi = a.re * b.im;
  aibr = a.im * b.re; aibi = a.im * b.im;

  result.re = (arbr - aibi) * cr - (arbi + aibr) * ci;
  result.im = (arbr - aibi) * ci + (arbi + aibr) * cr;
  
  return result;
}

/* calculate integrals in <V V^*> for arbitrary beam shapes.
 * set_gnewfcn() and set_beamfft() need to be called first 
 * for initialisations.
 */
dcomplex g1newfcn(double rho)
{
  dcomplex result;
  double   eps  = EPS2;
  int      iter = NITER2;

  gnewfcn_data.rho = rho;
  result = cqsimp2(g1newfcn1d, 0., 2.*PI, eps, &iter);
#if 0
  if (iter == 0)
    printf("Warning: Inaccurate theta integration in g1newfcn.\n");
#endif
  result.re /= rho; result.im /= rho;
  return result;
}


/* calculate integrals in <V V> for arbitrary beam shapes.
 * this function cannot be called directly - it is used as an 
 * integration routine in g2newfcn()
 */
dcomplex g2newfcn1d(double theta)
{
  double   u1, u2, arg;
  dcomplex a, b, result;
  double   cr, ci, arbr, arbi, aibr, aibi;

  u1 = gnewfcn_data.rho * cos(theta);
  u2 = gnewfcn_data.rho * sin(theta);

  a = beam_data.beamfft(gnewfcn_data.ui - u1, gnewfcn_data.vi - u2,
			gnewfcn_data.fi);
  b = beam_data.beamfft(gnewfcn_data.uj + u1, gnewfcn_data.vj + u2,
			gnewfcn_data.fj);

  arg = gnewfcn_data.x1 * u1 + gnewfcn_data.x2 * u2;
  cr = cos(arg); ci = sin(arg);

  arbr = a.re * b.re; arbi = a.re * b.im;
  aibr = a.im * b.re; aibi = a.im * b.im;

  result.re = (arbr - aibi) * cr - (arbi + aibr) * ci;
  result.im = (arbr - aibi) * ci + (arbi + aibr) * cr;
  
  return result;
}

/* calculate integrals in <V V> for arbitrary beam shapes.
 * set_gnewfcn() and set_beamfft() need to be called first 
 * for initialisations.
 */
dcomplex g2newfcn(double rho)
{
  dcomplex result;
  double   eps  = EPS2;
  int      iter = NITER2;

  gnewfcn_data.rho = rho;
  result = cqsimp2(g2newfcn1d, 0., 2.*PI, eps, &iter);
#if 0
  if (iter == 0)
    printf("Warning: Inaccurate theta integration in g2newfcn.\n");
#endif
  result.re /= rho; result.im /= rho;
  return result;
}

/*-----*-----------------------------------------------------------------*/


/* 
 * find the maximum of a function and determine the bin where it 
 * occurs
 * PARAMS
 *     fcn    [I]  real function whose maximum is to be found
 *     rbins  [I]  borders of bins, the i-th bin is [rbins[i], rbins[i+1]]
 *     na     [I]  number of bins
 *     maxbin [O]  bin where maximum occurs
 *     fmax   [O]  maximum of fcn
 *      
 */
void maxfcn(double (*fcn)(double), double *ra, int na, 
	    int *maxbin, double *fmax)
{
  int npts = 10;
  int i, k;
  double  x, xstep, xlo, xhi, f;

  *fmax = fabs((*fcn)(ra[1]));
  *maxbin = 1;
  for(k = 1; k <= na; k++) {
    xlo = ra[k];
    xhi = ra[k+1];
    xstep = (xhi - xlo) / ((double)(npts-1));
    for(i = 1; i <= npts; i++) {
      x = xlo + (double)(i-1) * xstep;
      f = fabs((*fcn)(x));
      if (f > *fmax) {
	*fmax = f;
	*maxbin = k;
      } 
    }
  } 
  
  return;
}

/* 
 * find the maximum of the modulus of a complex function and 
 * determine the bin where it occurs
 * PARAMS
 *     fcn    [I]  complex function whose maximum is to be found
 *     rbins  [I]  borders of bins, the i-th bin is [rbins[i], rbins[i+1]]
 *     na     [I]  number of bins
 *     maxbin [O]  bin where maximum occurs
 *     fmax   [O]  maximal modulus of fcn
 *      
 */
void cmaxfcn(dcomplex (*fcn)(double), double *ra, int na, 
	    int *maxbin, double *fmax)
{
  int npts = 10;
  int i, k, klo, khi;
  double  x, xstep, xlo, xhi, f;
  dcomplex fx;
  
  /* find the bins at whose edges the function is maximal */
  fx = (*fcn)(ra[1]);
  *fmax = modulus(fx.re, fx.im);
  *maxbin = 1;
  for(k = 1; k <= na; k++) {
    fx = (*fcn)(ra[k+1]);
    f = modulus(fx.re, fx.im);
    if (f > *fmax) {
      *fmax = f;
      *maxbin = k;
    } 
  }
  klo = (*maxbin - 1 >= 1 ? *maxbin - 1 : 1);
  khi = (*maxbin + 2 <= na ? *maxbin + 2 : na);

  /* search the bins around the maximum more thoroughly */
  for(k = klo; k <= khi; k++) {
    xlo = ra[k];
    xhi = ra[k+1];
    xstep = (xhi - xlo) / ((double)(npts-1));
    for(i = 1; i <= npts; i++) {
      x = xlo + (double)(i-1) * xstep;
      fx = (*fcn)(x);
      f = modulus(fx.re, fx.im);
      if (f > *fmax) {
	*fmax = f;
	*maxbin = k;
      } 
    }
  } 
  
  return;
}


/* This function determines the integral K_ij (x_m, x_{m+1}) 
 * in the the bins around the maximum of the integrand.
 * The integrals are given by 
 *   K_{ij} = \int_{x_m}^{x_{m+1}} dx I(x), 
 * where the integrand
 *   I(x) = (2\pi)^3 s_i^2 s_j^2 e^{-2\pi^2 (s_i^2 |u_i|^2 + s_j^2 |u_j|^2)}
 *          x^{-1} e^{-2\pi^2 (s_i^2+s_j^2) x^2} 
 *          I_0 (4 \pi | s_i^2 u_i + s_j^2 u_j| x),
 *        = e^{[ \ln(b) - a_1 - a_2 x^2 - \ln x + \ln I_0 (a_3)]},
 * where 
 *   a_1 = 2 \pi^2 (s_i^2 |u_i|^2 + s_j^2 |u_j|^2),
 *   a_2 = 2 \pi^2 (s_i^2+s_j^2),
 *   a_3 = 4 \pi  | s_i^2 u_i + s_j^2 u_j|,
 *   b   = (2 \pi)^3 s_i^2 s_j^2,
 * and I_0 modified Bessel function.
 * The values in the bins are returned in the variable temp.
 *
 * HISTORY
 *   31-05-2001 KSM, based on MPH Fortran code 
 *
 *****************************************************************
 * NOTE: This calculates real parts only, assuming no mosaicing.
 *       imaginary parts are set to zero.
 *****************************************************************
 * 
 * PARAMS
 *     ui, vi     [I]  uv position of first visibility
 *     xi, yi     [I]  offset of field centre corresponding 
 *                     to the first visibility in rad
 *     si         [I]  width of Gaussian beam for field corresponding to first
 *                     visibility
 *     uj, ... sj [I]  the same parameters for the second visibility
 *     rbins      [I]  borders of bins, the i-th bin is [rbins[i], rbins[i+1]]
 *     na         [I]  total number of bins
 *     tempr, tempi [O]  real and imaginary parts of overlap integrals 
 *                       for bins around maximum
 *     n_bins     [I]  number of bins calculated around maximum
 *     kmax       [O]  bin in which maxima for real and imaginary parts lie
 *     imagflag   [I]  flag: use complex calculations? (here ignored)
 *     flag       [I]  flag: calculate <V_i V_j^*> or <V_i V_j>?
 */
void calcints(double ui, double vi, double xi, double yi, double si, 
	      double uj, double vj, double xj, double yj, double sj,
	      double *rbins, int na, double *tempr, double *tempi, int n_bins,
	      int *kmax, int imagflag, int flag)
{
  int      k, klo, khi, kwdth, m, iter;
  double   value, eps;
  double   ai, aj, au, av, si2, sj2;
  
  /* for a real beam, the expression for <V_i V_j> 
     can be derived from  <V_i V_j^*> by replacing u_j -> -u_j */
  if (flag != 1) {
    uj = -uj; vj = -vj;
  }

  eps = EPS;             /*fractional accuracy in integration */

  /* determine the constants in the integrand */
  si2 = si * si;
  sj2 = sj * sj;
  ai = 2.*PI*PI * si2;
  aj = 2.*PI*PI * sj2;

  fcncom.ar1 = ai * (ui*ui+vi*vi) + aj*(uj*uj+vj*vj);
  fcncom.ar2 = ai + aj;
  au = ai*ui+aj*uj;
  av = ai*vi+aj*vj;
  fcncom.ar3 = 2.*sqrt(au*au+av*av);
  fcncom.b = 8.*PI*PI*PI*si2*sj2;
  
  /* determine the maximum of the integrand and the bin where it occurs */
  maxfcn(fcn, rbins, na, kmax, &fcncom.fmax);

  /* integrate over bins adjacent to the one with the maximum.
   * if the adjacent bin lies outside, set integral to zero.
   * the resulting integrals are returned in the variable temp. */
  kwdth = n_bins/2;
  klo = *kmax - kwdth;
  khi = *kmax + kwdth;
  if (klo < 1) {
    for (k = 1; k <= 1 - klo; k++) {
      tempr[k] = 0.; tempi[k] = 0.;
    }
    klo = 1;
  }
  if (khi > na) {
    for (k = n_bins - (khi - na) + 1; k <= n_bins; k++) {
      tempr[k] = 0.; tempi[k] = 0.;
    }
    khi = na;
  }

  /* numerical radial integration */
  for (k = klo; k <= khi; k++) {
    m = k - *kmax + kwdth + 1;
    iter = NITER;
    value = qsimp2(normfcn, rbins[k], rbins[k+1], eps, &iter);
    if ((iter == 0) && (fabs(value) > ABS_THRESH)) 
      printf("WARNING: inaccurate int. in calcints (bin %2i (%2i), %g)\n",
	     k, m-kwdth-1, fabs(value));
    tempr[m] = fcncom.fmax * value; /* multiply by fmax */
    tempi[m] = 0.; /* no imaginary parts for real integration */
  }

  return;
}

/* 
 * calculates an overlap integral using complex modified Bessel functions 
 * and complex integration routines. 
 * PARAMS
 *     ui, vi, wi [I]  uvw position of first visibility
 *     xi, yi     [I]  offset of field centre corresponding 
 *                     to the first visibility in rad
 *     si         [I]  width of Gaussian beam for field corresponding to first
 *                     visibility
 *     uj, ... sj [I]  the same parameters for the second visibility
 *     rbins      [I]  borders of bins, the i-th bin is [rbins[i], rbins[i+1]]
 *     na         [I]  total number of bins
 *     tempr, tempi [O]  real and imaginary parts of overlap integrals 
 *                       for bins around maximum
 *     n_bins     [I]  number of bins calculated around maximum
 *     kmax       [O]  bin in which maxima for real and imaginary parts lie
 *     imagflag   [I]  flag: use complex calculations? 
 *     flag       [I]  flag: calculate <V_i V_j^*> or <V_i V_j>?
 */
void calcints6(double ui, double vi, double wi, 
	       double xi, double yi, double si, 
	       double uj, double vj, double wj,
	       double xj, double yj, double sj,
	       double *rbins, int na, double *tempr, double *tempi, int n_bins,
	       int *kmax, int imagflag, int flag)
{
  int      k, klo, khi, kwdth, m, iter;
  double   eps, fmax;
  dcomplex value;

  /* for a real beam, the expression for <V_i V_j> 
     can be derived from  <V_i V_j^*> by replacing u_j -> -u_j */
  if (flag != 1) {
    uj = -uj; vj = -vj; wj = -wj;
  }

  eps = EPS;             /*fractional accuracy in integration */

  set_newfcn(ui, vi, wi, xi, yi, si, uj, vj, wj, xj, yj, sj); 

  /* determine the maximum of the integrand and the bin where it occurs */
  cmaxfcn(newfcn, rbins, na, kmax, &fmax);

  /* integrate over the bins adjacent to the one with the maximum.
   * if the adjacent bin lies outside, set integral to zero.
   * the resulting integrals are returned in the variable temp. */
  kwdth = n_bins/2;
  klo = *kmax - kwdth;
  khi = *kmax + kwdth;
  if (klo < 1) {
    for (k = 1; k <= 1 - klo; k++) {
      tempr[k] = 0.; tempi[k] = 0.;
    }
    klo = 1;
  }
  if (khi > na) {
    for (k = n_bins - (khi - na) + 1; k <= n_bins; k++) {
      tempr[k] = 0.; tempi[k] = 0.;
    }
    khi = na;
  }

  /* numerical radial integration */
  for (k = klo; k <= khi; k++) {
    m = k - *kmax + kwdth + 1;
    iter = NITER;	      /* number of steps in integration routine */
    value = cqsimp2(newfcn, rbins[k], rbins[k+1], eps, &iter);
    if ((iter == 0) && (modulus(value.re, value.im) > ABS_THRESH)) 
      printf("WARNING: inaccurate int. in calcints6 (bin %2i (%2i), %g)\n",
	      k, m-kwdth-1, modulus(value.re, value.im));
    tempr[m] = value.re;
    tempi[m] = value.im;
  }

  return;
}

/* 
 * calculates an overlap integral using complex integration routines. 
 * unlike calcints6, it can do arbitrary beam shapes. however, it is slower
 * than calcints6.
 * PARAMS
 *     ui, vi, wi [I]  uvw position of first visibility
 *     xi, yi     [I]  offset of field centre corresponding 
 *                     to the first visibility in rad
 *     si         [I]  width of Gaussian beam for field corresponding to first
 *                     visibility
 *     uj, ... sj [I]  the same parameters for the second visibility
 *     rbins      [I]  borders of bins, the i-th bin is [rbins[i], rbins[i+1]]
 *     na         [I]  total number of bins
 *     tempr, tempi [O]  real and imaginary parts of overlap integrals 
 *                       for bins around maximum
 *     n_bins     [I]  number of bins calculated around maximum
 *     kmax       [O]  bin in which maxima for real and imaginary parts lie
 *     imagflag   [I]  flag: use complex calculations? 
 *     flag       [I]  flag: calculate <V_i V_j^*> or <V_i V_j>?
 */
void calcints7(double ui, double vi, double wi,
	       double xi, double yi, double fi, 
	       double uj, double vj, double wj,
	       double xj, double yj, double fj,
	       double *rbins, int na, double *tempr, double *tempi, int n_bins,
	       int *kmax, int imagflag, int flag)
{
  int      k, klo, khi, kwdth, m, iter;
  double   eps, fmax;
  dcomplex value;

  eps = EPS;             /*fractional accuracy in integration */

  set_gnewfcn(ui, vi, wi, xi, yi, fi, uj, vj, wj, xj, yj, fj); 

  /* determine the maximum of the integrand and the bin where it occurs */
  if (flag == 1)
    cmaxfcn(g1newfcn, rbins, na, kmax, &fmax);
  else 
    cmaxfcn(g2newfcn, rbins, na, kmax, &fmax);

  /* integrate over the bins adjacent to the one with the maximum.
   * if the adjacent bin lies outside, set integral to zero.
   * the resulting integrals are returned in the variable temp. */
  kwdth = n_bins/2;
  klo = *kmax - kwdth;
  khi = *kmax + kwdth;
  if (klo < 1) {
    for (k = 1; k <= 1 - klo; k++) {
      tempr[k] = 0.; tempi[k] = 0.;
    }
    klo = 1;
  }
  if (khi > na) {
    for (k = n_bins - (khi - na) + 1; k <= n_bins; k++) {
      tempr[k] = 0.; tempi[k] = 0.;
    }
    khi = na;
  }

  /* numerical radial integration */
  for (k = klo; k <= khi; k++) {
    m = k - *kmax + kwdth + 1;
    iter = NITER;		/* number of steps in integration routine */
    if (flag == 1)		/* < V V^* > */
      value = cqsimp2(g1newfcn, rbins[k], rbins[k+1], eps, &iter);
    else			/* < V V > */
      value = cqsimp2(g2newfcn, rbins[k], rbins[k+1], eps, &iter);
    if (iter == 0) 
      printf("WARNING: inaccurate int. in calcints7 (bin %2i (%2i), %g)\n",
	      k, m-kwdth-1, modulus(value.re, value.im));
    tempr[m] = value.re;
    tempi[m] = value.im;
  }

  return;
}

/*
 * print utility used by testints
 */
void print_t_line(char s[], double ui, double vi, double wi,
		  double uj, double vj, double wj,
		  double temp[], int n_bins, int kmax)
{
  int i;
  
  printf("*** %s  %3.0f %3.0f %3.0f %3.0f %3.0f %3.0f:", 
	 s, ui, vi, wi, uj, vj, wj);
  for (i = 1; i <= n_bins; i++) printf(" %g", temp[i]);
  printf(" %2i\n", kmax);
  
  return;
}

/* 
 * DEBUGGING routine to compare analytical and numerical integration
 * routines. 
 */
void testints(double ui, double vi, double wi, 
	      double xi, double yi, double fi, double si, 
	      double uj, double vj, double wj, 
	      double xj, double yj, double fj, double sj,
	      double *rbins, int na, int n_bins, int imagflag, int flag) 
{

  int    kmax, i;
  double tempr[N_BINS+1], tempi[N_BINS+1];

  if (! imagflag) {
    calcints (ui, vi, xi, yi, si, uj, vj, xj, yj, sj, rbins, na, 
	      tempr, tempi, n_bins, &kmax, imagflag, flag);
    print_t_line("Bess1d ", ui, vi, 0., uj, vj, 0., tempr, n_bins, kmax);
    print_t_line("Bess1d ", ui, vi, 0., uj, vj, 0., tempi, n_bins, kmax);
  }

  calcints6(ui, vi, wi, xi, yi, si, uj, vj, wj, xj, yj, sj, rbins, na, 
	    tempr, tempi, n_bins, &kmax, imagflag, flag);
  print_t_line("cBess2d", ui, vi, wi, uj, vj, wj, tempr, n_bins, kmax);
  print_t_line("cBess2d", ui, vi, wi, uj, vj, wj, tempi, n_bins, kmax);

#if ! defined(WCORRECTIONS)
  calcints7(ui, vi, wi, xi, yi, fi, uj, vj, wj, xj, yj, fj, rbins, na, 
	    tempr, tempi, n_bins, &kmax, imagflag, flag);
  print_t_line("gnumint", ui, vi, wi, uj, vj, wj, tempr, n_bins, kmax);
  print_t_line("gnumint", ui, vi, wi, uj, vj, wj, tempi, n_bins, kmax);
#endif
  printf("\n");

  for (i = 1; i <= 1; i++) {
    xj += 100.*0.000436;

#if 1
    printf("  x: (%g %g) (%g %g)\n", xi, yi, xj, yj); 

    calcints6(ui, vi, wi, xi, yi, si, uj, vj, wj, xj, yj, sj, rbins, na, 
	      tempr, tempi, n_bins, &kmax, 1, flag);
    print_t_line("cBess2d", ui, vi, wi, uj, vj, wj, tempr, n_bins, kmax);
    print_t_line("cBess2d", ui, vi, wi, uj, vj, wj, tempi, n_bins, kmax);

#if ! defined (WCORRECTIONS)
    calcints7(ui, vi, wi, xi, yi, fi, uj, vj, wj, xj, yj, fj, rbins, na, 
	      tempr, tempi, n_bins, &kmax, imagflag, flag);
    print_t_line("gnumint", ui, vi, wi, uj, vj, wj, tempr, n_bins, kmax);
    print_t_line("gnumint", ui, vi, wi, uj, vj, wj, tempi, n_bins, kmax);
#endif

#endif

  }
    
  return;
}

/*-----*-----------------------------------------------------------------*/

/*
 * return the projection of the point given by ra, dec onto the plane 
 * tangent to (ra_ref, dec_ref).
 */
void sky_coords(double *x, double *y, double ra, double dec, 
		double ra_ref, double dec_ref)
{
  double alpha;

  alpha = ra - ra_ref;
  *x = sin(alpha) * cos(dec);
  *y = cos(dec_ref) * sin(dec) - sin(dec_ref) * cos(dec) * cos(alpha); 
  
  return;
}

/*
 * returns an array of column indices splitting a sparse covariance matrix 
 * into segments of roughly equal size; this is useful for MPI
 * PARAMS
 *     ints   [I]  struct containing the overlap integrals
 *     npe    [I]  number of processors used
 *     jidx   [O]  index array of matrix columns (dimension npe+1).
 *                 the processor of rank j calculates columns 
 *                 jidx[j+1] ... jidx[j+1]-1
 */ 
void get_mpi_integral_cols(int *icn, int nv, int nz, int npe, int *jidx)
{
  int i, j, p, size;

  /* determine the number of integrals each processor should roughly 
   * calculate
   */
  size = nz / npe;
  if (nz % npe)
    size++;

  /*
   * search through the column index to find the best division
   */
  jidx[1] = 1;
  for (i = 1; i <= npe; i++) {
    p = MIN(i*size, nz);
    j = jidx[i];
    while (icn[j+1]-1 < p)
      j++;
    jidx[i+1] = j+1;
  }
  jidx[npe+1] = nv+1;

  return;
}

#if 0
/*
 * experimental code, not used and not working properly
 */
void shrink_integrals(integral_struct *ints, int rank, int npe)
{
  int m, j, p, p0;
  int *jidx, newnz;
  int *newirn, *newicn, *newmaxbin1, *newmaxbin2;
  float *newintegral1r[N_BINS+1], *newintegral1i[N_BINS+1];
  float *newintegral2r[N_BINS+1], *newintegral2i[N_BINS+1];

  jidx = malloc_1imatrix(npe+1);
  get_mpi_integral_cols(ints->icn, ints->nv, ints->nz, npe, jidx);
  ints->j1 = jidx[rank+1];
  ints->j2 = jidx[rank+2]-1;

  /* allocate memory for reduced arrays */
  newnz = ints->icn[jidx[rank+2]] - ints->icn[jidx[rank+1]];
  newicn = malloc_1imatrix(ints->nv+1);
  newirn = malloc_1imatrix(newnz);
  newmaxbin1 = malloc_1imatrix(newnz);
  newmaxbin2 = malloc_1imatrix(newnz);
  for (m = 1; m <= ints->n_bins; m++) {
    newintegral1r[m] = malloc_1fmatrix(newnz);
    newintegral1i[m] = malloc_1fmatrix(newnz);
    newintegral2r[m] = malloc_1fmatrix(newnz);
    newintegral2i[m] = malloc_1fmatrix(newnz);
  }

  /* copy integrals to new array */
  p0 = ints->icn[ints->j1] - 1;
  for (j = 1; j <= ints->nv+1; j++) 
    newicn[j] = -1;
    
  for (j = ints->j1; j <= ints->j2; j++) {
    newicn[j] = ints->icn[j] - p0; 
	
    for (p = ints->icn[j]; p <= ints->icn[j+1]-1; p++) {
      newmaxbin1[p-p0] = ints->maxbin1[p];
      newmaxbin2[p-p0] = ints->maxbin2[p];
      newirn[p-p0] = ints->irn[p];
      for (m = 1; m <= ints->n_bins; m++) {
	newintegral1r[m][p-p0] = ints->integral1r[m][p]; 
	newintegral1i[m][p-p0] = ints->integral1i[m][p]; 
	newintegral2r[m][p-p0] = ints->integral2r[m][p]; 
	newintegral2i[m][p-p0] = ints->integral2i[m][p]; 
      }
    }
  }
  newicn[ints->j2+1] = ints->icn[ints->j2+1] - p0;
  
  /* free old memory */
  free_1imatrix(ints->icn); ints->icn = newicn;
  free_1imatrix(ints->irn); ints->irn = newirn;
  free_1imatrix(ints->maxbin1); ints->maxbin1 = newmaxbin1;
  free_1imatrix(ints->maxbin2); ints->maxbin2 = newmaxbin2;
  for (m = 1; m <= ints->n_bins; m++) {
    free_1fmatrix(ints->integral1r[m]); 
    ints->integral1r[m] = newintegral1r[m];
    free_1fmatrix(ints->integral1i[m]); 
    ints->integral1i[m] = newintegral1i[m];
    free_1fmatrix(ints->integral2r[m]); 
    ints->integral2r[m] = newintegral2r[m];
    free_1fmatrix(ints->integral2i[m]); 
    ints->integral2i[m] = newintegral2i[m];
  }

  free_1imatrix(jidx);
  return;
}
#endif


/* 
 * calculate the overlap integrals needed to derive covariance matrices 
 * from flat band powers. 
 * PARAMS
 *     common [I,O]  struct containing the overlap integrals
 *     freqn    [I]  index array indicating which observing fields the
 *                   visibilities belong to
 *     u, v, w  [I]  uvw positions of visibilities  
 *     nv       [I]  number of calculated visibilities
 *     ra, dec  [I]  RA and DEC of each observed fields
 *     ra_fre, dec_ref [I] RA and DEC of global reference centre
 *     freq     [I]  observing frequency for each field
 *     antwdth  [I]  width of 1/e diameter of aperture in wavelengths 
 *                    for each observed fields
 *     sigma    [I]  width of Gaussian beams for each field
 *     nf       [I]  number of observed fields
 *     rbins    [I]  borders of bins, the i-th bin is [rbins[i], rbins[i+1]]
 *     na       [I]  total number of bins
 *     gaussian_beam [I] flag: use Gaussian beam approximation?
 *     imagflag [I]  flag: use complex calculations? 
 *     rank     [I]  rank of processor; if MPI is not used, set to 0
 *     npe      [I]  number of processors; if MPI is not used set to 1
 */
void makeintegrals(integral_struct *common,
		   int *freqn, double *u, double *v, double *w, int nv, 
		   double *ra, double *dec, double ra_ref, double dec_ref,  
		   double *freq, double *antwdth, double *sigma, int nf,
		   double *rbins, int na, int gaussian_beam,
		   int imagflag, int rank, int npe)
{
  int    nz, n_bins;
  double tempr[N_BINS+1], tempi[N_BINS+1];
  double xpos[NFMAX+1], ypos[NFMAX+1];


  int    i, j, m, p, ifr, jfr, kmax1, kmax2;
  int    *jidx, int_total, int_done, percent;
  double ui, uj, vi, vj, wi, wj, si, sj, fi, fj, xi, yi, xj, yj;
  double anti, antj, amax, sep1, sep2;

  n_bins = common->n_bins;
  common->imaginary_flag = imagflag;

  /* determine relative coordinates of pointing centres */
  for (i = 1; i <= nf; i++) 
    sky_coords(&xpos[i], &ypos[i], ra[i], dec[i], ra_ref, dec_ref);

  /* 
   * build up index structure for sparse storage 
   */
  nz = 0;
  for(j = 1; j <= nv; j++) {
    common->freqn[j] = freqn[j];
    common->icn[j] = nz+1;
    antj = antwdth[freqn[j]];
    uj   = u[j]; vj   = v[j];
    for (i = j; i <= nv; i++) { /* only calculate elements in lower triangle*/
      anti = antwdth[freqn[i]];
      ui   = u[i]; vi   = v[i];

      amax = MAX(anti, antj);
      sep1 = sqrt((ui-uj)*(ui-uj) + (vi-vj)*(vi-vj));
      sep2 = sqrt((ui+uj)*(ui+uj) + (vi+vj)*(vi+vj));

#if CUTOFF
      /* skip correlations if separation of uv points is larger 
	 than the antenna width */
      if ((sep1 > 2.*amax) && (sep2 > 2.*amax)) continue;
#endif
      nz++;
      if (nz > common->nz) {
	printf("ERROR: insufficient sparse storage at i,j = %i %i\n", i, j);
	exit(1);
      }
      common->irn[nz] = i;

      /* initialise maxbins to zero; this serves as a marker to calculate
	 them later on */
      common->maxbin1[nz] = 0;
      common->maxbin2[nz] = 0;

#if CUTOFF 
      /* if separation of visibilities in uv plane is larger than
	 aperture, set integrals to zero */
      if (sep1 > 2.*amax) {	/* for u_i - u_j */
	common->maxbin1[nz] = 1;
	for(m = 1; m <= n_bins; m++) {
	  common->integral1r[m][nz] = 0.;
	  common->integral1i[m][nz] = 0.;
	}
      } 
      if (sep2 > 2.*amax) {	/* for u_i + u_j */
	common->maxbin2[nz] = 1;
	for(m = 1; m <= n_bins; m++) {
	  common->integral2r[m][nz] = 0.;
	  common->integral2i[m][nz] = 0.;
	}
      }
#endif
    }
  }
  common->icn[nv+1] = nz+1;
  if (rank == ROOT)
    printf( "  Total number of non-zero matrix elements = %i\n", nz);


  /* 
   * determine columns this process needs to calculate 
   */
  jidx = malloc_1imatrix(npe+1);
  get_mpi_integral_cols(common->icn, common->nv, common->nz, npe, jidx);
  int_total = common->icn[jidx[rank+2]] - common->icn[jidx[rank+1]];
  printf("  Processor %3i: calculating %8i integrals in rows %5i-%5i\n",
	 rank, int_total, jidx[rank+1], jidx[rank+2]-1);

  /*
   * calculate the overlap integrals
   */
  int_done = 0;
  percent = 1;

  for(j = jidx[rank+1]; j <= jidx[rank+2]-1; j++) {
    jfr = freqn[j];
    fj = freq[jfr]; sj = sigma[jfr];
    xj = xpos[jfr]; yj = ypos[jfr];
    uj = u[j]; vj = v[j]; wj = w[j];

    for (p = common->icn[j]; p <= common->icn[j+1]-1; p++) {
      i = common->irn[p];
      ifr = freqn[i];
      fi = freq[ifr]; si = sigma[ifr];
      xi = xpos[ifr]; yi = ypos[ifr];
      ui = u[i]; vi = v[i]; wi = w[i];

      /* 
       *  perform the numerical integration for < V_i V_j^* > 
       *  the non-zero elements have maxbin set to 0 (see above)
       */
      if (common->maxbin1[p] == 0) {
#if 0
	printf("<V V*>\n");
	printf("testints(%g, %g, %g, %g, %g, %g, %g, %g, %g, %g, %g, %g, %g,"
	       " %g, %i, %i %i %i)\n\n", 
	       ui, vi, wi, xi, yi, fi, si, uj, vj, wj, xj, yj, fj, sj, 
	       na, n_bins,  imagflag, 1);
	testints(ui, vi, wi, xi, yi, fi, si, uj, vj, wj, xj, yj, fj, sj, 
		 rbins, na, n_bins, imagflag, 1);
#endif
	if (gaussian_beam) {

	  if (!imagflag)
	    calcints (ui, vi, xi, yi, si, uj, vj, xj, yj, sj, rbins, na, 
		      tempr, tempi, n_bins, &kmax1, imagflag, 1);
	  else
	    calcints6(ui, vi, wi, xi, yi, si, uj, vj, wj, xj, yj, sj, 
		      rbins, na, tempr, tempi, n_bins, &kmax1, imagflag, 1);
	}
	else 
	    calcints7(ui, vi, wi, xi, yi, fi, uj, vj, wj, xj, yj, fj, 
		      rbins, na, tempr, tempi, n_bins, &kmax1, imagflag, 1);
	common->maxbin1[p] = kmax1;
	for(m = 1; m <= n_bins; m++) {
	  common->integral1r[m][p] = (float)tempr[m];
	  common->integral1i[m][p] = (float)tempi[m];
	}
      }                

      /* 
       *  perform the numerical integration for < V_i V_j > 
       *  the non-zero elements have maxbin set to 0 (see above)
       */
      if (common->maxbin2[p] == 0) { 
#if 0
	printf("<V V>\n");
	testints(ui, vi, wi, xi, yi, fi, si, uj, vj, wj, xj, yj, fj, sj, 
		 rbins, na, n_bins, imagflag, 2);
#endif
	/* perform the numerical integration for < V_i V_j > */
	if (gaussian_beam) {

	  if (!imagflag)
	    calcints (ui, vi, xi, yi, si, uj, vj, xj, yj, sj, rbins, na, 
		      tempr, tempi, n_bins, &kmax2, imagflag, 2);
	  else
	    calcints6(ui, vi, wi, xi, yi, si, uj, vj, wj, xj, yj, sj, 
		      rbins, na, tempr, tempi, n_bins, &kmax2, imagflag, 2);
	}
	else 
	  calcints7(ui, vi, wi, xi, yi, fi, uj, vj, wj, xj, yj, fj, 
		    rbins, na, tempr, tempi, n_bins, &kmax2, imagflag, 2);
	common->maxbin2[p] = kmax2;
	for(m = 1; m <= n_bins; m++) {
	  common->integral2r[m][p] = (float)tempr[m];
	  common->integral2i[m][p] = (float)tempi[m];
	}
      }                

    } 

    /* 
     * provide user with information about progress of calculation 
     */
    int_done += common->icn[j+1] - common->icn[j];
    if (int_done > percent * int_total / 10) {
      if (rank == ROOT) 
	printf("  Fraction of table completed = %3i percent\n", 
	       percent * 10);
      percent ++;
    }
  }

#if defined(MPI)
  /*
   * if MPI is used, broadcast integrals to other processors
   */

  if (npe>1) 
    {

  for (i = 0; i < npe; i++) {
    int sendcount, sendindex;

    sendcount = common->icn[jidx[i+2]] - common->icn[jidx[i+1]];
    sendindex = common->icn[jidx[i+1]];

    /* send maxbin arrays */
    MPI_Bcast(&common->maxbin1[sendindex], sendcount, MPI_INT, 
	      i, MPI_COMM_WORLD);
    MPI_Bcast(&common->maxbin2[sendindex], sendcount, MPI_INT, 
	      i, MPI_COMM_WORLD);
    /* send integrals */
    for (m = 1; m <= n_bins; m++) {
      MPI_Bcast(&common->integral1r[m][sendindex], sendcount, MPI_FLOAT, 
		i, MPI_COMM_WORLD);
      MPI_Bcast(&common->integral1i[m][sendindex], sendcount, MPI_FLOAT, 
		i, MPI_COMM_WORLD);
      MPI_Bcast(&common->integral2r[m][sendindex], sendcount, MPI_FLOAT, 
		i, MPI_COMM_WORLD);
      MPI_Bcast(&common->integral2i[m][sendindex], sendcount, MPI_FLOAT, 
		i, MPI_COMM_WORLD);
    }
  }

    } /* if npe>1  */
#endif

  free_1imatrix(jidx);

  return;
}

/*
 * copy a set of integrals from one struct to the second 
 * PARAMS
 *     int1 [I]  the integrals to be copied
 *     int2 [O]  the target 
 */
void copyintegrals(integral_struct *int1, integral_struct *int2)
{
  int i;

  int2->nv = int1->nv;
  int2->nz = int1->nz;
  int2->n_bins = int1->n_bins;
  int2->imaginary_flag = int1->imaginary_flag;
  int2->irn = int1->irn;
  int2->icn = int1->icn;
  int2->freqn = int1->freqn;
  int2->maxbin1 = int1->maxbin1;
  int2->maxbin2 = int1->maxbin2;
  for (i = 1; i <= int1->n_bins; i++) {
    int2->integral1r[i] = int1->integral1r[i]; 
    int2->integral2r[i] = int1->integral2r[i]; 
    int2->integral1i[i] = int1->integral1i[i]; 
    int2->integral2i[i] = int1->integral2i[i]; 
  }

  return;
}

/* determine the minimum number of adjacent bins needed to cover 
 * roughly a given width.
 */
int min_bins(double *rbins, int na, double width)
{
  int i, j;
  int imin = 1;
  double w;

  for (i = 1; i <= na; i++) {
    j = i;
    w = rbins[j+1] - rbins[i];
    while ((w < width) && (j < na)) {
      j++;
      w = rbins[j+1] - rbins[i];
    }
    if (j - i + 1 > imin) 
      imin = j - i + 1;
  }

  return imin;
}

static int initialised = 0;
static int state[NAMAX+1][NFMAX+1];
static integral_struct integral_store[NAMAX+1][NFMAX+1];

/*
 * returns the overlap integrals for a set of uv-positions
 * these can either be calculated or read from disk or memory
 * PARAMS
 *     integral_data [I,O]  struct containing the overlap integrals
 *     freqn    [I]  index array indicating which observing fields the
 *                   visibilities belong to
 *     u, v, w  [I]  uvw positions of visibilities  
 *     nv       [I]  number of calculated visibilities
 *     ra, dec  [I]  RA and DEC of each observed fields
 *     ra_fre, dec_ref [I] RA and DEC of global reference centre
 *     freq     [I]  observing frequency for each field
 *     antwdth  [I]  width of 1/e diameter of aperture in wavelengths 
 *                    for each observed fields
 *     sigma    [I]  width of Gaussian beams for each field
 *     nf       [I]  number of observed fields
 *     rbins    [I]  borders of bins, the i-th bin is [rbins[i], rbins[i+1]]
 *     na       [I]  total number of bins
 *     bin      [I]  index of current bin
 *     gaussian_beam [I] flag: use Gaussian beam approximation?
 *     imagflag [I]  flag: use complex calculations? 
 *     saveflag [I]  flag: save integrals to disk/memory?
 *     rank     [I]  rank of processor; if MPI is not used, set to 0
 *     npe      [I]  number of processors; if MPI is not used set to 1
 */
void get_integrals(integral_struct *integral_data, 
		   int *freqn, double *u, double *v, double *w, int nv,
		   double *ra, double *dec, double ra_ref, double dec_ref, 
		   double *freq, double *antwdth, double *sigma, int nf,
		   double *rbins, int na, int bin, int field, int ni,
		   int gaussian_beam, int imagflag, int saveflag, 
		   char prefix[], int rank, int npe) 
{
  int    i, j;
  int    nz;
  char   filename[FILENAME_LEN];
  int    n_bins = 0;
  double antmax = 0.;

  if (!initialised) {
    for (i = 1; i <= na; i++)
      for (j = 1; j <= ni; j++)
	state[i][j] = INT_STATE_NULL;
    if (saveflag & INT_LOAD_ONLY)
      for (i = 1; i <= na; i++)
	for (j = 1; j <= ni; j++)
	  state[i][j] = INT_STATE_DISK;
    initialised = 1;
  }
    
#ifdef FITSIO
  sprintf(filename, "%s%i_%i.fits", prefix, field, bin);
#elif defined ZLIB
  sprintf(filename, "%s%i_%i.dat.gz", prefix, field, bin);
#else
  sprintf(filename, "%s%i_%i.dat", prefix, field, bin);
#endif

  /* if integrals exist in memory, copy and exit */
  if (state[bin][field] & INT_STATE_MEMORY) {
      printf("  Loading integrals from memory...\n");
    copyintegrals(&integral_store[bin][field], integral_data);
    goto labelsavedisk;
  } 

  /* count non-zero elements of covariance matrix*/
  nz = countintegrals(antwdth, nf, freqn, u, v, nv);
  if (rank == ROOT) {
    if (nz < nv+1) {
      printf("WARNING: Matrix too sparse.\n");
      nz = nv+1;
    }
  }


  /* determine how many bins are needed to cover the aperture width
   * roughly (the factor 0.9 is arbitrarily chosen)
   */
  antmax = 0.;
  for (i = 1; i <= nf; i++)
    if (antwdth[i] > antmax)
      antmax = antwdth[i];
  n_bins = min_bins(rbins, na, 0.9 * antmax);
  if (rank == ROOT)
    printf("    [ Using semiexact == %i ]\n", n_bins);
  
  /* allocate memory for overlap integrals */
  newintegrals(integral_data, nv, nz, MAX(2*n_bins+1, 3));
  
  /* if integrals are on disk, load from disk */
  if (state[bin][field] & INT_STATE_DISK) {
    if (rank == ROOT)
      printf("  Loading integrals from disk...\n");
#ifndef FITSIO
    loadintegrals(filename, integral_data, freqn, rank, npe);
#else
    loadfitsintegrals(filename, integral_data, freqn, freq, nf);
#endif
  }
  else {
    /* calculate overlap integrals */
    if (rank == ROOT)
      printf("  Making look-up table of required integrals ...\n");
    makeintegrals(integral_data, freqn, u, v, w, nv, 
		  ra, dec, ra_ref, dec_ref, freq, antwdth, sigma, nf,
		  rbins, na, 1, imagflag, rank, npe);
  }

  /* store integrals in memory */
  if ((saveflag & INT_STORE_MEMORY) 
      && !(state[bin][field] & INT_STATE_MEMORY)) {
    copyintegrals(integral_data, &integral_store[bin][field]);
    state[bin][field] |= INT_STATE_MEMORY;
    printf("  Copying integrals to memory\n");
  }

 labelsavedisk:
  /* save integrals to disk */
  if ((saveflag & INT_STORE_DISK) && !(state[bin][field] & INT_STATE_DISK)) {
    if (rank == ROOT)
      printf("  Writing integrals to disk...\n");
#ifndef FITSIO
    saveintegrals(filename, integral_data, rank, npe);
#else
    savefitsintegrals(filename, integral_data);
#endif

    state[bin][field] |= INT_STATE_DISK;
  }

  return;
}

void dealloc_integrals(int na, int ni)
{
  int bin, field;

  for (bin = 1; bin <= na; bin++) 
    for (field = 1; field <= ni; field++) {
      if (state[bin][field] & INT_STATE_MEMORY)
	freeintegrals(& integral_store[bin][field]);
      state[bin][field] = state[bin][field] & (~INT_STATE_MEMORY);
  }

  return;
}

void reset_integrals(int na, int ni)
{
  int bin, field;

  dealloc_integrals(na, ni);
  for (bin = 1; bin <= na; bin++) 
    for (field = 1; field <= ni; field++) {
      state[bin][field] = INT_STATE_NULL;
  }
  initialised = 0;

  return;
}

/*
 * save the overlap integrals to an ASCII text file
 * PARAMS 
 *     filename [I]  filename to which the integrals are saved
 *     ints     [I]  struct containing the overlap integrals
 */
void saveintegrals(char filename[], integral_struct *ints, int rank, int npe)
{
  FILE *fp;
  int i, j, k, l;
  
  /* only one process should write out to the disk */
  if (rank != ROOT)
    return;

  fp = FOPEN(filename, "w");
  if (fp == NULL) {
    printf("ERROR: Could not open %s\n", filename);
    return;
  }
  
  /* write header with dimensions of data structures */
  FPRINTF(fp, "%6i %10i %2i %1i\n", ints->nv, ints->nz, ints->n_bins,
	  ints->imaginary_flag);
  /* save integrals and associated structures */
  for (j = 1; j <= ints->nv; j++)
    for (k = ints->icn[j]; k <= ints->icn[j+1]-1; k++) {
      i = ints->irn[k];
      /* save maxbin arrays */
      FPRINTF(fp, "%6i %6i %2i %2i", i, j, ints->maxbin1[k], ints->maxbin2[k]);
      /* save integrals */
      for (l = 1; l <= ints->n_bins; l++)
	FPRINTF(fp, " %.10g", ints->integral1r[l][k]);
      for (l = 1; l <= ints->n_bins; l++)
	FPRINTF(fp, " %.10g", ints->integral2r[l][k]);
      FPRINTF(fp, "\n");

      /* save imaginary parts of integrals */
      if (ints->imaginary_flag) {
	FPRINTF(fp, "%6i %6i", i, j);
	for (l = 1; l <= ints->n_bins; l++)
	  FPRINTF(fp, " %.10g", ints->integral1i[l][k]);
	for (l = 1; l <= ints->n_bins; l++)
	  FPRINTF(fp, " %.10g", ints->integral2i[l][k]);
	FPRINTF(fp, "\n");
      }
   }

  FCLOSE(fp);
  return;
}

/* 
 * load overlap integrals from an ASCII text file 
 * PARAMS 
 *     filename [I]  filename to which the integrals are saved
 *     ints   [I,O]  struct containing the overlap integrals
 *     freqn    [I]  index array indicating which observing fields the
 *                   visibilities belong to
 *     freq     [I]  observing frequency for each field
 *     nf       [I]  number of observed fields
 */
void loadintegrals(char filename[], integral_struct *ints, int *freqn,
		   int rank, int npe)
{
  FILE   *fp;
  int    j, i2, j2, k, l;
  int    nz, nv, nb, im, nread;
  char   buffer[256];
  float  d[14]; 

  /* only one process should access the disk */
  if (rank == ROOT) {

    fp = FOPEN(filename, "r");
    if (fp == NULL) {
      printf("ERROR: Could not open %s. Make sure that file exists\n", 
	     filename);
      return;
    }

    /* read header */
    FGETS(buffer, 255, fp);
    sscanf(buffer, "%i %i %i %i", &nv, &nz, &nb, &im);
    ints->imaginary_flag = im;
    if ((nv == ints->nv) && (nz == ints->nz) && (nb == ints->n_bins)) 
      printf("  Reading %i integrals for %i visibilities from %s...\n",
	     nz, nv, filename);
    else {
      printf("  Read: %i %i %i %i\n", nv, nz, nb, im);
      printf("  Expected: %i %i %i %i\n", ints->nv, ints->nz, ints->n_bins,
	     ints->imaginary_flag);
      myerror("ERROR: parameters in integral file do not match!");
    }
    if (nb > 7)
      printf("ERROR: Cannot parse %i integrals\n", nb);

    /* read body of file with integrals */
    j = 0;
    for (k = 1; k <= ints->nz; k++) {
      /* read real parts of integrals */
      FGETS(buffer, 255, fp);
      nread = sscanf(buffer, 
		     "%i %i %i %i %f %f %f %f %f %f %f %f %f %f %f %f %f %f",
		     &i2, &j2, &ints->maxbin1[k], &ints->maxbin2[k],
		     &d[0], &d[1], &d[2], &d[3], &d[4], &d[5], &d[6], &d[7],
		     &d[8], &d[9], &d[10], &d[11], &d[12], &d[13]);
      if (nread != 4 + 2*ints->n_bins)
	goto loadinterror;
      for (l = 1; l <= ints->n_bins; l++) {
	ints->integral1r[l][k] = d[l-1];
	ints->integral2r[l][k] = d[l-1+ints->n_bins];
      }
      
      ints->irn[k] = i2;
      while (j2 > j) {
	j++;
	ints->icn[j] = k;
      }
      
      /* read imaginary parts of integrals */
      if (ints->imaginary_flag) {
	FGETS(buffer, 255, fp);
	nread = sscanf(buffer, 
		       "%i %i %f %f %f %f %f %f %f %f %f %f %f %f %f %f",
		       &i2, &j2, 
		       &d[0], &d[1], &d[2], &d[3], &d[4], &d[5], &d[6], &d[7],
		       &d[8], &d[9], &d[10], &d[11], &d[12], &d[13]);
	if (nread != 2 + 2*ints->n_bins)
	  goto loadinterror;
	for (l = 1; l <= ints->n_bins; l++) {
	  ints->integral1i[l][k] = d[l-1];
	  ints->integral2i[l][k] = d[l-1+ints->n_bins];
	}
	
	if ((i2 != ints->irn[k]) || (j2 != j))
	  printf("ERROR reading imag. parts: index %i %i (expected %i %i)\n",
		 i2, j2, ints->irn[k], j); 
      }
    }
    ints->icn[ints->nv+1] = ints->nz+1;
    FCLOSE(fp);

    printf("  Finished reading visibilities.\n");
  }

  for (j = 1; j <= ints->nv; j++)
    ints->freqn[j] = freqn[j];

#if defined(MPI)
  /* 
   * if MPI is used, broadcast integrals from root process 
   * to all other processes 
   */
  if (npe > 1) {
    MPI_Bcast(&(ints->imaginary_flag), 1, MPI_INT, ROOT, MPI_COMM_WORLD); 
    MPI_Bcast(ints->icn+C_OFFSET, ints->nv+1, MPI_INT, ROOT, MPI_COMM_WORLD); 
    MPI_Bcast(ints->irn+C_OFFSET, ints->nz,   MPI_INT, ROOT, MPI_COMM_WORLD); 
    MPI_Bcast(ints->maxbin1+C_OFFSET, ints->nz, MPI_INT, ROOT, 
	      MPI_COMM_WORLD); 
    MPI_Bcast(ints->maxbin2+C_OFFSET, ints->nz, MPI_INT, ROOT, 
	      MPI_COMM_WORLD); 
    for (l = 1; l <= ints->n_bins; l++) {
      MPI_Bcast(ints->integral1r[l]+C_OFFSET, ints->nz, MPI_INT, ROOT, 
		MPI_COMM_WORLD); 
      MPI_Bcast(ints->integral1i[l]+C_OFFSET, ints->nz, MPI_INT, ROOT, 
		MPI_COMM_WORLD); 
      MPI_Bcast(ints->integral2r[l]+C_OFFSET, ints->nz, MPI_INT, ROOT, 
		MPI_COMM_WORLD); 
      MPI_Bcast(ints->integral2i[l]+C_OFFSET, ints->nz, MPI_INT, ROOT, 
		MPI_COMM_WORLD); 
    }
  }
#endif
  return;

 loadinterror:
  {
    char s[125];
    sprintf(s, "FATAL ERROR occured reading element %i from file %s.",
	    k, filename);
    myerror(s);
  }
  return;
}


#ifdef FITSIO

/* 
 * note: these functions aren't MPI safe
 * they are probably getting outdated as well
 */
void printerror( int status)
{
  /*****************************************************/
  /* Print out cfitsio error messages and exit program */
  /*****************************************************/
  
  
  if (status) {
    fits_report_error(stderr, status); /* print error report */
    
    exit( status );    /* terminate the program, returning error status */
  }
  
  return;
}

#define N_FITS_LEN 4*N_BINS+4

/* 
 * save the integrals to a FITS file with binary table extension 
 * PARAMS 
 *     filename [I]  filename to which the integrals are saved
 *     ints     [I]  struct containing the overlap integrals
 */
void savefitsintegrals(char filename[], integral_struct *ints)
{
  fitsfile *fptr;       /* pointer to the FITS file, defined in fitsio.h */
  int status;
  long firstrow, firstelem;
  
  int tfields;             /* table will have tfields columns */
  long nrows;              /* table will have nrows rows    */

  char extname[] = "VSA_integrals";           /* extension name */

  /* define the name, datatype, and physical units for the columns */
  char *ttype[N_FITS_LEN]; char stype[N_FITS_LEN][16];
  char *tform[N_FITS_LEN]; char sform[N_FITS_LEN][4];
  char *tunit[N_FITS_LEN]; char sunit[N_FITS_LEN][10];

  int i, j, k;
  int *iidx, *jidx;

  /* set pointers to some allocated memory - UGLY!!! */
  for (i = 0; i < N_FITS_LEN; i++) {
    ttype[i] = stype[i];
    tform[i] = sform[i];
    tunit[i] = sunit[i];
  }
    
  tfields = 2 + 1 + 2 * ints->n_bins;
  /* the first two columns for the indices i,j in the matrix */
  sprintf(ttype[0], "i");
  sprintf(tform[0], "1J");
  sprintf(tunit[0], " ");
  sprintf(ttype[1], "j");
  sprintf(tform[1], "1J");
  sprintf(tunit[1], " ");
  /* bins in which the maxima occur */
  sprintf(ttype[2], "max1");
  sprintf(tform[2], "1J");
  sprintf(tunit[2], " ");
  
  for (i = 0; i <= 1; i++)
    for (j = 0; j < ints->n_bins; j++) {
      sprintf(ttype[3+j+i*ints->n_bins], "int%ir%i", i+1, j+1);
      sprintf(tform[3+j+i*ints->n_bins], "1D");
      sprintf(tunit[3+j+i*ints->n_bins], " ");
    }

  if (ints->imaginary_flag) {
    tfields += 1 + 2 * ints->n_bins;
    sprintf(ttype[3+2*ints->n_bins], "max2");
    sprintf(tform[3+2*ints->n_bins], "1J");
    sprintf(tunit[3+2*ints->n_bins], " ");

    for (i = 0; i <= 1; i++)
      for (j = 0; j < ints->n_bins; j++) {
	sprintf(ttype[4+j+(2+i)*ints->n_bins], "int%ii%i", i+1, j+1);
	sprintf(tform[4+j+(2+i)*ints->n_bins], "1D");
	sprintf(tunit[4+j+(2+i)*ints->n_bins], " ");
      }
  }

#if 0
  for (i = 0; i < tfields; i++) 
    printf("%16s %2s %8s\n", ttype[i], tform[i], tunit[i]);
#endif

  remove(filename);               /* Delete old file if it already exists */

  nrows = ints->nz;
  
  status=0;

  /* open the FITS file containing a primary array and an ASCII table */
  if ( fits_create_file(&fptr, filename, &status) ) 
    printerror( status );

#if 0
  printf("Moving to second header unit..\n");
  if ( fits_movabs_hdu(fptr, 2, &hdutype, &status) ) /* move to 2nd HDU */
    printerror( status );
#endif

  /* append a new empty binary table onto the FITS file */
  if ( fits_create_tbl( fptr, BINARY_TBL, nrows, tfields, ttype, tform,
			tunit, extname, &status) )
    printerror( status );

  fits_write_key(fptr, TINT, "INTNV", &ints->nv, 
		 "number of visibilities", &status);
  fits_write_key(fptr, TINT, "INTNZ", &ints->nz, 
		 "number of non-zero elements", &status);
  fits_write_key(fptr, TINT, "INTNBINS", &ints->n_bins, 
		 "number of bins", &status);
  fits_write_key(fptr, TINT, "IMAGFLAG", &ints->imaginary_flag, 
		 "use imaginary parts or not", &status);


  firstrow  = 1;  /* first row in table to write   */
  firstelem = 1;  /* first element in row  (ignored in ASCII tables) */
  
  /* write columns  */
  iidx = malloc_1imatrix(ints->nz);
  jidx = malloc_1imatrix(ints->nz);
  for (j = 1; j <= ints->nv; j++)
    for (k = ints->icn[j]; k <= ints->icn[j+1]-1; k++) {
      iidx[k] = ints->irn[k];
      jidx[k] = j;
    }
  fits_write_col(fptr, TINT, 1, firstrow, firstelem, nrows, 
		 iidx+C_OFFSET, &status);
  fits_write_col(fptr, TINT, 2, firstrow, firstelem, nrows, 
		 jidx+C_OFFSET, &status);

#if 0
  for (k = 1; k <= 5; k++) {
    printf("%4i: %3i %3i, %3i ", k, iidx[k], jidx[k], ints->maxbin1[k]);
    for (j = 1; j <= ints->n_bins; j++)
      printf("%g ", ints->integral1r[j][k]);
    for (j = 1; j <= ints->n_bins; j++)
      printf("%g ", ints->integral2r[j][k]);
    printf("\n");
  }
#endif

  free_1imatrix(iidx);
  free_1imatrix(jidx);

  fits_write_col(fptr, TINT, 3, firstrow, firstelem, nrows, 
		 ints->maxbin1+C_OFFSET, &status);
  
  for (j = 1; j <= ints->n_bins; j++)
    fits_write_col(fptr, TFLOAT, 3+j, firstrow, firstelem, 
		   nrows, ints->integral1r[j]+C_OFFSET, &status);
  for (j = 1; j <= ints->n_bins; j++)
    fits_write_col(fptr, TFLOAT, 3+j+ints->n_bins, firstrow, firstelem, 
		   nrows, ints->integral2r[j]+C_OFFSET, &status);

  if (ints->imaginary_flag) {
    fits_write_col(fptr, TINT, 4+2*ints->n_bins, firstrow, firstelem, 
		   nrows, ints->maxbin2+C_OFFSET, &status);
    
    for (j = 1; j <= ints->n_bins; j++)
      fits_write_col(fptr, TFLOAT, 4+2*ints->n_bins+j, firstrow, firstelem, 
		     nrows, ints->integral1i[j]+C_OFFSET, &status);
    for (j = 1; j <= ints->n_bins; j++)
      fits_write_col(fptr, TFLOAT, 4+2*ints->n_bins+j, firstrow, firstelem, 
		     nrows, ints->integral2i[j]+C_OFFSET, &status);
  }

  if ( fits_close_file(fptr, &status) )       /* close the FITS file */
    printerror( status );

  return;
}

/* 
 * load overlap integrals from a FITS file 
 * PARAMS 
 *     filename [I]  filename to which the integrals are saved
 *     ints   [I,O]  struct containing the overlap integrals
 *     freqn    [I]  index array indicating which observing fields the
 *                   visibilities belong to
 *     freq     [I]  observing frequency for each field
 *     nf       [I]  number of observed fields
*/
void loadfitsintegrals(char filename[], integral_struct *ints, int *freqn,
		       double *freq, int nf)
{
  fitsfile *fptr;       /* pointer to the FITS file, defined in fitsio.h */
  int status, hdutype,  nfound, anynull, ii, j, k;
  long hdunum, frow, felem, nelem, longnull;
  char *ttype[N_FITS_LEN]; 
  double dblnull;
  int *iidx, *jidx;
  int nv, nz, nbin, imagflag;
  
  status = 0;

  if ( fits_open_file(&fptr, filename, READONLY, &status) )
    printerror( status );

  for (ii = 0; ii < N_FITS_LEN; ii++)      /* allocate space for the column labels */
    ttype[ii] = (char *) malloc(FLEN_VALUE);  /* max label length = 69 */
  
  /* move to the HDU */
  hdunum = 2;
  if ( fits_movabs_hdu(fptr, hdunum, &hdutype, &status) ) 
    printerror( status );

  if (hdutype != BINARY_TBL) {
    printf("Error: this HDU is not an ASCII or binary table\n");
    printerror( status );
  }

  /* read the column names from the TTYPEn keywords */
  fits_read_keys_str(fptr, "TTYPE", 1, N_FITS_LEN, ttype, &nfound, &status);
  fits_read_key(fptr, TINT, "INTNV", &nv, NULL, &status);
  fits_read_key(fptr, TINT, "INTNZ", &nz, NULL, &status);
  fits_read_key(fptr, TINT, "INTNBINS", &nbin, NULL, &status);
  fits_read_key(fptr, TINT, "IMAGFLAG", &imagflag, NULL, &status);
  if (nv != ints->nv)
    myerror("wrong number of visibilities in FITS file.");
  if (nz != ints->nz)
    myerror("wrong number of elements in FITS file.");
  if (nbin != ints->n_bins)
    myerror("wrong number of bins in FITS file.");
  if (imagflag != ints->imaginary_flag)
    myerror("wrong flag for imaginary parts in FITS file.");

#if 0
  for (i = 0; i < nfound; i++)
    printf("  Row %3i: %s\n", i+1, ttype[i]);
#endif
  
  frow      = 1;
  felem     = 1;
  nelem     = ints->nz;
  longnull = 0;
  dblnull  = 0.;

  /*  read the columns */ 
  iidx = malloc_1imatrix(ints->nz);
  jidx = malloc_1imatrix(ints->nz);
  fits_read_col(fptr, TINT, 1, frow, felem, nelem, &longnull,
		iidx+C_OFFSET, &anynull, &status);
  fits_read_col(fptr, TINT, 2, frow, felem, nelem, &longnull,
		jidx+C_OFFSET, &anynull, &status);
#if 0
  for (k = 1; k <= 5; k++) {
    printf("%4i: %3i %3i, %3i", k, iidx[k], jidx[k], ints->maxbin1[k]);
    printf("\n");
  }
#endif

  for (j = 1; j <= nv; j++)
    ints->freqn[j] = freqn[j];

  j = 0;
  for (k = 1; k <= ints->nz; k++) {
    ints->irn[k] = iidx[k];
    while (jidx[k] > j) {
      j++;
      ints->icn[j] = k;
    }
  }
  free_1imatrix(iidx);
  free_1imatrix(jidx);

  fits_read_col(fptr, TINT, 3, frow, felem, nelem, &longnull,
		ints->maxbin1+C_OFFSET, &anynull, &status);

  for (j = 1; j <= ints->n_bins; j++)
    fits_read_col(fptr, TDOUBLE, 3+j, frow, felem, nelem, 
		  &dblnull, ints->integral1r[j]+C_OFFSET, &anynull, &status);
  for (j = 1; j <= ints->n_bins; j++)
    fits_read_col(fptr, TDOUBLE, 3+j+ints->n_bins, frow, felem, nelem, 
		  &dblnull, ints->integral2r[j]+C_OFFSET, &anynull, &status);

  if (ints->imaginary_flag) {
    fits_read_col(fptr, TINT, 4+2*ints->n_bins, frow, felem, nelem, 
		  &longnull, ints->maxbin2+C_OFFSET, &anynull, &status);

    for (j = 1; j <= ints->n_bins; j++)
      fits_read_col(fptr, TDOUBLE, 4+2*ints->n_bins+j, frow, felem, nelem, 
		    &dblnull, ints->integral1i[j]+C_OFFSET, &anynull, &status);
    for (j = 1; j <= ints->n_bins; j++)
      fits_read_col(fptr, TDOUBLE, 4+3*ints->n_bins+j, frow, felem, nelem, 
		    &dblnull, ints->integral2i[j]+C_OFFSET, &anynull, &status);
  }

  for (ii = 0; ii < N_FITS_LEN; ii++)      /* free the memory for the column labels */
    free( ttype[ii] );
  
  if ( fits_close_file(fptr, &status) ) 
    printerror( status );
  
  return;
}
#endif
/*-----*-----------------------------------------------------------------*/
