/*  Time-stamp: <12 Nov 2002 18:06 GMT like_io.c maising> */
/*
 * this file contains various routines for interactive user input 
 * and reading and writing from data files
 *
 * HISTORY
 *   ??-05-01  KSM  original version
 */

#include <stdio.h>
#include <math.h>
#include <time.h>
#include <string.h>

#ifdef MPI
#include <mpi.h>
#endif

#include "like_io.h"
#include "like_integrals.h"
#include "my_malloc2.h"
#include "myslalib.h"
#include "params2.h"

#define PI 3.14159265358979323844
#define DEG2RAD (3.14159265358979323844/180.)
#define MIN2RAD (3.14159265358979323844/180./60.)
#define ROOT 0

/*-----*-----------------------------------------------------------------*/
int get_integral_flag(int flag)
{
  int flags[5] = { 0., 
		   INT_STORE_MEMORY,
		   INT_STORE_DISK,
		   INT_LOAD_ONLY,
		   INT_STORE_DISK | INT_SAVE_ONLY };
  if ((flag < 0) || (flag > 4))
    return 0;
  else return flags[flag];
}

#ifdef MPI
void broadcast_like_params(int *sparseflag, int *maximiseflag,
			   int *hessianflag,
			   int *galflag, int *saveflag, char outputfn[],
			   int *restart, int root)
{
  MPI_Bcast(sparseflag, 1, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(maximiseflag, 1, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(hessianflag, 1, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(galflag, 1, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(saveflag, 1, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(outputfn, FILENAME_LEN, MPI_CHAR, root, MPI_COMM_WORLD);
  MPI_Bcast(restart, 1, MPI_INT, root, MPI_COMM_WORLD);

  return;
}
#endif

/*
 * read parameters specific to the vsalike program
 */
void read_like_params(int *sparseflag, int *maximise_flag,
		      int *hessian_flag, int *galflag,
		      int *saveflag, char outputfn[], int *restart,
		      int rank, int npe)
{
  if (rank == ROOT) {

#if defined(SPARSE_LIKE)
    *sparseflag = 1;
#elif defined(DENSE_LIKE)
    *sparseflag = 0;
#else
    printf("Use sparse (1) or dense (0) matrix routines: ");
    scanf("%i", sparseflag);
#endif

#if defined(MAXIMISE_LIKE)
    *maximise_flag = 1;
    *hessian_flag = 0;
#elif defined(HESSIAN_LIKE)
    *maximise_flag = 0;
    *hessian_flag = 1;
#else
    *hessian_flag = 1;
    printf("Maximise likelihood (1) or calculate Hessian (0): ");
    scanf("%i", maximise_flag);
#endif

    printf("Estimate PS of Galaxy? (0 - no, 1 - yes): ");
    scanf("%i", galflag);

    printf("Recalculate integrals in each iteration (0), store integrals in memory (1) or calculate integrals and save to disk for further iterations (2), always load integrals from disk (3), save to disk and exit (4): ");
    scanf("%i", saveflag); *saveflag = get_integral_flag(*saveflag);

    printf("Name of output file for results: ");
    scanf("%s",  outputfn);

    *restart = 0;
  }

#ifdef MPI
  broadcast_like_params(sparseflag, maximise_flag, hessian_flag, 
			galflag, saveflag, outputfn, restart, ROOT);
#endif

  return;
}

void fread_like_params(char filename[], int *sparseflag, 
		       int *maximise_flag, int *hessian_flag, 
		       int *galflag, int *saveflag, 
		       char outputfn[], int *restart, int rank, int npe)
{
#if defined(SPARSE_LIKE)
  *sparseflag = 1;
#else
  *sparseflag = 0;
#endif

#if defined(MAXIMISE_LIKE)
  *maximise_flag = 1;
  *hessian_flag = 0;
#elif defined(HESSIAN_LIKE)  
  *maximise_flag = 0;
  *hessian_flag = 1;
#else
  *maximise_flag = 1;
  *hessian_flag = 1;
#endif
  *galflag = 0;
  *saveflag = 0;
  *restart = 0;
  sprintf(outputfn, "vsalike.out");

  if (rank == ROOT) {
    if (!open_param_file(filename)) {
#if ((! defined(MAXIMISE_LIKE)) && (! defined(HESSIAN_LIKE)))
      fget_param_int(maximise_flag, "MAXFLAG");
#endif
#if ((! defined(SPARSE_LIKE)) && (! defined(DENSE_LIKE)))
      fget_param_int(sparseflag, "SPARSEFLAG");
#endif
      fget_param_int(galflag, "GALFLAG");
      fget_param_int(saveflag, "STOREFLAG");  
      *saveflag = get_integral_flag(*saveflag);
      fget_param_string(outputfn, "OUTPUTFN");
      fget_param_int(restart, "RESTART");
      close_param_file();    
    }
  }

#ifdef MPI
  broadcast_like_params(sparseflag, maximise_flag, hessian_flag, 
			galflag, saveflag, outputfn, restart, ROOT);
#endif

  return;
}

/*-----*-----------------------------------------------------------------*/

#ifdef MPI
void broadcast_lfcn_params(int *plotflag, int *outflag, char *outputfn,
			   int *sparseflag, int *converge, int *saveflag, 
			   int root)
{
  MPI_Bcast(plotflag, 1, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(outflag, 1, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(outputfn, FILENAME_LEN, MPI_CHAR, root, MPI_COMM_WORLD);
  MPI_Bcast(sparseflag, 1, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(saveflag, 1, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(converge, 1, MPI_INT, root, MPI_COMM_WORLD);

  return;
}
#endif

void fread_lfcn_params(char filename[],
		       int *plotflag, int *outflag, char *outputfn,
		       int *sparseflag, int *converge, int *saveflag, 
		       int rank, int npe)
{
#if defined(SPARSE_LIKE)
  *sparseflag = 1;
#else
  *sparseflag = 0;
#endif
  *plotflag = 0;
  *outflag = 0;
  *converge = 1;
  *saveflag = 2;
  if (rank == ROOT) {
    if (!open_param_file(filename)) {
      fget_param_int(plotflag, "PRINTFLAG");
      fget_param_int(outflag, "WRITEFLAG");
      fget_param_string(outputfn, "LIKEFILE");
      fget_param_int(saveflag, "STOREFLAG");
      *saveflag = get_integral_flag(*saveflag);
      fget_param_int(converge, "CONVERGE");
#if ((! defined(SPARSE_LIKE)) && (! defined(DENSE_LIKE)))
      fget_param_int(sparseflag, "SPARSEFLAG");
#endif
      close_param_file();    
    }
  }

#ifdef MPI
  broadcast_lfcn_params(plotflag, outflag, outputfn, sparseflag, 
			converge, saveflag, ROOT);
#endif

  return;
}

/*
 * read parameters specific to the vsaeigen and vsalfcn programs
 */
void read_lfcn_params(int *plotflag, int *outflag, char *outputfn,
		      int *sparseflag, int *converge, int *saveflag, 
		      int rank, int npe)
{
  if (rank == ROOT) {
    printf("Write out likelihood to screen (0 - no, 1 - yes): ");
    scanf("%i", plotflag);
    
    printf("Write likelihood to a file (0 - no, 1 - yes): ");
    scanf("%i", outflag);

    if (*outflag) {
      *outflag = 1;
      printf("Output filename for likelihood : ");
      scanf("%s", outputfn); 
    }

#if defined(SPARSE_LIKE)
    *sparseflag = 1;
#elif defined(DENSE_LIKE)
    *sparseflag = 0;
#else
    printf("Use sparse (1) or dense (0) matrix routines: ");
    scanf("%i", sparseflag);
#endif
    
    printf("Recalculate integrals in each iteration (0), store integrals in memory (1) or calculate integrals and save to disk for further iterations (2), always load integrals from disk (3), save to disk and exit (4): ");
    scanf("%i", saveflag); *saveflag = get_integral_flag(*saveflag);
    
    printf("Run until converged (1) or single iteration first (0): ");
    scanf("%i", converge);
  }
#ifdef MPI
  broadcast_lfcn_params(plotflag, outflag, outputfn, sparseflag, 
			converge, saveflag, ROOT);
#endif

  return;
}
/*-----*-----------------------------------------------------------------*/

void init_lim(double *cpsmin, double *cpsmax, double *cpsstep, int *ncps,
	      double *gpsmin, double *gpsmax, double *gpsstep, int *ngps,
	      double *spmin, double *spmax, double *spstep, int *nsp, 
	      int *gal_flag)
{
  *gal_flag = 0;
  *nsp = 1;
  *ngps = 1;
  *ncps = 201;
  *gpsmin = 0.;
  *gpsstep = 0.;
  *gpsmax = 0.;
  *spmin = 0.;
  *spmin = 0.;
  *spstep = 0.;
  *cpsmin  = 0.;
  *cpsstep  = 0.;
  *cpsmax  = 2.e-9;

  return;
}

#ifdef MPI
void broadcast_lim(double *cpsmin, double *cpsmax, int *ncps,
		   double *gpsmin, double *gpsmax, int *ngps,
		   double *spmin, double *spmax, int *nsp, 
		   int *gal_flag, int root)
{
  MPI_Bcast(gal_flag, 1, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(ncps, 1, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(ngps, 1, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(nsp, 1, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(cpsmin, 1, MPI_DOUBLE, root, MPI_COMM_WORLD);
  MPI_Bcast(cpsmax, 1, MPI_DOUBLE, root, MPI_COMM_WORLD);
  MPI_Bcast(gpsmin, 1, MPI_DOUBLE, root, MPI_COMM_WORLD);
  MPI_Bcast(gpsmax, 1, MPI_DOUBLE, root, MPI_COMM_WORLD);
  MPI_Bcast(spmin, 1, MPI_DOUBLE, root, MPI_COMM_WORLD);
  MPI_Bcast(spmax, 1, MPI_DOUBLE, root, MPI_COMM_WORLD);
}
#endif

void fread_lim(char filename[],
	       double *cpsmin, double *cpsmax, double *cpsstep, int *ncps,
	       double *gpsmin, double *gpsmax, double *gpsstep, int *ngps,
	       double *spmin, double *spmax, double *spstep, int *nsp, 
	       int *gal_flag, int rank, int npe)
{
  init_lim(cpsmin, cpsmax, cpsstep, ncps, gpsmin, gpsmax, gpsstep, ngps,
	   spmin, spmax, spstep, nsp, gal_flag);

  if (rank == ROOT) {
    /* ... input cmb spectrum limits in each annulus and no. of steps*/
    if (!open_param_file(filename)) {
      fget_param_int(gal_flag, "GALFLAG");
      fget_param_int(ncps, "NCPS");
      fget_param_int(ngps, "NGPS");
      fget_param_int(nsp, "NSP");
      fget_param_double(cpsmin, "CPSMIN");
      fget_param_double(cpsmax, "CPSMAX");
      fget_param_double(gpsmin, "GPSMIN");
      fget_param_double(gpsmax, "GPSMAX");
      fget_param_double(spmin, "SPMIN");
      fget_param_double(spmax, "SPMAX");
      close_param_file();    
    }
  }

#ifdef MPI
  broadcast_lim(cpsmin, cpsmax, ncps, gpsmin, gpsmax, ngps,
		spmin, spmax, nsp, gal_flag, ROOT);
#endif
  if (*ncps > 1)
    *cpsstep = (*cpsmax - *cpsmin) / ((double)(*ncps-1));
  if (*ngps > 1)
    *gpsstep = (*gpsmax - *gpsmin) / ((double)(*ngps-1));
  if (*nsp > 1)
    *spstep = (*spmax - *spmin) / ((double)(*nsp-1));

  return;
}

/*
 * read limits and step sizes on the parameters to be estimated
 */
void read_lim(double *cpsmin, double *cpsmax, double *cpsstep, int *ncps,
	      double *gpsmin, double *gpsmax, double *gpsstep, int *ngps,
	      double *spmin, double *spmax, double *spstep, int *nsp, 
	      int *gal_flag, int rank, int npe)
{
  init_lim(cpsmin, cpsmax, cpsstep, ncps, gpsmin, gpsmax, gpsstep, ngps,
	   spmin, spmax, spstep, nsp, gal_flag);

  if (rank == ROOT) {
    /* ... input cmb spectrum limits in each annulus and no. of steps*/
    printf("\n");
    printf("Input range of l^2 C_l/2pi for CMB: ");
    scanf("%lf %lf", cpsmin, cpsmax);

    printf("Input number of CMB PS steps: ");
    scanf("%i", ncps);
    
    /* ... input galaxy spectrum limits in each annulus and no. of steps*/
    printf("Estimate PS of Galaxy? (0 - no, 1 - yes): ");
    scanf("%i", gal_flag);
    if (*gal_flag) {
      printf("Input range of l^2 C_l/2pi for Galaxy: ");
      scanf("%lf %lf", gpsmin, gpsmax);
      printf("Input number of GAL PS steps: ");
      scanf("%i", ngps);
      printf("Input limits on GAL flux spectral index: ");
      scanf("%lf %lf", spmin, spmax);
      printf("Input number of steps for spectral index: ");
      scanf("%i", nsp);
    }
  }

#ifdef MPI
  broadcast_lim(cpsmin, cpsmax, ncps, gpsmin, gpsmax, ngps,
		spmin, spmax, nsp, gal_flag, ROOT);
#endif  
  if (*ncps > 1)
    *cpsstep = (*cpsmax - *cpsmin) / ((double)(*ncps-1));
  if (*ngps > 1)
    *gpsstep = (*gpsmax - *gpsmin) / ((double)(*ngps-1));
  if (*nsp > 1)
    *spstep = (*spmax - *spmin) / ((double)(*nsp-1));

  return;
}
/*-----*-----------------------------------------------------------------*/

/*
 * read information about the primary beam shape
 */
void read_beam_info(int *gaussian_beam,
		    char beamfn[], double *beamcell, double *beamfreq)
{
  printf("Model telescope primary beam as Gaussian (1) or read beam profile from file (0): ");
#ifdef DEBUG
  *gaussian_beam = 1;
#else
  scanf("%i", gaussian_beam);
#endif
  if (*gaussian_beam == 0) {
#ifndef DEBUG
    printf("Enter filename for beam profile: ");
    scanf("%s", beamfn);
    
    printf("Enter cell size: ");
    scanf("%lf", beamcell);
    printf("Enter reference frequency: ");
    scanf("%lf", beamfreq);
#else
    sprintf(beamfn, "../challenge/vsabeam34.1_0.2.dat");
    *beamcell = 0.2;
    *beamfreq = 34.1;
#endif
  }

  return;
}

#ifdef MPI
void broadcast_initial_ps(double *cps1dinit, double *gps1dinit, int na, 
			  double *sp, int *read_flag, int root)
{
  MPI_Bcast(cps1dinit+C_OFFSET, na, MPI_DOUBLE, root, MPI_COMM_WORLD);
  MPI_Bcast(gps1dinit+C_OFFSET, na, MPI_DOUBLE, root, MPI_COMM_WORLD);
  MPI_Bcast(sp, 1, MPI_DOUBLE, root, MPI_COMM_WORLD);
  MPI_Bcast(read_flag, 1, MPI_DOUBLE, root, MPI_COMM_WORLD);

  return;
}
#endif

void fread_initial_ps(char filename[],
		      double *cps1dinit, double *gps1dinit, double *sp,
		      int na, int galflag, int *read_flag, int rank, int npe)
{
  int i;
  char msg[80];

  for (i = 1; i <= na; i++) {
    cps1dinit[i] = 5.e-10;
    gps1dinit[i] = 0.;
  }
  *sp = 0.;
  
  if (rank == ROOT) {
    if (!open_param_file(filename)) {
      fget_param_int(read_flag, "INITIALPS");
      if (*read_flag == 1) {
	for (i = 1; i <= na; i++) {
	  sprintf(msg, "ICPS_BIN%i", i); 
	  fget_param_double(&cps1dinit[i], msg);
	}
	if (galflag) {
	  for (i = 1; i <= na; i++) {
	    sprintf(msg, "IGPS_BIN%i", i); 
	    fget_param_double(&gps1dinit[i], msg);
	  }
	  fget_param_double(sp, "SPINDEX");
	}
      }
      close_param_file();
    }
  }

#ifdef MPI
  broadcast_initial_ps(cps1dinit, gps1dinit, na, sp, read_flag, ROOT);
#endif

  return;
}

/*
 * read initial values for power spectrum estimates
 */
void read_initial_ps(double *cps1dinit, double *gps1dinit, double *sp,
		     int na, int galflag, int *read_flag, int rank, int npe)
{
  int i;

  for (i = 1; i <= na; i++) {
    cps1dinit[i] = 5.e-10;
    gps1dinit[i] = 0.;
  }
  *sp = 0.;
  
  if (rank == ROOT) {
    printf("Set initial power spectrum (1), use defaults (0), or read from log file (2): ");
    scanf("%i", read_flag);
  
    if (*read_flag == 1) {
      printf("Enter CMB value for each bin: ");
      for (i = 1; i <= na; i++) 
	scanf("%lf", &cps1dinit[i]);
      if (galflag) {
	printf("Enter Gal. value for each bin: ");
	for (i = 1; i <= na; i++) 
	  scanf("%lf", &gps1dinit[i]);
	printf("Enter spectral index of Galaxy: ");
	scanf("%lf", sp);
      }
    }
  }

#ifdef MPI
  broadcast_initial_ps(cps1dinit, gps1dinit, na, sp, read_flag, ROOT);
#endif
  return;
}

/*
 * write out power spectrum parameters to a log file
 */
int write_param_log(char logfile[], double *cps1d, double *gps1d, double sp,
		    double *rbins, int na, int galflag, int iter, int ibin,
		    int rank, int npe)
{
  FILE *fp;
  int i;

  if (rank == ROOT) {
    fp = fopen(logfile, "w+");
    if (fp) {
      if (! galflag)
	for (i = 1; i <= na; i++)
	  gps1d[i] = 0;
      for (i = 1; i <= na; i++)
	fprintf(fp, "%g %g %g\n", (rbins[i] + rbins[i+1])/2., 
		cps1d[i], gps1d[i]);
      fprintf(fp, "%i %g\n", galflag, sp);
      fprintf(fp, "%i %i\n", iter, ibin);
      fclose(fp);
    }
    else
      printf("ERROR: cannot open logfile %s\n", logfile);
  }

  return 0;
}

/*
 * read power spectrum parameters from a log file
 */
int read_param_log(char logfile[], double *cps1d, double *gps1d, 
		   double *sp, double *rbins, int na, 
		   int galflag, int *iter, int *ibin, int rank, int npe)
{
  FILE *fp;
  int i, idum;
  double dum;

  if (rank == ROOT) {
    fp = fopen(logfile, "r");
    if (fp) {
      for (i = 1; i <= na; i++)
	if (fscanf(fp, "%lg %lg %lg", &dum, &cps1d[i], &gps1d[i]) != 3) {
	  printf("ERROR reading log file %s\n", logfile);
	  return 1;
	}
      fscanf(fp, "%i %lg", &idum, sp);
      if (idum != galflag) {
	printf("ERROR: galactic flag doesn't match (%i != %i)\n", 
	       idum, galflag);
	return 1;
      }
      fscanf(fp, "%i %i", iter, ibin);
      if ((*ibin <= 1) || (*ibin >= na)) {
	printf("ERROR: bin %i in %s does not exist\n", *ibin, logfile);
	return 1;
      }
      
      fclose(fp);
    }  
    else
      printf("ERROR: cannot open logfile %s\n", logfile);
  }

#ifdef MPI
  broadcast_initial_ps(cps1d, gps1d, na, sp, &idum, ROOT);
  MPI_Bcast(iter, 1, MPI_INT, ROOT, MPI_COMM_WORLD);
  MPI_Bcast(ibin, 1, MPI_INT, ROOT, MPI_COMM_WORLD);
#endif  

  return 0;
}

void init_inps(int *nf, double *freq, double *freq0, int *ni, int *fieldn,
	       double *ra, double *dec, double *ra_ref, double *dec_ref,
	       double *sigma, double *antwdth, double *antmax,
	       char datafn[NFMAX+1][FILENAME_LEN])
{
  *nf = 1;
  freq[1] = 30.;
  *freq0 = 30.;
  *ni = 1;
  fieldn[1] = 1;
  ra[1] = *ra_ref = 0.;
  dec[1] = *dec_ref = 30.;
  sigma[1] = 4.6;
  antwdth[1] = 0.;
  *antmax = 0.;
  sprintf(datafn[1], "VSA.vbn"); 

  return;
}

#ifdef MPI
void broadcast_inps(int *nf, double *freq, double *freq0,
		    int *ni, int *fieldn,
		    double *ra, double *dec, double *ra_ref, double *dec_ref,
		    double *sigma, 
		    char datafn[NFMAX+1][FILENAME_LEN], int root)
{
  int i;

  MPI_Bcast(nf, 1, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(freq0, 1, MPI_DOUBLE, root, MPI_COMM_WORLD);
  MPI_Bcast(freq+C_OFFSET, *nf, MPI_DOUBLE, root, MPI_COMM_WORLD);
  MPI_Bcast(ra+C_OFFSET, *nf, MPI_DOUBLE, root, MPI_COMM_WORLD);
  MPI_Bcast(dec+C_OFFSET, *nf, MPI_DOUBLE, root, MPI_COMM_WORLD);
  MPI_Bcast(sigma+C_OFFSET, *nf, MPI_DOUBLE, root, MPI_COMM_WORLD);
  MPI_Bcast(ni, 1, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(fieldn+C_OFFSET, *nf, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(ra_ref, 1, MPI_DOUBLE, root, MPI_COMM_WORLD);
  MPI_Bcast(dec_ref, 1, MPI_DOUBLE, root, MPI_COMM_WORLD);
  for (i = 1; i <= *nf; i++)
    MPI_Bcast(datafn[i], FILENAME_LEN, MPI_CHAR, root, MPI_COMM_WORLD);

  return;
}
#endif

void fread_inps(char filename[],
		int *nf, double *freq, double *freq0, int *ni, int *fieldn,
		double *ra, double *dec, double *ra_ref, double *dec_ref,
		double *sigma, double *antwdth,
		double *antmax, char datafn[NFMAX+1][FILENAME_LEN], 
		int rank, int npe)
{
  int i, status;
  char msg[128];
  double rah, ram, ras, decd, decm, decs;
  
  init_inps(nf, freq, freq0, ni, fieldn, ra, dec, ra_ref, dec_ref, 
	    sigma, antwdth, antmax, datafn);

  if (rank == ROOT) {
    if (!open_param_file(filename)) {
      fget_param_int(nf, "NFREQ");
      fget_param_int(ni, "NMOSAIC");
      if (*nf > NFMAX) 
	myerror("ERROR: too many data sets - recompile program!");
      for (i = 1; i <= *nf; i++) {
	sprintf(msg, "FREQ%i", i); 
	fget_param_double(&freq[i], msg);
      }
      for (i = 1; i <= *nf; i++) {
	fieldn[i] = 1;
	sprintf(msg, "MOSAIC%i", i); 
	fget_param_int(&fieldn[i], msg);
      }
      *ra_ref = 0.; *dec_ref = 0.;
      for (i = 1; i <= *nf; i++) {
	sprintf(msg, "RA%i", i);
	fget_param_deg(&rah, &ram, &ras, msg);
	sprintf(msg, "DEC%i", i);
	fget_param_deg(&decd, &decm, &decs, msg);

	/* convert to radians */
	sla_ctf2r(rah, ram, ras, &ra[i], &status); 
	sla_daf2r(decd, decm, decs, &dec[i], &status); 
	*ra_ref += ra[i]; *dec_ref += dec[i];
      }
      *ra_ref /= *nf; *dec_ref /= *nf;
      for (i = 1; i <= *nf; i++) 
	printf("  Field centre %2i: %10.5f %10.5f arcmin, %10.5f %10.5f rad\n",
	       i, ra[i] / MIN2RAD, dec[i] / MIN2RAD, ra[i], dec[i]);
      printf("  Reference centre:  %10.5f %10.5f arcmin, %10.5f %10.5f rad\n",
	     *ra_ref / MIN2RAD, *dec_ref / MIN2RAD, *ra_ref, *dec_ref);
      for (i = 1; i <= *nf; i++) {
	sprintf(msg, "FWHM%i", i); 
	fget_param_double(&sigma[i], msg);
      }
      fget_param_double(freq0, "REFFREQ");
      for (i = 1; i <= *nf; i++) {
	sprintf(msg, "DATAFN%i", i); 
	fget_param_string(datafn[i], msg);
      }
      close_param_file();
    }
    else 
      printf("ERROR: Cannot read %s - using defaults instead\n", filename);
  }
  
#ifdef MPI
  broadcast_inps(nf, freq, freq0, ni, fieldn, ra, dec, ra_ref, dec_ref, sigma, 
		 datafn, ROOT);
#endif      
  *antmax = 0.;
  for(i = 1; i <= *nf; i++) {
    sigma[i] *= DEG2RAD;
    sigma[i] /= (2.*sqrt(2.*log(2.))); /* sigma of primary beam */
				       /* FT sigma -> 1./(2 \pi \sigma)*/
    antwdth[i] = 2./(PI*sigma[i]* sqrt(2.));
    if (rank == ROOT)
      printf("  1/e diameter of aperture at %6.2f  GHz = %6.2f wavelengths\n", 
	     freq[i], antwdth[i]);
    if (antwdth[i] > *antmax) *antmax = antwdth[i];
  }

  return;
}

/* 
 * read in various telescope parameters
 */
void read_inps(int *nf, double *freq, double *freq0,
	       int *ni, int *fieldn, 
	       double *ra, double *dec, double *ra_ref, double *dec_ref,
	       double *sigma, double *antwdth,
	       double *antmax, char datafn[NFMAX+1][FILENAME_LEN], 
	       int rank, int npe)
{
  double rah, ram, ras, decd, decm, decs;
  int    status, i;

  init_inps(nf, freq, freq0, ni, fieldn, ra, dec, ra_ref, dec_ref, 
	    sigma, antwdth, antmax, datafn);

  if (rank == ROOT) {
    printf("\n");
    printf("Input number of indpendent mosaics (max. %i): ", NFMAX);
    scanf("%i", ni);
    printf("Input number of observed datasets (max. %i): ", NFMAX);
    scanf("%i", nf);
    if (*nf > NFMAX) myerror("ERROR: too many data sets - recompile program!");
    printf("Input observing frequencies (GHz): ");
    for (i = 1; i <= *nf; i++) 
      scanf("%lf", &freq[i]);
    if (*ni > 1) {
      printf("Input no of mosaic the fields belong to (1, ..., %i): ", *ni);
      for (i = 1; i <= *nf; i++) 
	scanf("%i", &fieldn[i]);
    }
    else {
      for (i = 1; i <= *nf; i++) 
	fieldn[i] = 1;
    }
    *ra_ref = 0.; *dec_ref = 0.;
    for (i = 1; i <= *nf; i++) {
      printf("Enter RA (h min s) and DEC (deg min s) of field centre: ");
      scanf("%lf %lf %lf %lf %lf %lf", & rah, & ram, & ras, 
	    & decd, & decm, & decs);

      /* convert to radians */
      sla_ctf2r(rah, ram, ras, &ra[i], &status); 
      sla_daf2r(decd, decm, decs, &dec[i], &status); 
      *ra_ref += ra[i]; *dec_ref += dec[i];
    }
    *ra_ref /= *nf; *dec_ref /= *nf;
    for (i = 1; i <= *nf; i++) 
      printf("  Field centre %2i: %10.5f %10.5f arcmin, %10.5f %10.5f rad\n",
	     i, ra[i] / MIN2RAD, dec[i] / MIN2RAD, ra[i], dec[i]);
    printf("  Reference centre:  %10.5f %10.5f arcmin, %10.5f %10.5f rad\n",
	   *ra_ref / MIN2RAD, *dec_ref / MIN2RAD, *ra_ref, *dec_ref);

    printf("Input FWHM of primary beam at each frequency (degrees): ");
    for (i = 1; i <= *nf; i++) 
      scanf("%lf", &sigma[i]);
    
    printf("Input reference frequency for calculation (GHz): ");
    scanf("%lf", freq0);

    printf("\n");
    printf("Input visibilities data file for each frequency: ");
    for(i = 1; i <= *nf; i++) {
      get_param_filename(datafn[i], "", "r");
    }
  }

#ifdef MPI
  broadcast_inps(nf, freq, freq0, ni, fieldn, ra, dec, ra_ref, dec_ref, sigma, 
		 datafn, ROOT);
#endif      

  /* calculate sigma of primary and 1/e diameter of antenna */
  printf("\n");
  *antmax = 0.;
  for(i = 1; i <= *nf; i++) {
    sigma[i] *= DEG2RAD;
    sigma[i] /= (2.*sqrt(2.*log(2.))); /* sigma of primary beam */
                                       /* FT sigma -> 1./(2 \pi \sigma)*/
    antwdth[i] = 2./(PI*sigma[i]* sqrt(2.));
    if (rank == ROOT)
      printf("  1/e diameter of aperture at %6.2f  GHz = %6.2f wavelengths\n", 
	     freq[i], antwdth[i]);
    if (antwdth[i] > *antmax) *antmax = antwdth[i];
  }

  return;
}
/*-----*-----------------------------------------------------------------*/

/*
 * determine minimum and maximum baseline lengths from data file
 */
void uvlimits(int nf, char datafn[NFMAX+1][FILENAME_LEN], int *nv,
	      double *uvmin, double *uvmax, int rank, int npe)
{
  int    i, idum;
  double uvdist, upos, vpos, vr, vi, rms;
  char   input_line[256];
  FILE   *fp;

  /* read data files to find min and max uv distances*/
  *uvmin = 1.e10;
  *uvmax = 0.;
  *nv = 0;
  if (rank == ROOT) {
    for(i = 1; i <= nf; i++) {
      fp = fopen(datafn[i], "r");
      if (!fp)
	printf("ERROR: Could not load %s\n", datafn[i]);
      while (fgets(input_line, 250, fp) != NULL) {
	if ((input_line[0] != '*')) {
	  if (sscanf(input_line, "%i %lf %lf %lf %lf %lf\n", 
		     &idum, &upos, &vpos, &vr, &vi, &rms)) {
	    (*nv)++;
	    uvdist = sqrt(upos*upos + vpos*vpos);
	    if (uvdist < *uvmin) *uvmin = uvdist;
	    if (uvdist > *uvmax) *uvmax = uvdist;
	  }
	  else
	    printf("Warning: Cannot parse string: %s\n", input_line);
	}
      }
      fclose(fp);
    }
  }

#ifdef MPI
  MPI_Bcast(nv, 1, MPI_INT, ROOT, MPI_COMM_WORLD); 
  MPI_Bcast(uvmin, 1, MPI_DOUBLE, ROOT, MPI_COMM_WORLD); 
  MPI_Bcast(uvmax, 1, MPI_DOUBLE, ROOT, MPI_COMM_WORLD); 
#endif  

  return;
}
/*-----*-----------------------------------------------------------------*/

/*
 * choose number and position of annuli for PS estimation
 */
void fannuli(char filename[], double uvmin, double uvmax, double antmax, 
	     double *r, int *na, int rank, int npe)
{
  int  i, autoflag;
  double  umin, umax, ustep;
  char    id[30];

  autoflag = 1;
  /* write out automatic u range and number of annuli*/
  if (rank == ROOT) {
    umin = uvmin - antmax/2.;
    if (umin < 0.) umin = 0.;
    umax = uvmax + antmax/2.;
    *na = (int)((umax-umin)/antmax);
    if (*na > NAMAX) myerror("ERROR: too many annuli - recompile program");;
    ustep = (umax - umin)/((double)*na);
    for (i = 1; i <= *na+1; i++) {
      r[i] = umin + ustep*(i-1);
    }
    
    /* use different u range and number of annuli if required */
    printf("\n");
    if (!open_param_file(filename)) {
      fget_param_int(&autoflag, "UVAUTOFLAG");
      /*
       * autoflag = 1: automatic bins
       * autoflag = 0: uniformly spaced bins given by umin and umax
       * autoflag = -1: unevenly spaced bins given by bin edges
       */
      if (autoflag != 1) {
	fget_param_int(na, "NBINS");
	if (autoflag == 0) {
	  fget_param_double(&umin, "UVMIN");
	  fget_param_double(&umax, "UVMAX");
	  ustep = (umax - umin)/((double)*na);
	  for (i = 1; i <= *na+1; i++) {
	    r[i] = umin + ustep*(i-1);
	  }
	}
	if (autoflag == -1) {
	  for (i = 1; i <= *na+1; i++) {
	    sprintf(id, "BINLIMIT%d", i);
	    fget_param_double(&r[i], id);
	  }
	  umin = r[1]; umax = r[*na+1];
	  ustep = (umax - umin)/((double)*na);
	}
      }
      
      close_param_file();
    }
    else 
      printf("ERROR: Could not open parameter file!\n");
    if (*na > NAMAX) myerror("ERROR: too many annuli - recompile " 
			     "program with different NAMAX in vsa.h!");
    printf("\n");
    printf("  uv-range measured from %7.2f to %7.2f wavelengths.\n", 
	   uvmin, uvmax);
    printf("  u-range = %6.2f to %6.2f wavelengths.\n", umin, umax);
    printf("  number of bins = %i\n", *na);
    printf("  width of bin = %6.2f wavelengths.\n", ustep);
    printf("\n");
    for (i = 1; i <= *na; i++)
      printf("  Bin %2i:  (%8.4f %8.4f)\n", i, r[i], r[i+1]);
  }

#ifdef MPI
  MPI_Bcast(na, 1, MPI_INT, ROOT, MPI_COMM_WORLD);
  MPI_Bcast(r+C_OFFSET, *na+1, MPI_DOUBLE, ROOT, MPI_COMM_WORLD);
#endif

  return;
}

void annuli(double uvmin, double uvmax, double antmax, double *r, int *na,
	    int rank, int npe)
{
  int  i, autoflag;
  double  umin, umax, ustep;

  /* write out automatic u range and number of annuli*/
  if (rank == ROOT) {
    printf("\n");
    printf("  Power spectrum will be calculated in several bins.\n");
    printf("\n");

    umin = uvmin - antmax/2.;
    if (umin < 0.) umin = 0.;
    umax = uvmax + antmax/2.;
    *na = (int)((umax-umin)/antmax);
    printf("  uv-range measured from %7.2f to %7.2f wavelengths.\n", 
	   uvmin, uvmax);
    printf("  estimate power spectrum from %7.2f to %7.2f wavelengths.\n", 
	   umin, umax);
    printf("  number of bins = %i\n", *na);
    if (*na > NAMAX) myerror("ERROR: too many annuli - recompile program");;
    ustep = (umax-umin)/((double)*na);
    printf("  width  of bins = %6.2f wavelengths", ustep);
    
    /* use different u range and number of annuli if required*/
    printf("\n");
    printf("Use automatic u-range and bin width above?\n"
	   "  (-1 - enter bins manually, 0 - no, 1 - yes): ");
    scanf("%i", &autoflag);
    if (autoflag == 0) {
      printf("Input lower and upper limits on u (wavelengths): ");
      scanf("%lf %lf", &umin, &umax);
      printf("Input number of bins in this range (%i max.): ", NAMAX);
      scanf("%i", na);
      if (*na > NAMAX) myerror("ERROR: too many annuli - recompile " 
			       "program with different NAMAX in vsa.h!");
      ustep = (umax - umin)/((double)*na);
      printf("\n");
      printf("  u-range = %6.2f to %6.2f wavelengths.\n", umin, umax);
      printf("  number of bins = %i\n", *na);
      ustep = (umax - umin)/((double)*na);
      printf("  width of bin = %6.2f wavelengths.\n", ustep);
      printf("\n");
    }
    for(i = 1; i <= *na+1; i++) {
      r[i] = umin + ustep*(i-1);
    }
    if (autoflag == -1) {
      for(i = 1; i <= *na+1; i++) {
	printf("Enter left edge of bin %i (or right edge of %i): ", i, i-1);
	scanf("%lf", &r[i]);
      }
    }

    getchar();
    for (i = 1; i <= *na; i++)
      printf("  Bin %2i:  (%8.4f %8.4f)\n", i, r[i], r[i+1]);
  }
  
#ifdef MPI
  MPI_Bcast(na, 1, MPI_INT, ROOT, MPI_COMM_WORLD);
  MPI_Bcast(r+C_OFFSET, *na+1, MPI_DOUBLE, ROOT, MPI_COMM_WORLD);
#endif

  return;
}
/*-----*-----------------------------------------------------------------*/

/*
 * read visibility data from ASCII file in "vis" format or "wis" format
 * NOTE: the sixth column in the vis-file is the noise rms on real
 *       or imaginary part, and NOT on the modulus of the visibility
 */
void read_data(int nf, int *freqn,
	       double *u, double *v, double *w,
	       double *visr, double *visi, double *rms, int *nvmax,
	       double *freq, double umin, double umax, 
	       char datafn[NFMAX+1][FILENAME_LEN], 
	       int rank, int npe)
{
  int     i, idum, count, nv, wis_flag;
  double  upos, vpos, wpos, vr, vi, rmserr, uvdist;
  char    data_input[256], msg[256];
  FILE    *fp;

  nv = 0;

  if (rank == ROOT) {
    /*...begin loop over observed frequencies*/
    for(i = 1; i <= nf; i++) {
      count = 0;
      
      /* 
       * check whether files have ".wis" or ".wbn" format (7 columns
       * incl. w component) instead of "vis" format (6 columns only)
       */
      wis_flag = 0;
      if (strstr(datafn[i], ".wis") || strstr(datafn[i], ".wbn"))
	wis_flag = 1;

      fp = fopen(datafn[i], "r");
      if (!fp)
	printf("ERROR: Couldn't open %s!\n", datafn[i]);
      
      /* read data and noise*/
      while (fgets(data_input, 250, fp) != NULL) {

	/* skip comments */
	if (data_input[0] == '*')
	  continue;

	if (!wis_flag) {	/* 6 columns in data file */
	  if (sscanf(data_input, "%i %lf %lf %lf %lf %lf", 
		     &idum, &upos, &vpos, &vr, &vi, &rmserr) != 6) {
	    sprintf(msg, "Missing values, line %i in %s\n", count+1, 
		    datafn[i]); 
	    myerror(msg);
	  }
	  wpos = 0.;
	}
	else {			/* 7 columns incl w-component */
	  if (sscanf(data_input, "%i %lf %lf %lf %lf %lf %lf", 
		     &idum, &upos, &vpos, &wpos, &vr, &vi, &rmserr) != 7) {
	    sprintf(msg, "Missing values, line %i in %s\n", count+1, 
		    datafn[i]); 
	    myerror(msg);
	  }
	}
	
	/* ignore data point if outside region of PS estimation*/
	uvdist = sqrt(upos*upos + vpos*vpos);
	if ((uvdist < umin) || (uvdist >= umax))
	  continue;
	
	count++;
	nv++;
	if (nv > *nvmax) {
	  printf("ERROR: more than %i data points\n", *nvmax);
	  return;
	}
	u[nv] = upos; v[nv] = vpos; w[nv] = wpos;
	visr[nv] = vr; visi[nv] = vi; rms[nv] = rmserr; 
	freqn[nv] = i;
      }
      
      fclose(fp);
      
      printf("  At %6.2f GHz read %5i visibilities from %s.\n", 
	     freq[i], count, datafn[i]);
    } /*...close loop over observed frequencies*/
    
    /* write out total number of visibilities*/
    if (nf > 1) {
      printf("\n");
      printf("  Total number of visibilities = %6i.\n", nv);
    }
    if (nv < *nvmax)
      printf("  Only used %6i out of %6i read visibilities.\n", nv, *nvmax);
    *nvmax = nv;
  }
#ifdef MPI
  MPI_Bcast(nvmax, 1, MPI_INT, ROOT, MPI_COMM_WORLD);
  MPI_Bcast(freqn+C_OFFSET, *nvmax, MPI_INT, ROOT, MPI_COMM_WORLD);
  MPI_Bcast(u+C_OFFSET, *nvmax, MPI_DOUBLE, ROOT, MPI_COMM_WORLD);
  MPI_Bcast(v+C_OFFSET, *nvmax, MPI_DOUBLE, ROOT, MPI_COMM_WORLD);
  MPI_Bcast(w+C_OFFSET, *nvmax, MPI_DOUBLE, ROOT, MPI_COMM_WORLD);
  MPI_Bcast(visr+C_OFFSET, *nvmax, MPI_DOUBLE, ROOT, MPI_COMM_WORLD);
  MPI_Bcast(visi+C_OFFSET, *nvmax, MPI_DOUBLE, ROOT, MPI_COMM_WORLD);
  MPI_Bcast(rms+C_OFFSET, *nvmax, MPI_DOUBLE, ROOT, MPI_COMM_WORLD);
#endif

  return;
}

void broadcast_min_params(int *iusemin, int *itol, double *tol, 
			  double *lambda, int *itmax, int root)
{
#ifdef MPI
  MPI_Bcast(iusemin, 1, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(itol, 1, MPI_INT, root, MPI_COMM_WORLD);
  MPI_Bcast(tol, 1, MPI_DOUBLE, root, MPI_COMM_WORLD);
  MPI_Bcast(lambda, 1, MPI_DOUBLE, root, MPI_COMM_WORLD);
  MPI_Bcast(itmax, 1, MPI_INT, root, MPI_COMM_WORLD);
#endif

  return;
}

/*
 * read parameters required for minimisation algorithms
 */
void read_min_params(int *iusemin, int *itol, double *tol, double *lambda,
		     int *itmax, int rank, int npe)
{
  if (rank == ROOT) {
    printf("\n");
    printf("Minimiser - Powell (0), simpow (1), or amoeba (2): ");
    scanf("%i", iusemin);
    printf("Set convergence of function (1) or minimum (0): ");
    scanf("%i", itol);
    printf("Input fractional convergence tolerance: ");
    scanf("%lf", tol);
    printf("Input charistic scale-length: ");
    scanf("%lf", lambda);
    *itmax = 100;
    if (*iusemin == MIN_POWELL) {
      printf("Input maximum number of Powell iterations: ");
      scanf("%i", itmax);
    }
    if (*iusemin == MIN_SIMPOW) {
      printf("Input maximum number of simpow iterations: ");
      scanf("%i", itmax);
    }
  }

#ifdef MPI
  broadcast_min_params(iusemin, itol, tol, lambda, itmax, ROOT);
#endif

  return;
}

void fread_min_params(char filename[],
		      int *iusemin, int *itol, double *tol, double *lambda,
		      int *itmax, int rank, int npe)
{
  *iusemin = 0;
  *itol = 1;
  *tol = 1.e-3;
  *lambda = 0.1;
  *itmax = 100;
  if (rank == ROOT) {
    if (!open_param_file(filename)) {
      fget_param_int(iusemin, "MINIMISER");
      fget_param_int(itol, "TOLFLAG");
      fget_param_double(tol, "TOLERANCE");
      fget_param_double(lambda, "LAMBDA");
      fget_param_int(itmax, "MAXITER");
      close_param_file();    
    }
  }

#ifdef MPI
  broadcast_min_params(iusemin, itol, tol, lambda, itmax, ROOT);
#endif

  return;
}

/*-----*-----------------------------------------------------------------*/

/*
 * write message to log file and flush to disk
 */
int log_file(char filename[], char error_text[])
{
  FILE *flog;
  time_t tp;

  flog = fopen(filename, "a");
  if (! flog)
    return 1;
  time(& tp);
  fprintf(flog, "%s  - %s\n", ctime(&tp), error_text);
  fflush(flog);
  fclose(flog);

  return 0;
}


/*
 * write likelihood estimates to ASCII file and Fortran binary file
 */
void write_like(char outputfn[FILENAME_LEN], double *rbins, int na, 
		double cpsmin, double cpsmax, int ncps,
		double gpsmin, double gpsmax, int ngps,
		double spmin, double spmax, int nsp, double *likeval,
		int rank, int npe)
{
  FILE   *fp;
  int    il, i;

  if (rank == ROOT) {
    if ((fp = fopen(outputfn, "w+"))) {
      fprintf(fp, "%18.11f %18.11f %18.11f\n", 
	      2.*PI*rbins[1], 2.*PI*rbins[na+1], (double)na);
      fprintf(fp, "%18.11f %18.11f %18.11f\n", spmin,  spmax,  (double)nsp);
      fprintf(fp, "%18.11f %18.11f %18.11f\n", cpsmin, cpsmax, (double)ncps);
      fprintf(fp, "%18.11f %18.11f %18.11f\n", gpsmin, gpsmax, (double)ngps);
      for (il = 1; il <= na*nsp*ngps*ncps; il++)
	fprintf(fp, "%18.11f\n", likeval[il]);
      /* extra information not contained in original ".like" files */
      /* list number of extra variables for each dimension */
      fprintf(fp, "%6i %6i %6i %6i\n", na+1, 0, 0, 0);
      for (i = 1; i <= na+1; i++)
	fprintf(fp, "%18.11f\n", rbins[i]);

      fprintf(fp, "# likelihood file produced by MADCOW version > 1.6\n");
      
      fclose(fp);
    }
    else 
      printf("ERROR: Cannot write to %s\n", outputfn);
  }

  return;
}

void read_like(char outputfn[FILENAME_LEN], double *rbins, int na, 
	       double cpsmin, double cpsmax, int ncps,
	       double gpsmin, double gpsmax, int ngps,
	       double spmin, double spmax, int nsp, double *likeval,
	       int rank, int npe)
{
  FILE   *fp;
  int    i, il;
  double rdum1, rdum2, rdum3;
  int    idum1, idum2, idum3, idum4, newflag;
  double eps = 1.e-6;
  char   input_line[256];

  newflag = 0;
  if (rank == ROOT) {
    if ((fp = fopen(outputfn, "r"))) {
      /* skip comments */
      do {
	fgets(input_line, 250, fp);
      } while (input_line[0] == '#');
      /* l_min, l_max, na */
      sscanf(input_line, "%lf %lf %lf", &rdum1, &rdum2, &rdum3); 
      if ((fabs((rdum1 - 2*PI*rbins[1]) / rdum1) > eps)
	  || (fabs((rdum2 - 2*PI*rbins[na+1]) / rdum2) > eps))
	myerror("Wrong bin size in like file");
      if ((int)rdum3 != na)
	myerror("Wrong number of bins in like file.");

      /* spmin,  spmax,  (double)nsp */
      fscanf(fp, "%lf %lf %lf\n",  &rdum1, &rdum2, &rdum3);  
      
      /* cpsmin, cpsmax, (double)ncps */
      fscanf(fp, "%lf %lf %lf\n",  &rdum1, &rdum2, &rdum3);  
      if ((fabs((rdum1 - cpsmin) / rdum1) > eps) 
	  || (fabs((rdum2 - cpsmax) / rdum2) > eps))
	myerror("Wrong CMB step size in like file");
      if ((int)rdum3 != ncps)
	myerror("Wrong number of CMB steps in like file.");
      
      /*gpsmin, gpsmax, (double)ngps)*/
      fscanf(fp, "%lf %lf %lf\n",  &rdum1, &rdum2, &rdum3);  

      for (il = 1; il <= na*nsp*ngps*ncps; il++)
	fscanf(fp, "%lf\n", &likeval[il]);
      
      /* the new section in like file */
#if 0
      /* skip comments */
      do {
	fgets(input_line, 250, fp);
      } while ((fgets(input_line, 250, fp) != NULL) && (input_line[0] == '#'));
#endif

      if (fscanf(fp, "%i %i %i %i\n", &idum1, &idum2, &idum3, &idum4) != EOF) {
	if (idum1 > na + 1)
	  myerror("Too many bins in like file");
	for (i = 1; i <= idum1; i++) {
	  fscanf(fp, "%lf\n", &rdum1); 
	  if (fabs(rdum1 - rbins[i]) / rdum1 > eps)
	    myerror("Wrong bin edge in like file");
	}      
      }

      fclose(fp);
    }
    else 
      printf("ERROR: Cannot read from %s\n", outputfn);
  }
#ifdef MPI
  MPI_Bcast(likeval+C_OFFSET, na*nsp*ngps*ncps, MPI_DOUBLE, 
	    ROOT, MPI_COMM_WORLD);
#endif  

  return;
}
