/*
 * wrapper routines to make calls to LAPACK less irksome
 * KSM, 2001
 */
/*  Time-stamp: <28 Jan 2002 13:20 GMT like_lapack.c maising> */


#include "stdio.h"
#include "math.h"
#include "my_malloc2.h"
#include "clapack.h"
#include "like_lapack.h"

#define DIVIDE_CONQUER 1	/* use LAPACK divide & conquer algorithm? */

#define MALLOC1 malloc_1dmatrix
#define MALLOC1I malloc_1imatrix
#define FREE1 free_1dmatrix
#define FREE1I free_1imatrix

/*************************************************************************
 * calculate Cholesky decomposition of matrix A 
 * PARAMS:
 *     a    [I,O] n1 x n2 dimensional matrix
 *     n    [I] dimension of a
 */
void cholesky_decompose(double **a, int n)
{
  int uplo2, info;
  uplo2 = 0;			/* uplo = 'U' --> lower triangular in C */

  CDPOTRF(&uplo2, &n, a[1]+1, &n, &info);
  if (info != 0)
    printf("ERROR in dpotrf == %i\n", info);
  return;
}


/*************************************************************************
 * compute Cholesky decomposition of Matrix in LAPACK packed storage.
 * the matrix is assumed to be lower triangular 
 * PARAMS:
 *     a    [I,O] n1 x n2 dimensional matrix
 *     n    [I]   dimension of a
 */
void sp_cholesky_decompose(double *a, int n)
{
  int uplo2, info;

  uplo2 = 1;			/* UPLO='L'; lower trianguar */
  CDPPTRF(&uplo2, &n, a+1, &info);
  if (info != 0)
    printf("ERROR in dpptrf: %i\n", info);
  return;
}


/*************************************************************************
 * solve symmetric eigenvalue problem Ax=lambda x; 
 * the input matrices are destroyed upon exit
 * PARAMS
 *     a    [I,O]  on entry, the matrix A; on exit, the eigenvectors
 *     eval [O]    the eigenvalues lambda
 *     nc   [I]    dimension of A and B
 */
void sym_eigenval(double **a, double *eval, int nc)
{
  {
    int uplo2, jobz2, info;
    double *work;
    int lwork, *iwork, liwork;

    uplo2 = 1;			/* UPLO='L' */
    jobz2 = 1;			/* JOBZ='V' */
    lwork = 1 + 6*nc + 2*nc*nc;
    liwork = 3+5*nc;
    work = MALLOC1(lwork);
    iwork = MALLOC1I(liwork);
    CDSYEVD( &jobz2, &uplo2, &nc, a[1]+1, &nc, eval+1, work+1, &lwork,
	     iwork+1, &liwork, &info );
    if (info !=0)
      printf("ERROR in dsyevd == %i\n", info);
    else {
      printf("  Optimal lwork  %8.0f (used %i)\n", work[1], lwork);
      printf("  Optimal liwork %8i (used %i)\n", iwork[1], liwork);
    }
    FREE1(work);
    FREE1I(iwork);
  }
  return;
}

/*************************************************************************
 * solve general eigenvalue problem Ax=lambda Bx; 
 * the input matrices are destroyed upon exit
 * PARAMS
 *     a    [I,O]  on entry, the matrix A; on exit, the eigenvectors
 *     b    [I,O]  on entry, the matrix B; on exit, the Cholesky decomposition
 *                 B = L L^T
 *     eval [O]    the eigevalues lambda
 *     nc   [I]    dimension of A and B
 */
void gen_sym_eigenval(double **a, double **b, double *eval, int nc)
{
  int itype, uplo2, jobz2, info;
  
  itype = 1;			/*  A*x = (lambda)*B*x */
  jobz2 = 1;			/* jobz = 'V'; compute eigenvectors */
  uplo2 = 1;			/* uplo = 'L'; lower triangles */
  info = 0;
  
  printf("  Solving general symmetric eigenvalue problem...\n");
#if (! DIVIDE_CONQUER)
  /* use LAPACK routine dsygv */
  {
    double *work;
    int lwork;
    
    lwork = 3*nc;
    work = MALLOC1(lwork);
    CDSYGV(&itype, &jobz2, &uplo2, &nc,
	   a[1]+1, &nc,  b[1]+1, &nc, eval+1, work+1, &lwork, &info);
    if (info != 0) 
      printf("ERROR in cdsygv = %i\n", info); 
    FREE1(work);
  }
#else
  /* use dsygvd, divide/conquer algorithm */
  {
    double *work;
    int lwork, liwork, *iwork, lgn;
    
    lgn = (int)(log(nc)/log(2.))+1;
    lwork = 1 + (5+2*lgn)*nc + 2*nc*nc;
    work = MALLOC1(lwork);
    liwork = 3+5*nc;
    iwork = MALLOC1I(liwork);
    CDSYGVD(&itype, &jobz2, &uplo2, &nc, a[1]+1, &nc, b[1]+1, &nc, 
	    eval+1, work+1, &lwork, iwork+1, &liwork, &info);
    if (info != 0) {
      printf("ERROR in cdsygvd = %i\n", info); 
    }
    else {
      printf("  Optimum lwork  %8.0f (used %i)\n", work[1], lwork);
      printf("  Optimum liwork %8i (used %i)\n", iwork[1], liwork);
    }
    FREE1(work);
    FREE1I(iwork);
  }
#endif

  return;
}

/*************************************************************************
 * solve general eigenvalue problem Ax=lambda Bx; 
 * calculate only eigenvalues and vectors in a given interval
 * the input matrices are destroyed upon exit
 * PARAMS
 *     a      [I]  on entry, the matrix A; on exit destroyed
 *     b    [I,O]  on entry, the matrix B; on exit, the Cholesky decomposition
 *                 B = L L^T
 *     nc     [I]  dimension of A and B
 *     valmin [I]  the minimum value for calculated eigenvalues 
 *     valmax [I]  the maximum value for calculated eigenvalues 
 *     eigval [O]  the eigenvalues lambda
 *     eigvec [O]  the eigenvectors 
 *     nev    [O]  the number of eigenvalues and vectors calculated
 */
void gen_sym_exp_eigenval(double **a, double **b, int nv,
			  double valmin, double valmax,
			  double *eigval, double **eigvec, int *nev)
{
  int itype, uplo2, jobz2, range2, info;

  if (nv == 0)
    myerror("matrices have zero dimension");
  if (nv == 1) {
    eigval[1] = 1; eigvec[1][1] = 1.;
    return;
  }

  /* other LAPACK arguments */
  itype = 1;			/*  A*x = (lambda)*B*x */
  jobz2 = 1;			/* jobz = 'V'; compute eigenvectors */
  uplo2 = 1;			/* uplo = 'L'; lower triangles */
  info = 0;

  printf("  Solving general symmetric eigenvalue problem...\n");
  {
    double *work;
    int lwork, liwork, *iwork, *ifail, idum;
    char c;
    double abstol;

    range2 = 1;	    /* all eigenvalues in the half-open interval (VL,VU] */
                    /* will be found */
    c = 'S';
    abstol = 2. * CDLAMCH(&c);
    ifail = MALLOC1I(nv);
    liwork = 5*nv;
    iwork = MALLOC1I(liwork);

    lwork = -1;
    work = MALLOC1(1);
    CDSYGVX(&itype, &jobz2, &range2, &uplo2, &nv, a[1]+1, &nv, b[1]+1, &nv,
	    &valmin, &valmax, &idum, &idum,  &abstol, 
	    nev, eigval+1, eigvec[1]+1, &nv, work+1,
	    &lwork, iwork+1, ifail+1, &info);
    if (info != 0)
      printf("ERROR in work space query for cdspgvd = %i\n", info);
    lwork = (int)(work[1]);
    FREE1(work);
    work = MALLOC1(lwork);
    CDSYGVX(&itype, &jobz2, &range2, &uplo2, &nv, a[1]+1, &nv, b[1]+1, &nv,
	    &valmin, &valmax, &idum, &idum,  &abstol, 
	    nev, eigval+1, eigvec[1]+1, &nv, work+1,
	    &lwork, iwork+1, ifail+1, &info);
    if (info != 0) {
      printf("ERROR in cdspgvx = %i\n", info); 
    }
    else {
      printf("  Optimum lwork  %8.0f (used %i)\n", work[1], lwork);
      printf("  Optimum liwork %8i (used %i)\n", iwork[1], liwork);
    }
    FREE1(work);
    FREE1I(iwork);
    FREE1I(ifail);
  }

  return;
}

/*************************************************************************
 * solve general eigenvalue problem Ax=lambda Bx; 
 * the input matrices are in LAPACK packed storage and destroyed upon exit
 * PARAMS
 *     ac     [I]  on entry, the matrix A; on exit destroyed
 *     bc   [I,O]  on entry, the matrix B; on exit, the Cholesky decomposition
 *                 B = L L^T
 *     nv     [I]  dimension of A and B
 *     eigval [O]  the eigenvalues lambda
 *     eigvec [O]  the eigenvectors
 */
void sp_gen_sym_eigenval(double *ac, double *bc, int nv, 
			 double *eigval, double **eigvec)
{
  int itype, uplo2, jobz2, info;

  /* other LAPACK arguments */
  itype = 1;			/*  A*x = (lambda)*B*x */
  jobz2 = 1;			/* jobz = 'V'; compute eigenvectors */
  uplo2 = 1;			/* uplo = 'L'; lower triangles */
  info = 0;

  printf("  Solving general symmetric eigenvalue problem...\n");
#if (!DIVIDE_CONQUER)
  {
    double *work;
    work = MALLOC1(3*nv);
    CDSPGV(&itype, &jobz2, &uplo2, &nv,
	   ac+1, bc+1, eigval+1, eigvec[1]+1, &nv, work+1, &info);
    if (info != 0) 
      printf("ERROR in cdspgv = %i\n", info); 
    FREE1(work);
  }
#else
  {
    double *work;
    int lwork, liwork, *iwork, lgn;

    lgn = (int)(log(nv)/log(2.))+1;
    lwork = 1 + (5+2*lgn)*nv + 2*nv*nv;
    work = MALLOC1(lwork);
    liwork = 3+5*nv;
    iwork = MALLOC1I(liwork);
    CDSPGVD(&itype, &jobz2, &uplo2, &nv,
	    ac+1, bc+1, eigval+1, eigvec[1]+1, &nv, work+1, &lwork,
	    iwork+1, &liwork, &info);
    if (info != 0) {
      printf("ERROR in cdspgvd = %i\n", info); 
    }
    else {
      printf("  Optimum lwork  %8.0f (used %i)\n", work[1], lwork);
      printf("  Optimum liwork %8i (used %i)\n", iwork[1], liwork);
    }
    FREE1(work);
    FREE1I(iwork);
  }
#endif

  return;
}

/*************************************************************************
 * solve general eigenvalue problem Ax=lambda Bx; 
 * calculate only eigenvalues and vectors in a given interval
 * the input matrices are in LAPACK packed storage and destroyed upon exit
 * PARAMS
 *     a      [I]  on entry, the matrix A; on exit destroyed
 *     b    [I,O]  on entry, the matrix B; on exit, the Cholesky decomposition
 *                 B = L L^T
 *     nc     [I]  dimension of A and B
 *     valmin [I]  the minimum value for calculated eigenvalues 
 *     valmax [I]  the maximum value for calculated eigenvalues 
 *     eigval [O]  the eigenvalues lambda
 *     eigvec [O]  the eigenvectors 
 *     nev    [O]  the number of eigenvalues and vectors calculated
 */
void sp_gen_sym_exp_eigenval(double *a, double *b, int nv,
			     double valmin, double valmax,
			     double *eigval, double **eigvec, int *nev)
{
  int itype, uplo2, range2, jobz2, info, idum;

  /* other LAPACK arguments */
  itype = 1;			/*  A*x = (lambda)*B*x */
  jobz2 = 1;			/* jobz = 'V'; compute eigenvectors */
  range2 = 1;			/* range = 'V'; all eigenvalues in the */
				/* interval (VL,VU] will be found. */
  uplo2 = 1;			/* uplo = 'L'; lower triangles */
  info = 0;

  printf("  Solving general symmetric eigenvalue problem...\n");
  {
    double *work;
    int lwork, liwork, *iwork, *ifail;
    char c;
    double abstol;

    c = 'S';
    abstol = 2. * CDLAMCH(&c);
    lwork = 8*nv;
    liwork = 5*nv;
    work = MALLOC1(lwork);
    iwork = MALLOC1I(liwork);
    ifail = MALLOC1I(nv);
    CDSPGVX(&itype, &jobz2, &range2, &uplo2, &nv, a+1, b+1,
	    &valmin, &valmax, &idum, &idum, &abstol, 
	    nev, eigval+1, eigvec[1]+1, &nv, work+1, iwork+1,
	    ifail+1, &info);
    if (info != 0)
      printf("ERROR in cdspgvx = %i\n", info); 

    FREE1(work);
    FREE1I(iwork);
    FREE1I(ifail);
  }

  return;
}

