/* 
 * Various matrix and vector operation and conversion routines for
 * various storage formats
 *   - C style array 
 *   - compressed column storage (SLAP)
 *   - LAPACK compressed storage
 */
/*  Time-stamp: <11 Jul 2002 18:38 BST like_matvec.c maising> */

#include <stdio.h>
#include <math.h>
#include "my_malloc2.h"
#include "like_matvec.h"

#define MALLOC1 malloc_1dmatrix
#define MALLOC2 malloc_2dmatrix

/*-----------------------------------------------------------------------*/

/*************************************************************************
 * convert upper or lower triangular  matrix to LAPACK packed storage 
 * PARAMS:
 *     p0  [I]  input matrix
 *     nv  [I]  dimension of p0
 *     lu  [I]  lu == 0: upper triangle, lu ==1: lower triangle
 *     pp2 [O]  pointer to LAPACK packed matrix
 */
void p0_p2(double **p0, int nv, int lu, double **pp2)
{
  int ndim, i, j;

  ndim = nv*(nv+1)/2;
  (*pp2) = MALLOC1(ndim);
  switch (lu) {
  case 0: 		/* upper triangular j >= i*/
    for (i = 1; i <= nv; i++)
      for (j = i; j <= nv; j++)
	(*pp2)[i+j*(j-1)/2] = p0[i][j];
    break;

  case 1:			/* lower triangular j <= i*/
    for (i = 1; i <= nv; i++)
      for (j = 1; j <= i; j++)
	(*pp2)[i+(2*nv-j)*(j-1)/2] = p0[i][j];
    break;
  default: printf("ERROR in p0_p2: Matrix type not recognised.\n");
    break;
  }

  return;
}

/*************************************************************************
 * convert matrix from LAPACK packed storage to ordinary storage 
 * PARAMS:
 *     p2  [I]  input matrix in LAPACK packed storage
 *     nv  [I]  dimension of p2
 *     lu  [I]  lu == 0: upper triangle, lu ==1: lower triangle
 *     pp0 [O]  pointer to output matrix
 */
void p2_p0(double *p2, int nv, int lu, double ***pp0)
{
  int i, j;

  (*pp0) = MALLOC2(nv, nv);
  switch (lu) {
  case 0:                       /* upper triangular j >= i*/
    for (i = 1; i <= nv; i++) {
      for (j = 1; j < i; j++)
	(*pp0)[i][j] = 0.;
      for (j = i; j <= nv; j++)
	(*pp0)[i][j] = p2[i+j*(j-1)/2];
    }
    break;
  
  case 1:			/* lower triangular j <= i*/
    for (i = 1; i <= nv; i++) {
      for (j = 1; j <= i; j++)
	(*pp0)[i][j] = p2[i+(2*nv-j)*(j-1)/2];
      for (j = i+1; j <= nv; j++)
	(*pp0)[i][j] = 0.;
    }
    break;

  default: printf("ERROR in p2_p0: Matrix type not recognised.\n");
    break;
  }
  return;
}

/*************************************************************************
 * convert matrix from SLAP to LAPACK packed storage (lower triangular)
 * PARAMS:
 *     icn [I]  column index array for compressed column storage
 *     irn [I]  row index array
 *     c   [I]  array elements
 *     nv  [I]  dimension of matrix
 *     lu  [I]  lu == 0: upper triangle, lu ==1: lower triangle
 *     pp2 [O]  pointer to output matrix in LAPACK packed storage
 */
void p1_p2(int *icn, int *irn, double *c, int nv, int lu, double **pp2)
{
  int i, j, k, ndim;

  ndim = nv*(nv+1)/2;
  (*pp2) = MALLOC1(ndim);
  for (k = 1; k <= ndim; k++) 
    (*pp2)[k] = 0.;
  switch (lu) {
  case 0:			/* upper triangular */
    for (j = 1; j <= nv; j++)
      for (k = icn[j]; k <= icn[j+1]-1; k++) {
	i = irn[k];
	(*pp2)[i+j*(j-1)/2] = c[k];
      }
    break;

  case 1: 		        /* lower triangular */
    for (j = 1; j <= nv; j++)
      for (k = icn[j]; k <= icn[j+1]-1; k++) {
	i = irn[k];
	(*pp2)[i+(j-1)*(2*nv-j)/2] = c[k];
      }
    break;

  default: printf("ERROR in p1_p2: Matrix type not recognised.\n");
    break;
  }
  return;
}

/*************************************************************************
 * convert matrix from SLAP to conventional storage (symmetric matrix) 
 * PARAMS:
 *     icn [I]  column index array for compressed column storage
 *     irn [I]  row index array
 *     c   [I]  array elements
 *     nv  [I]  dimension of matrix
 *     lu  [I]  lu == 2: symmetrise matrix; lu == 0, 1: general matrix 
 *     pp0 [O]  pointer to output matrix
 */
void p1_p0(int *icn, int *irn, double *c, int nv, int lu, double ***pp0)
{
  int i, j, k;

  (*pp0) = MALLOC2(nv, nv);

  for (i = 1; i <= nv; i++) 
    for (j = 1; j <= nv; j++) 
      (*pp0)[i][j] = 0.;

  switch (lu) {
  case 0: case 1:		/* general matrix */
    for (j = 1; j <= nv; j++)
      for (k = icn[j]; k <= icn[j+1]-1; k++) {
	i = irn[k];
	(*pp0)[i][j] = c[k];
      }
    break;

  case 2: 		/* symmetric matrix */
    for (j = 1; j <= nv; j++)
      for (k = icn[j]; k <= icn[j+1]-1; k++) {
	i = irn[k];
	(*pp0)[i][j] = c[k];
	if (i != j)
	  (*pp0)[j][i] = c[k];
      }
    break;

  default: printf("ERROR in p1_p0: Matrix type not recognised.\n");
    break;
  }
  return;
}

/*************************************************************************
 * derived from p1_p0; however, in this case, only the the sourtheastern
 * nv x nv part of a (2*nv) x (2*nv) matrix is created
 * PARAMS:
 *     icn [I]  column index array for compressed column storage
 *     irn [I]  row index array 
 *     c   [I]  array elements
 *     nv  [I]  dimension of submatrix; the total matrix is 2*nv
 *     lu  [I]  lu == 2: symmetrise matrix; lu == 0, 1: general matrix 
 *     pp0 [O]  pointer to output matrix (nv x nv)
 */
void p1_p0_2(int *icn, int *irn, double *c, int nv, int lu, double ***pp0)
{
  int i, j, k;

  (*pp0) = MALLOC2(nv, nv);

  for (i = 1; i <= nv; i++) 
    for (j = 1; j <= nv; j++) 
      (*pp0)[i][j] = 0.;

  switch (lu) {
  case 0: case 1:		/* general matrix */
    for (j = nv+1; j <= 2*nv; j++)
      for (k = icn[j]; k <= icn[j+1]-1; k++) {
	i = irn[k];
	(*pp0)[i-nv][j-nv] = c[k];
      }
    break;

  case 2: 		/* symmetric matrix */
    for (j = nv+1; j <= 2*nv; j++)
      for (k = icn[j]; k <= icn[j+1]-1; k++) {
	i = irn[k];
	(*pp0)[i-nv][j-nv] = c[k];
	if (i != j)
	  (*pp0)[j-nv][i-nv] = c[k];
      }
    break;

  default: printf("ERROR in p1_p0_2: Matrix type not recognised.\n");
    break;
  }
  return;
}

/*************************************************************************
 * multiply a matrix A in compressed column storage by a vector x
 * PARAMS:
 *     y    [I,O] resulting vector
 *     nv   [I] dimension of matrix
 *     icn  [I] column index array for compressed column storage
 *     irn  [I] row index array
 *     a    [I] array elements
 *     nz   [I] number of elements in sparse array
 *     x    [I] vector
 */
void p1_mult_matvec(double *y, int nv, 
		    int *icn, int *irn, double *a, int nz, double *x)
{
  int i, j, k;

  for (k = 1; k <= nv; k++)
    y[k] = 0.;

  for (j = 1; j <= nv; j++)
    for (k = icn[j]; k <= icn[j+1]-1; k++) {
      i = irn[k];
      if (i == j)
	y[i] += a[k] * x[i];
      else {
	y[i] += a[k] * x[j];
	y[j] += a[k] * x[i];
      }
    }
  return;
}

/*************************************************************************
 * multiply a matrix A in LAPACK compressed storage by a vector x
 * PARAMS:
 *     y    [I,O] resulting vector
 *     nv   [I] dimension of matrix
 *     a    [I] matrix elements in LAPACK storage
 *     x    [I] vector
 */
void p2_mult_matvec(double *y, int nv, double *a, double *x)
{
  int i, j, k;

  for (i = 1; i <= nv; i++)
    y[i] = 0.;
  i = 1;
  j = 1;
  y[1] = a[1] * x[1];

  for (k = 2; k <= nv * (nv+1)/2; k++) {
    i++;
    if (i > nv) {
      j++;
      i = j;
      y[i] += a[k] * x[i];
    }
    else {
      y[i] += a[k] * x[j];
      y[j] += a[k] * x[i];
    }
  }
    
  return;
}

/*************************************************************************
 * multiply a matrix A by a vector x
 * PARAMS:
 *     y    [I,O] resulting vector
 *     n1   [I] dimension of y
 *     a    [I] n1 x n2 dimensional matrix
 *     n2   [I] dimension of x
 *     x    [I] vector
 */
void mult_matvec(double *y, int n1, double **a, int n2, double *x)
{
  int i, j;

  for (i = 1; i <= n1; i++) {
    y[i] = 0.;
    
    for (j = 1; j <= n2; j++)
      y[i] += a[i][j] * x[j];
  }
  return;
}

/*************************************************************************
 * multiply a matrix A by matrix X: Y = A X
 * PARAMS:
 *     y    [I,O] resulting n1 x n2 - dimensional output matrix
 *     n1   [I] 1st dimension of a
 *     n2   [I] 2nd dimension of a
 *     a    [I] n1 x n3 dimensional matrix
 *     n3   [I] 2nd dimension of x
 *     x    [I] n3 x n2 dimensional matrix
 */
void mult_matmat(double **y, int n1, int n2, double **a, int n3, double **x)
{
  int i, j, k;

  for (i = 1; i <= n1; i++) 
    for (j = 1; j <= n2; j++) {
      y[i][j] = 0.;
      for (k = 1; k <= n3; k++)
	y[i][j] += a[i][k] * x[k][j];
    }
  return;
}


/*************************************************************************
 * calculate the Frobenius norm \sqrt{\sum_{ij} a_{ij}^2} of a matrix
 * PARAMS:
 *     a    [I] n1 x n2 dimensional matrix
 *     n1   [I] 1st dimension of a
 *     n2   [I] 2nd dimension of a
 * RETURNS
 *     the norm of a
 */
double frobenius_norm(double **a, int n1, int n2)
{
  int i, j;
  double norm;

  norm = 0.;
  for (i = 1; i <= n1; i++)
    for (j = 1; j <= n2; j++) 
      norm += a[i][j] * a[i][j];
  return sqrt(norm);
}

