/* 
 * wrapper routines to make ScaLAPACK calls less irksome
 * KSM, 2001
 */ 
/*  Time-stamp: <28 Jan 2002 13:21 GMT like_scalapack.c maising> */


#include <stdio.h>
#include <math.h>
#include "my_malloc2.h"
#include "cscalapack.h"
#include "like_scalapack.h"

#define ROOT 0
#define MAX(x, y) (((x) > (y) ? (x) : (y)))

/*************************************************************************
 * solve general eigenvalue problem Ax=lambda Bx using ScaLAPACK routine
 * pdsygvx()
 * calculate only eigenvalues and vectors in a given interval
 * the input matrices are destroyed upon exit
 * PARAMS
 *     a      [I]  on entry, the matrix A; on exit destroyed
 *     b    [I,O]  on entry, the matrix B; on exit, the Cholesky decomposition
 *                 B = L L^T
 *     nc     [I]  dimension of A and B
 *     valmin [I]  the minimum value for calculated eigenvalues 
 *     valmax [I]  the maximum value for calculated eigenvalues 
 *     eigval [O]  the eigenvalues lambda
 *     eigvec [O]  the eigenvectors 
 *     nev    [O]  the number of eigenvalues and vectors calculated
 */
void sca_gen_sym_exp_eigenval(double **a, int desca[], double **b, int descb[],
			      double valmin, double valmax, 
			      double *eigval, int nv,
			      double **z, int descz[], int *nevec)
{
  static int  itype = 1;        /*  A*x = (lambda)*B*x */
  static char jobz = 'V';	/* 'V':  Compute eigenvalues and vectors. */
  static char range = 'V';	/* 'V' */
  static char uplo = 'L';	/* 'L':  Lower triangles are stored. */
  int info, neval;
  int rank, npe, icontxt;
  int nprow, npcol;
  int prow, pcol;
  double *work, *gap;
  int lwork, liwork, *iwork, *ifail;
  int *iclustr;

  if (nv == 0)
    myerror("matrices have zero dimension");
  if (nv == 1) {
    eigval[1] = 1; z[1][1] = 1.;
    return;
  }

  icontxt = desca[1];
  cblacs_pinfo(&rank, &npe);
  cblacs_gridinfo(icontxt, &nprow, &npcol, &prow, &pcol);
#if 0
  printf("icontxt %i rank %i np %i %i prow pcol %i %i\n", icontxt, 
	 rank, nprow, npcol, prow, pcol);
#endif

  /* other LAPACK arguments */
  info = 0;
  if (rank == ROOT)
    printf("  Solving general symmetric eigenvalue problem...\n");
  
  ifail = malloc_1imatrix(nv);
  iclustr = malloc_1imatrix(2*nprow*npcol);
  gap = malloc_1dmatrix(nprow*npcol);
#if 0
  printf("Allocating %i %i %i\n", nv, 2*nprow*npcol, nprow*npcol);
#endif
  
  {
    int nb, nn, np0, mq0, md, nnp;

    nb = desca[4];
    nn = MAX(nb, nv); nn = MAX(nn, 2);
    np0 = cnumroc(nn, nb, 0, 0, nprow);
    md = MAX(nb, nv); md = MAX(md, 2);
    mq0 = cnumroc(md, nb, 0, 0, npcol);
    lwork = 5*nv+ MAX(5*nn, np0 * mq0 + 2 * nb * nb) +
      (int)ceil((double)nv/(double)(nprow*npcol))*nn;
    nnp = MAX(nv, nprow*npcol); nnp = MAX(nnp, 4);
    liwork = 6*nnp;
  }
  work = malloc_1dmatrix(lwork);
  iwork = malloc_1imatrix(liwork);
  lwork = -1;
  liwork = -1;
  cpdsygvx(itype, &jobz, &range, &uplo, nv, 
	   a[C_OFFSET]+C_OFFSET, 1, 1, desca, 
	   b[C_OFFSET]+C_OFFSET, 1, 1, descb,
	   valmin, valmax, 0, 0, -1., &neval, nevec, eigval+C_OFFSET, -1.,
	   z[C_OFFSET]+C_OFFSET, 1, 1, descz,
	   work+C_OFFSET, lwork, iwork+C_OFFSET, liwork, 
	   ifail+C_OFFSET, iclustr+C_OFFSET, gap+C_OFFSET, &info);
  if (info != 0)
    printf("ERROR in work space query for cpdsygvx = %i\n", info);
  lwork = (int)(work[1]);
  liwork = iwork[1];
  free_1dmatrix(work);
  work = malloc_1dmatrix(lwork);
  free_1imatrix(iwork);
  iwork = malloc_1imatrix(liwork);

  cpdsygvx(itype, &jobz, &range, &uplo, nv, 
	   a[C_OFFSET]+C_OFFSET, 1, 1, desca, 
	   b[C_OFFSET]+C_OFFSET, 1, 1, descb,
	   valmin, valmax, 0, 0, -1., &neval, nevec, eigval+C_OFFSET, -1.,
	   z[C_OFFSET]+C_OFFSET, 1, 1, descz,
	   work+C_OFFSET, lwork, iwork+C_OFFSET, liwork, 
	   ifail+C_OFFSET, iclustr+C_OFFSET, gap+C_OFFSET, &info);
  if (info != 0) {
    if (rank == ROOT) {
      printf("ERROR in cpdsygvx = %i:\n", info); 
      if (info % 2 != 0)
	printf("  One or more eigenvectors failed to converge. PLEASE REPORT TO scalapack@cs.utk.edu!\n");
      else if ((info / 2) % 2 != 0)
	printf("  Some eigenvectors could not be reorthogonalised.\n");
      else if ((info / 4) % 2 != 0)
	printf("  Not enough workspace to compute all eigenvectors.\n");
      else if ((info / 8) % 2 != 0)
	printf("  Some eigenvalues could not be computed. PLEASE REPORT TO scalapack@cs.utk.edu!\n");
      else if ((info / 16) % 2 != 0)
	printf("  Matrix B was not positive definite.\n");
    }
  }
#if 0
  else 
    if (rank == ROOT) {
      printf("  Optimum lwork  %8.0f (used %i)\n", work[1], lwork);
      printf("  Optimum liwork %8i (used %i)\n", iwork[1], liwork);
    }
#endif
  if (neval != *nevec)
    if (rank == ROOT)
      printf("ERROR: less eigenvectors (%i) than eigenvalues (%i)!\n", 
	     *nevec, neval);
  free_1dmatrix(work);
  free_1dmatrix(gap);
  free_1imatrix(iwork);
  free_1imatrix(ifail);
  free_1imatrix(iclustr);

  return;
}
