/* 
 * program to calculate maximum likelihood estimates of parameters      
 * for an interferometer observation of CMB and Galaxy fluctuations at  
 * several frequencies 
 */

/*  Time-stamp: <13 Nov 2002 11:55 GMT maxlike.c maising> */

/* HISTORY
 *   ??-??-1997  MPH   original version in Fortran
 *   09-04-2001  KSM   C translation
 */

#include <stdio.h>
#include <math.h>

#ifdef MPI
#include <mpi.h>
#endif
#if defined (SCALAPACK)
#include "cscalapack.h"
#endif

#include "globals.h"
#include "my_malloc2.h"
#include "like_temp.h"
#include "like_io.h"
#include "like_integrals.h"
#include "like_covariance.h"
#include "like_func.h"
#include "nrminimizer2.h"
#include "hessian.h"
#include "matinvert.h"

#define PI 3.14159265358979323844
#define ROOT 0			/* MPI root process */

/*-----*-----------------------------------------------------------------*/

/*
 * minimise function with a set of minimising routines
 * PARAMS
 *     func    [I]  function to be minimised
 *     x     [I,O]  initial and final parameter vector
 *     np      [I]  number of parameters
 *     iusemin [I]  flag to identify mimiser to use (0 - powell, 2 - amoeba)
 *     itol    [I]  use relative or absolute convergence?
 *     tol     [I]  convergence tolerance
 *     lambda  [I]  typical scale
 *     itmax   [I]  maximum number of iterations
 *     rank    [I]  rank of processor given by MPI; if MPI is not used, 0
 *     npe     [I]  number of processors; if MPI is not used set to 1
 *     rank    [I]
 */
double minimise(double (*func)(double []), double *x, int np,
	      int iusemin, int itol, double tol, double lambda, 
	      int itmax, int rank, int npe)
{
  double **xi, **p, *y, fret;
  int    i, j, iter;

  p = malloc_2dmatrix(np+1, np); xi = malloc_2dmatrix(np, np);
  y = malloc_1dmatrix(np+1);
    
  switch (iusemin) {
  case MIN_POWELL:   /* ... use NR powell method*/

    if (rank == ROOT)
      printf("setting up initial direction set ... \n");
    for(i = 1; i <= np; i++) {
      for(j = 1; j <= np; j++) {
	xi[i][j] = 0.;
      }
      xi[i][i] = lambda;
    }

    if (rank == ROOT)
      printf("Calling powell ...\n");
    xpowell(x, xi, np, tol, &iter, itmax, &fret, itol, func);
    if ((rank == ROOT) && (iter < itmax))
      printf("Powell converged after %i iterations.\n", iter);
    break;

  case MIN_SIMPOW:   /* ... use simpow method*/
    
    if (rank == ROOT)
      printf("setting up initial direction set ... \n");
    for(i = 1; i <= np; i++) {
      for(j = 1; j <= np; j++) {
	xi[i][j] = 0.;
      }
      xi[i][i] = lambda;
    }
    if (rank == ROOT)
      printf("Calling simpow ...\n");
#if 0
    simpow(x,xi,np,npmax,tol,iter,itmax,fret,itol);
#else
    printf("Sorry, simpow not implemented.\n");
#endif
    if ((rank == ROOT) && (iter < itmax))
      printf("Simpow converged after %i iterations.\n", iter);
    break;
    
  case MIN_AMOEBA: /* ... use NR amoeba method*/
    if (rank == ROOT)
      printf("setting up initial simplex ...\n");
    for(j = 1; j <= np; j++)
      p[1][j] = x[j];
    y[1] = func(x);

    for(i = 2; i <= np+1; i++) {
      for(j = 1; j <= np; j++) {
	p[i][j] = p[1][j];
	if (j == (i-1)) 
	  p[i][j] = p[1][j] + lambda;
	x[j] = p[i][j];
      }
      y[i] = func(x);
    }
      
    if (rank == ROOT)
      printf("Calling amoeba ...\n");
    xamoeba(p, y, np, tol, func, &iter, itol);
    if (rank == ROOT)
      printf("Amoeba converged after %i function evaluations.\n", iter);
    for(j = 1; j <= np; j++)
      x[j] = p[1][j];
    fret = y[1];
    break;
  default:
    if (rank == ROOT)
      printf("Unkown minimiser %i\n", iusemin);
    break;
  }
  free_2dmatrix(p); free_2dmatrix(xi); free_1dmatrix(y);

  return fret;
}

int main(int argc, char *argv[])
{
  int    *freqn, fieldn[NFMAX+1];
  double *u, *v, *w, *visr, *visi, *vis, *rms;
  double rbins[NAMAX+1+1];
  double freq[NFMAX+1], sigma[NFMAX+1], antwdth[NFMAX+1];
  double cps1d[NAMAX+1], gps1d[NAMAX+1], sp;
  double *x;

  int    galflag, sparseflag, maximise_flag, hessian_flag, imagflag;
  int    ia, ita, read_ps_flag, irestart, saveflag;
  int    nf, ni, na, np, i;
  double antmax, uvmin, uvmax;
  char   datafn[NFMAX+1][FILENAME_LEN];
  double ra[NFMAX+1], dec[NFMAX+1];
  double ra_ref, dec_ref;
  double fnorm;

  int    nv;				/* number of visibilities */
  double freq0;  /* rec freq */
  int    rank = 0;
  int    npe  = 1;
  int    std_input = 1;
  char   infile[FILENAME_LEN];
  char   outputfn[FILENAME_LEN];
  
#ifdef SCALAPACK
  int    icontxt;
  char   major_order='C';
  int    prow, pcol, nprow, npcol;
  int    mypnum, nprocs;
#endif

  /*
   * read command line arguments
   */
  if (argc > 1) {
    sprintf(infile, "%s", argv[1]);
    std_input = 0;
  }

#ifdef MPI
  /*
   * Initialise MPI.
   */
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &npe);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Process %i is running...\n", rank);
#endif
#ifdef SCALAPACK
  /*
   * Initialise ScaLAPACK.
   */
  cblacs_pinfo(&mypnum, &nprocs);
  if ((nprocs != npe) || (mypnum != rank))
    printf("  Rank and processor number do not match on process %i!\n", rank);

  nprow = (int)sqrt(nprocs) + 1;
  while (nprocs%nprow > 0) nprow--;
  npcol = nprocs / nprow;
  if (rank == ROOT)
    printf("  Using %i x %i processor grid.\n", nprow, npcol);
  cblacs_get(-1, 0, &icontxt);
  cblacs_gridinit(&icontxt, &major_order, nprow, npcol);
  cblacs_gridinfo(icontxt, &nprow, &npcol, &prow, &pcol); 
#endif

  /* initialise checks for memory leaks */
  init_mymalloc();

  /* read input parameters*/
  if (std_input)
    read_inps(&nf, freq, &freq0, &ni, fieldn,
	      ra, dec, &ra_ref, &dec_ref, sigma, antwdth, &antmax,
	      datafn, rank, npe);
  else
    fread_inps(infile, &nf, freq, &freq0, &ni, fieldn, 
	       ra, dec, &ra_ref, &dec_ref, sigma, antwdth, &antmax,
	       datafn, rank, npe);

  /* find limits on uv positions*/
  uvlimits(nf, datafn, &nv, &uvmin, &uvmax, rank, npe);
  if (rank == ROOT)
    printf("  Minimum and maximum u: %f %f\n", uvmin, uvmax);

  /* determine annuli to use*/
  if (std_input)
    annuli(uvmin, uvmax, antmax, rbins, &na, rank, npe);
  else
    fannuli(infile, uvmin, uvmax, antmax, rbins, &na, rank, npe);

  /* 
   * read in input data (assuming units are lambda and Jy), plot uv tracks
   */
  if (rank == ROOT)
    printf("  Number of visibilities: %i.\n", nv);
  freqn = malloc_1imatrix(nv); 
  u = malloc_1dmatrix(nv); v = malloc_1dmatrix(nv); w = malloc_1dmatrix(nv);
  visr = malloc_1dmatrix(nv); visi = malloc_1dmatrix(nv); 
  rms = malloc_1dmatrix(nv);

  /* read visibility data from file */
  read_data(nf, freqn, u, v, w, visr, visi, rms, &nv, freq, 
	    rbins[1], rbins[na+1], datafn, rank, npe);

  /* copy real and imaginary parts to one vector */
  vis = malloc_1dmatrix(2*nv);
  for (i = 1; i <= nv; i++) {
    vis[i] = visr[i];
    vis[i+nv] = visi[i];
  }
  free_1dmatrix(visr); free_1dmatrix(visi); 

  /* read in parameters controlling likelihood calculation */ 
  if (std_input)
    read_like_params(&sparseflag, &maximise_flag, &hessian_flag, 
		     &galflag, &saveflag, outputfn, &irestart, rank, npe);
  else
    fread_like_params(infile, &sparseflag, &maximise_flag,
		      &hessian_flag, &galflag, 
		      &saveflag, outputfn, &irestart, rank, npe);

  /* check whether there is more then one pointing (mosaicing) */
  imagflag = 0;
  for (i = 2; i <= nf; i++)
    if ((ra[i] != ra[1]) || (dec[i] != dec[1]))
      imagflag = 1;
  if (rank == ROOT) {
    if (imagflag)
      printf("  Using mosaicing algorithm...\n");
    else 
      printf("  No mosaicing required...\n");
  }

  if (std_input)
    read_initial_ps(cps1d, gps1d, &sp, na, galflag, &read_ps_flag,
		    rank, npe);
  else
    fread_initial_ps(infile, cps1d, gps1d, &sp, na, galflag, 
		     &read_ps_flag, rank, npe);
  if (rank == ROOT) {
    printf("  Initial CMB PS: "); 
    for (ita = 1; ita <= na; ita++)
      printf("%g ", cps1d[ita]);
    printf("\n");
    printf("  Initial GAL PS: "); 
    for (ita = 1; ita <= na; ita++)
      printf("%g ", gps1d[ita]);
    printf("\n");
  }


  /* 
   * calculate overlap integrals 
   */
  if (saveflag & INT_SAVE_ONLY) {
    integral_struct integral_data;
    get_integrals(&integral_data, freqn, u, v, w, nv, 
		  ra, dec, ra_ref, dec_ref, freq, antwdth, sigma, nf,
		  rbins, na, 1, 1, ni, 1, imagflag, saveflag, 
		  "integrals", rank, npe);
    if (!(saveflag & INT_STORE_MEMORY))
      freeintegrals(&integral_data);
  
    goto label_End;
  }


  /* 
   * create initial parameter vector
   */
  np = 2*na+1;			/* na CMB params, na GAL, 1 spectral idx */
  x = malloc_1dmatrix(np);
  for (ia = 1; ia <= na; ia++) {
    x[ia] = cps1d[ia] * 1.e9;
    x[ia+na] = gps1d[ia] * 1.e9;
  }
  x[2*na+1] = sp;
  if (!galflag)
    np = na;
    
  if (rank == ROOT)
    printf("  Initialising data for calls of loglike function.\n");
  init_loglike_func(np, na, sparseflag, ra, dec, ra_ref, dec_ref,
		    freqn, u, v, w, vis, rms, nv, freq, antwdth, sigma, nf, 
		    freq0, ps_conversion(T_CMB, freq0), rbins, 1, fieldn, ni, 
		    imagflag, saveflag,  1, rank, npe);

  /* 
   * maximise log(likelihood) (actually minimise -log(likelihood))
   */
  if (maximise_flag) {

    /* rescale function value at starting point to 1 */
    set_fnorm(0.);
    fnorm = loglike_func(x) - 1.;
    set_fnorm(fnorm);

    if (rank == ROOT) 
      printf("\n  Beginning maximisation of log[likelihood].\n\n");

    do {

      reset_loglike_func_count();

      /* 
       * choose minimiser, convergence tolerance and scale-lengths
       * and perform maximisation of log likelihood
       */
      {
	int    iusemin, itol;
	double lambda, tol;
	int    itmax;

	if (std_input)
	  read_min_params(&iusemin, &itol, &tol, &lambda, &itmax, rank, npe);
	else
	  fread_min_params(infile, &iusemin, &itol, &tol, &lambda, &itmax, 
			   rank, npe);

	minimise(loglike_func, x, np, iusemin, itol, tol, lambda, itmax, 
		 rank, npe);
	for (ia = 1; ia <= np; ia++)
	  x[ia] = fabs(x[ia]);
      }
      
      /* write out number of evaluations required*/
      if (rank == ROOT)
	printf("Number  of likelihood evaluations = %i\n", 
	       query_loglike_func_count());
      
      /* 
       * write out peak of likelihood
       */
      if (rank == ROOT) {
	printf("\nMaximum of log(likelihood) occurs at:\n\n");
	for(ita = 1; ita <= na; ita++) {
	  double leff;
	  
	  leff = PI * (rbins[ita]+rbins[ita+1]);
	  printf("In annulus %2i centred at l = %8.2f\n", ita, leff);
	  printf("\n");
	  printf("  CMB power spectrum = %g\n", x[ita]);
	  if (galflag)
	    printf("  GAL power spectrum = %g\n", x[na+ita]);
	}
	if (galflag)
	  printf("  GAL spectral index = %g\n", x[2*na*1]);
	printf("\n");
	
	write_param_log(outputfn, x, x+na, x[2*na+1], rbins, na, galflag,
			query_loglike_func_count(), 1, rank, npe);
      }
      
      /* restart from current maximum likelihood estimates? */
      if (std_input) {
	if (rank == ROOT) {
	  printf("\n");
	  printf("Restart minimiser from this point (1) or not (0): ");
	  scanf("%i", &irestart);
	}
#ifdef MPI
	MPI_Bcast(&irestart, 1, MPI_INT, ROOT, MPI_COMM_WORLD);
#endif
      }
    } while (irestart);

  }
    
  /* 
   * calculate approximate Hessian at likelihood peak 
   * using subroutine hessian()
   */
  if (hessian_flag) {
    double ferr;
    double *diag;
    double **hesian;
    int    i, j;
    
    /* rescale so it equals unity at the maximum */
    set_fnorm(0.);
    fnorm = loglike_func(x) - 1.;
    set_fnorm(fnorm);
    reset_loglike_func_count();

    if (rank == ROOT)
      printf("\n  Calculating Hessian at current func point ...\n\n");

    ferr = 1.e-4;      /* relative error in func */ 

    hesian = malloc_2dmatrix(np, np);
    diag = malloc_1dmatrix(np);
    for(i = 1; i <= np; i++)
      diag[i] = 1.;    /* diagonal scaling */
    hessian(loglike_func, x, np, hesian, diag, ferr);

    if (rank == ROOT) {
      printf("  Hessian matrix is:\n");
      for(i = 1; i <= np; i++)
	for(j = 1; j <= i; j++) { 
	  printf("  H[%2i][%2i] = %8.4f\n", i, j, hesian[i][j]);
	  hesian[j][i] = hesian[i][j];
	}
    }
        
    /* 
     * invert Hessian matrix to obtain covariances
     */
    if (rank == ROOT)
      printf("\n  Inverting Hessian to obtain covariances ...\n\n");
    matinvert(hesian, np);

    if (rank == ROOT) {
      char msg[125];

      printf("Covariance matrix is:\n");
      for(i = 1; i <= np; i++)
	for(j = 1; j <= i; j++) { 
	  sprintf(msg, "  Cov[%2i][%2i] = %8.4f\n", i, j, hesian[i][j]);
	  printf(msg);
	  write_file(outputfn, msg);
	}

      printf("  Correlation matrix is:\n");
      for(i = 1; i <= np; i++) {
	diag[i] = sqrt(fabs(hesian[i][i]));
	for(j = 1; j <= i; j++) {
	  sprintf(msg, "  Cor[%2i][%2i] = %8.4f\n", 
		 i, j, hesian[i][j]/diag[i]/diag[j]);
	  printf(msg);
	  write_file(outputfn, msg);
	}
      }
    }
    
    /* 
     * write out maximum likelihood position and errors to file
     */
    if (rank == ROOT) {
      char msg[125];

      for(i = 1; i <= na; i++) {
	double leff;
	
	leff = PI * (rbins[i]+rbins[i+1]);
	x[i] = fabs(x[i]);
	sprintf(msg, "l, l^2 C_l/2pi, error = %g %g %g\n",
		leff, x[i], diag[i]);
	printf(msg);
	write_file(outputfn, msg);

	if (galflag == 1) {
	  x[na+i] = fabs(x[na+i]);
	  sprintf(msg, "l, l^2 C_l/2pi, error = %g %g %g\n",
		  leff, x[na+i], diag[na+i]);
	  printf(msg);
	  write_file(outputfn, msg);
	}
      }
      if (galflag == 1) {
	sprintf(msg, "Galactic spectral index = %f\n", sp);
	printf(msg);
	write_file(outputfn, msg);
      }
    }

    free_1dmatrix(diag);
    free_2dmatrix(hesian);
  } /* hessian_flag */

  free_1dmatrix(x);

 label_End:
  dealloc_integrals(na, ni);
  free_1imatrix(freqn); 
  free_1dmatrix(u); free_1dmatrix(v); free_1dmatrix(w);
  free_1dmatrix(vis); free_1dmatrix(rms);

  /* check for memory leaks */
  done_mymalloc();

  /*
   * exit ScaLAPACK and MPI
   */
#ifdef SCALAPACK
  if (prow >= 0) {
    cblacs_gridexit(icontxt);
    cblacs_exit(0);
  }
#endif
#if defined(MPI) && (!defined(SCALAPACK))
  MPI_Finalize();
#endif

  return 0;
}
/*-----*-----------------------------------------------------------------*/
