/* 
 * A program to calculate the likelihood for certain parameters      
 * given an interferometer observation of CMB and Galaxy fluctuations
 * at several frequencies;
 * includes mosaicing and w-distortions.
 */
/*  Time-stamp: <12 Nov 2002 18:21 GMT redlike.c maising> */
/*
 * HISTORY
 *     ??-??-97  MPH  single-field Fortran version
 *     ??-??-01  KSM  C version
 */

/*-----*-----------------------------------------------------------------*/
/* 	$Id: redlike.c,v 1.9 2002/11/13 12:13:46 maising Exp $	 */

#ifndef lint
static char vcid[] = "$Id: redlike.c,v 1.9 2002/11/13 12:13:46 maising Exp $";
#endif /* lint */

#include <stdio.h>
#include <math.h>

#if defined(MPI)
#include <mpi.h>
#endif
#if defined (SCALAPACK)
#include "cscalapack.h"
#endif

#include "my_malloc2.h"
#include "globals.h"
#include "like_temp.h"
#include "like_io.h"
#include "like_sort.h"
#include "like_integrals.h"
#include "like_covariance.h"
#include "like_func.h"

#define PI 3.14159265358979323844
#define ROOT 0			/* MPI root process */

/*-----*-----------------------------------------------------------------*/

int main(int argc, char *argv[])
{
  double *u, *v, *w, *visr, *visi, *rms;
  double *u2, *v2, *w2, *vis2, *rms2;
  int    nv, nv2;               /* number of visibilities */
  int    *freqn, *freqn2, nva[NAMAX+1], fieldn[NFMAX+1];
  double rbins[NAMAX+1+1];
  double cps1d[NAMAX+1], gps1d[NAMAX+1], sp;
  double cps1dmax[NAMAX+1], gps1dmax[NAMAX+1], spindmax[NAMAX+1];
  double cps1dold[NAMAX+1], gps1dold[NAMAX+1];

  double chisqmax[NAMAX+1], likemax[NAMAX+1];
  double freq[NFMAX+1], sigma[NFMAX+1], antwdth[NFMAX+1], freq0;
  double ra[NFMAX+1], dec[NFMAX+1];
  double ra_ref, dec_ref;
  double *likeval, *chisqval, *lndetval;

  int    galflag, plotflag, outflag, sparseflag, imagflag;
  int    converge, saveflag, read_ps_flag;  
  int    itsp, itg, itc, ita, iter, ind, idx;
  int    ncps, ngps, nsp, nf, ni, na, i, j, ia, firstbin;
  double antmax, uvmin, uvmax, rmin, rmax;
  double cpsmin, cpsmax, cpsstep;
  double gpsmin, gpsmax, gpsstep;
  double spmin, spmax, spstep;
  char   datafn[NFMAX+1][FILENAME_LEN];
  char   outputfn[FILENAME_LEN], logfn[FILENAME_LEN], parlogfn[FILENAME_LEN];
  char   infile[FILENAME_LEN];
  int    std_input = 1;

  integral_struct integral_data;
  cm_struct cm_data;
  int    rank = 0, npe = 1;
#ifdef SCALAPACK
  int    icontxt;
  char   major_order='C';
  int    prow, pcol, nprow, npcol;
  int    mypnum, nprocs;
#endif

  /*
   * read command line arguments
   */
  if (argc > 1) {
    sprintf(infile, "%s", argv[1]);
    std_input = 0;
  }

#ifdef MPI
  /*
   * Initialise MPI.
   */
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &npe);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Process %i is running...\n", rank);
#endif
#ifdef SCALAPACK
  /*
   * Initialise ScaLAPACK.
   */
  cblacs_pinfo(&mypnum, &nprocs);
  if ((nprocs != npe) || (mypnum != rank))
    printf("  Rank and processor number do not match on process %i!\n", rank);

  nprow = (int)sqrt(nprocs) + 1;
  while (nprocs%nprow > 0) nprow--;
  npcol = nprocs / nprow;
  if (rank == ROOT)
    printf("  Using %i x %i processor grid.\n", nprow, npcol);
  cblacs_get(-1, 0, &icontxt);
  cblacs_gridinit(&icontxt, &major_order, nprow, npcol);
  cblacs_gridinfo(icontxt, &nprow, &npcol, &prow, &pcol); 
#endif

  /* initialise checks for memory leaks */
  init_mymalloc();

  /* read input parameters*/
  if (std_input)
    read_inps(&nf, freq, &freq0, &ni, fieldn, 
	      ra, dec, &ra_ref, &dec_ref, sigma, antwdth, &antmax,
	      datafn, rank, npe);
  else
    fread_inps(infile, &nf, freq, &freq0, &ni, fieldn,
	       ra, dec, &ra_ref, &dec_ref, sigma, antwdth, &antmax,
	       datafn, rank, npe);

  /* find limits on uv positions*/
  uvlimits(nf, datafn, &nv, &uvmin, &uvmax, rank, npe);
  if (rank == ROOT)
    printf("  Minimum and maximum u: %f %f\n", uvmin, uvmax);

  /* determine annuli to use*/
  if (std_input)
    annuli(uvmin, uvmax, antmax, rbins, &na, rank, npe);
  else
    fannuli(infile, uvmin, uvmax, antmax, rbins, &na, rank, npe);

  /* 
   * read in input data (assuming units are lambda and Jy), plot uv tracks
   */
  if (rank == ROOT)
    printf("  Number of visibilities: %i.\n", nv);
  freqn = malloc_1imatrix(nv);
  u = malloc_1dmatrix(nv); v = malloc_1dmatrix(nv); w = malloc_1dmatrix(nv);
  visr = malloc_1dmatrix(nv); visi = malloc_1dmatrix(nv); 
  rms = malloc_1dmatrix(nv);

  /* read visibility data from file */
  read_data(nf, freqn, u, v, w, visr, visi, rms, &nv, freq, 
	    rbins[1], rbins[na+1], datafn, rank, npe);

  /* sort visibilities by radius in uv-plane */
  sort_data(freqn, u, v, w, visr, visi, rms, nv, radius);
  /* count visibilities per bin */
  for (ia = 1; ia <= na; ia++)
    nva[ia] = 0;
  for (j = 1; j <= nv; j++)
    for (ia = 1; ia <= na; ia++)
      if (radius(u[j], v[j], w[j]) >= rbins[ia] && 
	  radius(u[j], v[j], w[j]) < rbins[ia+1])
	nva[ia] ++;

  if (rank == ROOT)
    for (ia = 1; ia <= na; ia++)
      printf("  In bin %2i: %5i visibilities.\n", ia, nva[ia]);

  /* read in limits on CMB and GAL parameters*/
  if (std_input) 
    read_lim(&cpsmin, &cpsmax, &cpsstep, &ncps, &gpsmin, &gpsmax, &gpsstep, 
	     &ngps, &spmin, &spmax, &spstep, &nsp, &galflag, rank, npe);
  else 
    fread_lim(infile,
	      &cpsmin, &cpsmax, &cpsstep, &ncps, &gpsmin, &gpsmax, &gpsstep, 
	      &ngps, &spmin, &spmax, &spstep, &nsp, &galflag, rank, npe);


  /* read in parameters controlling likelihood calculation*/
  if (std_input) 
    read_lfcn_params(& plotflag, &outflag, outputfn,
		     &sparseflag, &converge, &saveflag, rank, npe);
  else
    fread_lfcn_params(infile, & plotflag, &outflag, outputfn,
		      &sparseflag, &converge, &saveflag, rank, npe);
  /* write to log file */
  if (rank == ROOT) {
    sprintf(logfn, "%s.log", outputfn);
    log_file(logfn, vcid);
    for (i = 1; i <= nf; i++)
      log_file(logfn, datafn[i]);
  }

  /* check whether there is more then one pointing (mosaicing) */
  imagflag = 0;
  for (i = 2; i <= nf; i++)
    if ((ra[i] != ra[1]) || (dec[i] != dec[1]))
      imagflag = 1;
#if defined(WCORRECTIONS)
  imagflag = 1;
#endif
  if (rank == ROOT) {
    if (imagflag)
      printf("  Using mosaicing algorithm...\n");
    else 
      printf("  No mosaicing required...\n");
  }

  /* 
   * set initial assumed CMB and GAL power spectrum
   */
  if (rank == ROOT)
    sprintf(parlogfn, "%s.plog", outputfn);

  if (std_input)
    read_initial_ps(cps1dmax, gps1dmax, &sp, na, galflag, &read_ps_flag,
		    rank, npe);
  else
    fread_initial_ps(infile, cps1dmax, gps1dmax, &sp, na, galflag, 
		     &read_ps_flag, rank, npe);

  /*
   * initialise likelihood values
   */
  firstbin = 1;
  iter = 0;
  likeval = malloc_1dmatrix(na*nsp*ngps*ncps);
  chisqval = malloc_1dmatrix(na*nsp*ngps*ncps);
  lndetval = malloc_1dmatrix(na*nsp*ngps*ncps);
  for (i = 1; i <= na*nsp*ngps*ncps; i++)
    likeval[i] = 0.;

  /* 
   * read parameters from previous iteration 
   */
  if (read_ps_flag == 2) {
    if (rank == ROOT)
      printf("  Reading parameters from logfile %s.\n", parlogfn);
    read_param_log(parlogfn, cps1dmax, gps1dmax, 
		   &sp, rbins, na, galflag, &iter, &firstbin, rank, npe);
    firstbin++;		       /* first bin after the last saved bin */
    iter--;		       /* this gets incremented again later */
    if (firstbin > na) {
      firstbin = 1; iter++; 
    }
    read_like(outputfn, rbins, na, cpsmin, cpsmax, ncps, 
	      gpsmin, gpsmax, ngps, spmin, spmax, nsp, likeval, rank, npe);
  }

  if (rank == ROOT) {
    printf("  Initial CMB PS: "); 
    for (ita = 1; ita <= na; ita++)
      printf("%g ", cps1dmax[ita]);
    printf("\n");
  }

  if (rank == ROOT) {
    printf("  Initial GAL PS: "); 
    for (ita = 1; ita <= na; ita++)
      printf("%g ", gps1dmax[ita]);
    printf("\n");
  }

  /* 
   * iterate likelihood calculation
   */
  do {
    iter++;

    /* update CMB and GAL power spectrum */
    for(ita = 1; ita <= na; ita++) {
      cps1d[ita] = cps1dmax[ita];
      gps1d[ita] = gps1dmax[ita];
      cps1dold[ita] = cps1dmax[ita];
      gps1dold[ita] = gps1dmax[ita];
    }

    /* start loop over annuli*/
    for (ita = firstbin; ita <= na; ita++) {
      if (rank == ROOT)
	printf("\n  Beginning analysis of annulus %2i\n\n",ita);

      /* set likelihood values in this bin to zero */
      for (i = (ita-1)*nsp*ngps*ncps+1; i <= ita*nsp*ngps*ncps; i++) {
	likeval[i] = 0.;
	chisqval[i] = 0.;
	lndetval[i] = 0.;
      }

      /* find minimum and maximum radius sensitive to power in this bin */
      rmin = rbins[ita] - antmax;
      rmax = rbins[ita+1] + antmax;

      /* loop over independent mosaics */
      for (ind = 1; ind <= ni; ind++) {

	/* calculate reduced number of visibilities for this bin */
	nv2 = 0;
	for (j = 1; j <= nv; j++)
	  if ((fieldn[freqn[j]] == ind)
	      && (radius(u[j], v[j], w[j]) >= rmin)
	      && (radius(u[j], v[j], w[j]) < rmax))
	    nv2 ++;
	if (rank == ROOT)
	  printf("  Number of observed visibilities = %i\n", nv2);

	/* skip bin if empty */
	if (nv2 == 0) {
	  likemax[ita] = 0.; chisqmax[ita] = 0.; 
	  cps1dmax[ita] = 0.; gps1dmax[ita] = 0.; spindmax[ita] = 0.;
	  continue;
	}

	freqn2 = malloc_1imatrix(nv2);
	u2 = malloc_1dmatrix(nv2); v2 = malloc_1dmatrix(nv2); 
	w2 = malloc_1dmatrix(nv2);
	vis2 = malloc_1dmatrix(2*nv2); rms2 = malloc_1dmatrix(nv2);

	/* copy data */
	i = 0;
	for (j = 1; j <= nv; j++)
	  if ((fieldn[freqn[j]] == ind)
	      && (radius(u[j], v[j], w[j]) >= rmin)
	      && (radius(u[j], v[j], w[j]) < rmax)) {
	    i++;
	    freqn2[i] = freqn[j];
	    u2[i] = u[j]; v2[i] = v[j]; w2[i] = w[j];
	    vis2[i] = visr[j]; vis2[i+nv2] = visi[j]; rms2[i] = rms[j];
	  }

	/*
	 * calculate the integrals required to form the covariance matrix
	 */
	if (rank == ROOT)
	  printf("  Obtaining overlap integrals ...\n");
	get_integrals(&integral_data, freqn2, u2, v2, w2, nv2, 
		      ra, dec, ra_ref, dec_ref, freq, antwdth, sigma, nf,
		      rbins, na, ita, ind, ni, 1, imagflag, saveflag, 
		      "integrals", rank, npe);
	if (rank == ROOT) { 
	  char msg[255];
	  
	  sprintf(msg, "Finished %i integrals for %i visibilities.", 
		  integral_data.nz, integral_data.nv);
	  log_file(logfn, msg);
	}
	if (saveflag & INT_SAVE_ONLY)
	  goto label_freeintegrals;

	/* 
	 * calculate the log(likelihood) ...
	 */
	if (rank == ROOT)
	  printf("\n  Calculating log(likelihood) ...\n\n");
	likemax[ita] = -1.e30;

	/* 
	 * calculate the covariance matrices
	 */
	{
	  int cnz;

	  cnz = countcms(&integral_data);
	  if (rank == ROOT)
	    printf("  Number of entries in covariance matrix: %i\n", cnz);
	  if (rank == ROOT)
	    printf("  Creating covariance matrix...\n");
	  newcms(&cm_data, nv2, cnz);
	  if (rank == ROOT)
	    printf("  Creating index table for covariance matrix...\n");
	  makeidxcms(&cm_data, &integral_data);
	}

	/* update Galactic flux spectral index */
	for(itsp = 1; itsp <= nsp; itsp++) {
	  sp = spmin + spstep * (double)(itsp-1);

	  /* update GAL power-spectrum */
	  for(itg = 1; itg <= ngps; itg++) {
	    gps1d[ita] = gpsmin + gpsstep * (double)(itg-1);

	    /* update CMB power-spectrum */
	    for(itc = 1; itc <= ncps; itc++) {
	      double ll, chisq, lndet;

	      cps1d[ita] = cpsmin + cpsstep * (double)(itc-1);

	      idx = (ita-1)*nsp*ngps*ncps +(itsp-1)*ngps*ncps+(itg-1)*ncps+itc;

	      /*
	       * calculate log[likelihood]
	       */
	      ll = loglike(cps1d, gps1d, na, sp, sparseflag,
			   vis2, rms2, nv2, freq, nf, freq0, 
			   ps_conversion(T_CMB, freq0), &chisq, &lndet,
			   &integral_data, &cm_data, rank, npe);
	      likeval[idx] += ll;
	      chisqval[idx] += chisq;
	      lndetval[idx] += lndet;

	      if (ind == ni) {
		/* write out current parameters and likelihood */
		if ((plotflag) && (rank == ROOT)) {
		  printf("  CMB PS, GAL PS, Index = %g %g %g\n",
			 cps1d[ita], gps1d[ita], sp);
		  printf("  lndet,chisq,loglike   = %8.3f %8.3f %8.3f\n",
			 0.5*lndetval[idx], 0.5*chisqval[idx], likeval[idx]);
		}

		/* store maximum likelihood in present annulus */
		if (likeval[idx] > likemax[ita]) {
		  likemax[ita] = likeval[idx];
		  chisqmax[ita] = chisqval[idx];
		  cps1dmax[ita] = cps1d[ita];
		  gps1dmax[ita] = gps1d[ita];
		  spindmax[ita] = sp;
		}
	      }
	    } /*...close CMB power-spectrum loop*/
	  } /*...close GAL power-spectrum loop*/
	} /*...close spectral index loop*/

	freeidxcms(&cm_data);
	freecms(&cm_data);

      label_freeintegrals:
	if (! (saveflag & INT_STORE_MEMORY))
	  freeintegrals(&integral_data);
	free_1imatrix(freqn2); 
	free_1dmatrix(u2); free_1dmatrix(v2); free_1dmatrix(w2);
	free_1dmatrix(vis2); free_1dmatrix(rms2);
      } /* loop over independent mosaics */

      /* reset CMB and GAL power spectrum to use for next annulus*/
      cps1d[ita] = cps1dold[ita];
      gps1d[ita] = gps1dold[ita];

      if (rank == ROOT)
	printf("  Maximum in iteration %2i, bin %2i occurs at %g %g %g\n",
	       iter, ita, cps1dmax[ita], gps1dmax[ita], spindmax[ita]);
#if 1
      /* log results to file - I find this useful on supercomputers 
       *  that don't flush their output regularly
       */
      if (rank == ROOT) {
	char msg[255];

	sprintf(msg, "Iteration %2i bin %2i  max %g %g %g", iter, ita, 
		cps1dmax[ita], gps1dmax[ita], spindmax[ita]);
	log_file(logfn, msg);
      }
#endif

      /*
       * write out likelihood values to file
       */
      write_param_log(parlogfn, cps1dmax, gps1dmax, spindmax[ita],
		      rbins, na, galflag, iter, ita, rank, npe);
      if (outflag)
	write_like(outputfn, rbins, na, cpsmin, cpsmax, ncps,
		   gpsmin, gpsmax, ngps, spmin, spmax, nsp,
		   likeval, rank, npe);

    } /* ... loop over annuli */

    /* 
     * write out peak of likelihood
     */
    if (rank == ROOT) {
      printf("\nMaximum of log(likelihood) occurs at:\n\n");
      for(ita = 1; ita <= na; ita++) {
	double leff;

	leff = PI * (rbins[ita]+rbins[ita+1]);
	printf("In annulus %2i centred at l = %8.2f\n", ita, leff);
	printf("\n");
	printf("  CMB power spectrum = %8.2g\n", cps1dmax[ita]);
	printf("  GAL power spectrum = %8.2g\n", gps1dmax[ita]);
	printf("  GAL spectral index = %8.2g\n", spindmax[ita]);
	printf("  Chi-sq,log[likehd] = %8.2f %8.2f\n",
	       0.5 * chisqmax[ita], 0.5*likemax[ita]);
	printf("\n");
      }
    }

    /* 
     * restart from current maximum likelihood estimates
     */
    firstbin = 1;
    if (converge) {
      /* check for convergence of likelihood maximum.
       * if either the change in one parameter is greater than the step size,
       * or the total change is more than (na)*0.6, start new iteration 
       */
      double eps = 0.6 * na;
      double error, ec, eg;
      int    flag;

      error = 0.;
      flag = 0;
      for (ita = 1; ita <= na; ita ++) {
	ec = 0.; eg = 0.;
	if (cpsstep > 0.)
	  ec = fabs(cps1dmax[ita] - cps1dold[ita]) / cpsstep;
	if (gpsstep > 0.)
	  eg = fabs(gps1dmax[ita] - gps1dold[ita]) / gpsstep;
	if ((ec > 1.001) || (eg > 1.001))
	  flag = ita;
	error += ec + eg;
      }
      if (rank == ROOT)
	printf("  Convergence of maximum: %g %g, %i\n", error, eps, flag);
      if (!((fabs(error) > eps) || (flag)) || (iter > 10))
	converge = 0;
    }
  } while (converge);

  if (rank == ROOT)
    printf("  Likelihood iteration converged after iteration %i.\n", iter);

  /* 
   * free allocated memory
   */
  reset_integrals(na, ni);
  free_1dmatrix(likeval);
  free_1dmatrix(chisqval);
  free_1dmatrix(lndetval);
  free_1imatrix(freqn); free_1dmatrix(u); free_1dmatrix(v); free_1dmatrix(w);
  free_1dmatrix(visr); free_1dmatrix(visi); free_1dmatrix(rms);

  /* check for memory leaks */
  done_mymalloc();

  /*
   * exit ScaLAPACK and MPI
   */
#ifdef SCALAPACK
  if (prow >= 0) {
    cblacs_gridexit(icontxt);
    cblacs_exit(0);
  }
#endif
#if defined(MPI) && (!defined(SCALAPACK))
  MPI_Finalize();
#endif

  return 0;
}
/*-----*-----------------------------------------------------------------*/
